/*
 * File:   HashT.h
 * Author: peter
 *
 * Created on 03. Dezember 2012, 13:54
 */

#include <math.h>
#include <cstdlib>
#include <time.h>
#include "Position.h"

#ifndef HASHT_H
#define	HASHT_H

class HashT {
public:
    HashT() {};
    HashT(unsigned int size):size(size) {
        hashMask=(size-1)>>2;
        ageMask=17870283321406128128UL; // highest 5 bits
        depthMask=558446353793941504UL; // next 5 bits
        map = new unsigned long[size];
        clearHash();
        
        srand(time(NULL));
        for (int i=0;i<64;i++) {
            kb[i]=((unsigned long)rand()<<32)|rand();
            qb[i]=((unsigned long)rand()<<32)|rand();
            bb[i]=((unsigned long)rand()<<32)|rand();
            nb[i]=((unsigned long)rand()<<32)|rand();
            rb[i]=((unsigned long)rand()<<32)|rand();
            pb[i]=((unsigned long)rand()<<32)|rand();
            kw[i]=((unsigned long)rand()<<32)|rand();
            qw[i]=((unsigned long)rand()<<32)|rand();
            bw[i]=((unsigned long)rand()<<32)|rand();
            nw[i]=((unsigned long)rand()<<32)|rand();
            rw[i]=((unsigned long)rand()<<32)|rand();
            pw[i]=((unsigned long)rand()<<32)|rand();            
        }
        side2move=((unsigned long)rand()<<32)|rand();
        for (unsigned int i=0;i<4;i++)
            castling[i]=((unsigned long)rand()<<32)|rand();
        for (unsigned int i=0;i<8;i++)
            files[i]=((unsigned long)rand()<<32)|rand();
    };
    
    void clearHash() {
        for (unsigned int i=0;i<size;i++)
            map[i]=0L;
    }

    unsigned long getHashMask() {
        return hashMask;
    }
    
    unsigned int getSize() {
        return size;
    }
    
    unsigned long getHashKey(Position *pos) {     
        unsigned long fullHash=0L;
        for (int i=0;i<64;i++) {
            if ((*pos).figOnPos((*pos).kings[0],i))       fullHash^=kw[i];
            if ((*pos).figOnPos((*pos).queens[0],i))      fullHash^=qw[i];
            if ((*pos).figOnPos((*pos).bishops[0],i))     fullHash^=bw[i];
            if ((*pos).figOnPos((*pos).knights[0],i))     fullHash^=nw[i];
            if ((*pos).figOnPos((*pos).rooks[0],i))       fullHash^=rw[i];
            if ((*pos).figOnPos((*pos).pawns[0],i))       fullHash^=pw[i];
            if ((*pos).figOnPos((*pos).kings[1],i))       fullHash^=kb[i];
            if ((*pos).figOnPos((*pos).queens[1],i))      fullHash^=qb[i];
            if ((*pos).figOnPos((*pos).bishops[1],i))     fullHash^=bb[i];
            if ((*pos).figOnPos((*pos).knights[1],i))     fullHash^=nb[i];
            if ((*pos).figOnPos((*pos).rooks[1],i))       fullHash^=rb[i];
            if ((*pos).figOnPos((*pos).pawns[1],i))       fullHash^=pb[i];                        
        }

        if (pos->onMove)
            fullHash^=side2move;
        
        short int cast=0;
        cast=(pos->white_castling_long<<3)^(pos->white_castling_short<<2)^(pos->black_castling_long<<1)^(pos->black_castling_short);
        fullHash^=castling[cast];
        
        fullHash^=files[pos->ep%8];
        
        return fullHash;
    };
    
    void insert(unsigned long hashKey, int score, unsigned int currentDepth) {
        score+=100000;
        unsigned int shortenHash=(unsigned int)hashKey&hashMask;
        
        // first: update score if hashKey already stored in map
        unsigned short int replaceAgeIndex=-1;
        unsigned short int storedAge=0;
        for (unsigned short int i=0;i<4;i++) {
            unsigned long storedKey=map[shortenHash+i*2];
            unsigned long storedValue=map[shortenHash+i*2+1];

            if (storedKey==hashKey) {
                if (((storedValue&depthMask)>>54)>=currentDepth)
                    map[shortenHash+i*2+1]=(((unsigned long)score)&(((unsigned long)currentDepth)<<54));
                return;
            }
            
            if ((unsigned short int)((storedValue&ageMask)>>59)>storedAge) // &ageMask nicht notwendig weil age=highest 5 bits
                replaceAgeIndex=i;
        }

        // second: if hashKey not found, set on position where hashAge is oldest
        map[shortenHash+replaceAgeIndex*2]=hashKey;
        map[shortenHash+replaceAgeIndex*2+1]=(((unsigned long)score)&(((unsigned long)currentDepth)<<54));
    }
    
    unsigned int contains(unsigned long hashKey, unsigned int currentDepth) {

        unsigned int shortenHash=(unsigned int)hashKey&hashMask;
        
        for (unsigned short int i=0;i<4;i++) {
            unsigned long storedKey=map[shortenHash+i*2];
            
            unsigned long hashDepth=map[shortenHash+i*2+1]&depthMask;
            hashDepth>>=54;
            
            if (storedKey==hashKey && hashDepth==currentDepth) { // <= oder >= ???
                return i;
            }
        }
        return -1;
    }
    
    int getScoreAt(unsigned long hashKey, int position) {
        
        unsigned int shortenHash=(unsigned int)(hashKey&hashMask);
        
        return (map[shortenHash+position*2+1]&~(ageMask|depthMask))-100000; // besser: &scoreMask, noch nicht implementiert
    }
    
    void updateAges() {
        for (unsigned int i=0;i<size;i+=2) {
            unsigned long age = (map[i]&ageMask)>>59;
            age++;
            age<<=59;
            map[i]=(map[i]&(~ageMask))^age;
        }
    }
    
    virtual ~HashT() {};
    

private:
    unsigned int size;
    unsigned long hashMask;
    unsigned long depthMask;
    unsigned long ageMask;
    
    unsigned long kb[64];
    unsigned long qb[64];
    unsigned long bb[64];
    unsigned long nb[64];
    unsigned long rb[64];
    unsigned long pb[64];
    unsigned long kw[64];
    unsigned long qw[64];
    unsigned long bw[64];
    unsigned long nw[64];
    unsigned long rw[64];
    unsigned long pw[64];         
    unsigned long side2move;
    unsigned long castling[4];
    unsigned long files[8];
    
    unsigned long *map;
};

#endif	/* HASHT_H */