/* 
 * File:   Position.cpp
 * Author: peter
 * 
 * Created on 31. Jänner 2012, 18:23
 */

#include "Position.h"
#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include "Heap.h"
#include "HashT.h"

using namespace std;

typedef boost::ptr_list<string> PositionList;
typedef boost::ptr_list<string>::iterator PositionIterator;

extern HashT *hashTable;

Position::Position() {
    
    ep=0;
    targetmap=0L;
    
    kings[1]=Position::BIN[4];
    queens[1]=Position::BIN[3];
    bishops[1]=Position::BIN[2] | Position::BIN[5];
    knights[1]=Position::BIN[1] | Position::BIN[6];
    rooks[1]=Position::BIN[0] | Position::BIN[7];
    pawns[1]=Position::BIN[8] | Position::BIN[9] | Position::BIN[10] | Position::BIN[11] | Position::BIN[12] | Position::BIN[13] | Position::BIN[14] | Position::BIN[15];

    kings[0]=Position::BIN[60];
    queens[0]=Position::BIN[59];
    bishops[0]=Position::BIN[58] | Position::BIN[61];
    knights[0]=Position::BIN[57] | Position::BIN[62];
    rooks[0]=Position::BIN[56] | Position::BIN[63];
    pawns[0]=Position::BIN[48] | Position::BIN[49] | Position::BIN[50] | Position::BIN[51] | Position::BIN[52] | Position::BIN[53] | Position::BIN[54] | Position::BIN[55];

    updateBitboardLongs();
    
    white_castling_short=true;
    white_castling_long=true;
    black_castling_short=true;
    black_castling_long=true;      
    
    onMove=false;
    
}

Position::Position(const Position &pos) {
    this->ep=0;
    this->targetmap=0L;
    this->aw=pos.aw;
    this->ab=pos.ab;
    this->all=pos.all;
    this->kings[0]=pos.kings[0];
    this->kings[1]=pos.kings[1];
    this->pawns[0]=pos.pawns[0];
    this->pawns[1]=pos.pawns[1];
    this->bishops[0]=pos.bishops[0];
    this->bishops[1]=pos.bishops[1];
    this->knights[0]=pos.knights[0];
    this->knights[1]=pos.knights[1];
    this->queens[0]=pos.queens[0];
    this->queens[1]=pos.queens[1];
    this->rooks[0]=pos.rooks[0];
    this->rooks[1]=pos.rooks[1];
    this->white_castling_short=pos.white_castling_short;
    this->white_castling_long=pos.white_castling_long;
    this->black_castling_short=pos.black_castling_short;
    this->black_castling_long=pos.black_castling_long;       
    this->onMove=pos.onMove;
}

Position::Position(const string &fen) {
    
    white_castling_short=false;
    white_castling_long=false;
    black_castling_short=false;
    black_castling_long=false;
    
    ep=0;
    targetmap=0L;

    aw=0L;
    ab=0L;
    all=0L;
    
    pawns[0]=0L;
    pawns[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    bishops[0]=0L;
    bishops[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    queens[0]=0L;
    queens[1]=0L;
    knights[0]=0L;
    knights[1]=0L;
    kings[0]=0L;
    kings[1]=0L;
    
    size_t found;
    found=fen.find(" ");
    string fengrp0 = fen.substr(0,int(found));
    string rest = fen.substr(int(found)+1,fen.size()-int(found));
    string fengrp1 = rest.substr(0,1);
    rest = rest.substr(2,rest.size()-2);
    found=rest.find(" ");
    string fengrp2 = rest.substr(0,int(found));
    rest = rest.substr(int(found)+1,rest.size()-int(found));
    string fengrp3 = rest;

    // 1. group: figures
    int pos=0;
    for (unsigned int i=0;i<fengrp0.length();i++) {
        if (fengrp0.at(i)=='/') continue;
        else if (fengrp0.at(i)=='1') pos=pos+1;
        else if (fengrp0.at(i)=='2') pos=pos+2;
        else if (fengrp0.at(i)=='3') pos=pos+3;
        else if (fengrp0.at(i)=='4') pos=pos+4;
        else if (fengrp0.at(i)=='5') pos=pos+5;
        else if (fengrp0.at(i)=='6') pos=pos+6;
        else if (fengrp0.at(i)=='7') pos=pos+7;
        else if (fengrp0.at(i)=='8') pos=pos+8;
        else if (fengrp0.at(i)=='k') { kings[1]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='q') { queens[1]=queens[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='n') { knights[1]=knights[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='b') { bishops[1]=bishops[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='r') { rooks[1]=rooks[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='p') { pawns[1]=pawns[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='K') { kings[0]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='Q') { queens[0]=queens[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='N') { knights[0]=knights[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='B') { bishops[0]=bishops[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='R') { rooks[0]=rooks[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='P') { pawns[0]=pawns[0]|Position::BIN[pos]; pos++;}
    }
    
    // 2. group: color on move, fengrp1
    if (fengrp1=="b")
        onMove=true;
    else
        onMove=false;

    // 3. group: castles
    if (fengrp2.at(0)!='-') {
        for (unsigned int i=0;i<fengrp2.size();i++) {
            if (fengrp2.at(i)=='K')   white_castling_short=true;
            if (fengrp2.at(i)=='Q')   white_castling_long=true;
            if (fengrp2.at(i)=='k')   black_castling_short=true;
            if (fengrp2.at(i)=='q')   black_castling_long=true;
        }
    }
    
    // 4. group: en passant
    int tmp_ep=0;
    if (fengrp3.length()==2) {
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='6') tmp_ep=1;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='6') tmp_ep=2;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='6') tmp_ep=3;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='6') tmp_ep=4;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='6') tmp_ep=5;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='6') tmp_ep=6;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='6') tmp_ep=7;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='6') tmp_ep=8;
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='3') tmp_ep=9;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='3') tmp_ep=10;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='3') tmp_ep=11;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='3') tmp_ep=12;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='3') tmp_ep=13;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='3') tmp_ep=14;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='3') tmp_ep=15;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='3') tmp_ep=16;
    }
    ep=tmp_ep;
    
    updateBitboardLongs();
}

// return List of Child-Positions (with lastmove-string), used by Eval.cpp commands
PositionList Position::getMMChilds() {
    PositionList returnList;
    calc_moves(true);
    for (int i=0;i<mvSize[cd];i++) {
        Position *tmp_pos = new Position(*this);
        tmp_pos->move_figure(mvD[cd][i][0],mvD[cd][i][1],mvI[cd][i]);
        tmp_pos->last_move=getMoveString(i);
        string *move=new string(tmp_pos->last_move);
        returnList.push_back(move);
        // FIXME: check for legal-moves fehlt
    }
    return returnList;
}

// get divide sums
void Position::getDivide(int color, int curr_depth, const int &depth, const bool &modus, long &sum, long &total) {
    cd = curr_depth-1;
    calc_moves(true);  
    doBackup(cd);

    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            move_figure(mvD[cd][i][0],mvD[cd][i][1],mvI[cd][i]);
            if (curr_depth==1) {
                sum=0;
                last_move=getMoveString(i);
            }
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            if (!catchKing) {
                getDivide(abs(color-1), (curr_depth+1),depth, modus, sum, total);
                cd--;
                if (curr_depth==1)
                    cout << last_move << " " << sum << "     " << getFen() << endl;                                
            } else {
                if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
                    sum--;
                    total--;
                }                
            }
            
            undoMove(cd);
        }
    }
    if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
        sum+=mvSize[cd];
        total+=mvSize[cd];
    }
}

// do a full information perft on a given position, used by "perft/perft all/perft list" commands
void Position::getPerftInfo(int curr_depth, int depth, Perftres *perft) {
    
    cd = curr_depth-1;
    doBackup(cd);
    calc_moves(true);
    
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],mvI[cd][i]);
            
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            
            if (!catchKing) {
                getPerftInfo((curr_depth+1),depth, perft);
                cd--;

                // count captures
                if (ret)
                    perft->perftres[1+(curr_depth-1)*7]++;                    

                // count ep-moves
                // if (mvD[cd][i][2]!=0L) { // <-- funktioniert nicht mehr, kill / [2] nicht mehr vorhanden
                if (mvD[cd][i][1]!=0L) {                
                    perft->perftres[4+(curr_depth-1)*7]++;
                    perft->perftres[1+(curr_depth-1)*7]++;
                } 

                // count promotions
                if (mvI[cd][i]>0)
                    perft->perftres[5+(curr_depth-1)*7]++;                                
                
                // count castlings
                if ((mvD[cd][i][1]&kings[!onMove])!=0L && (mvD[cd][i][0]<<2==mvD[cd][i][1] || mvD[cd][i][0]>>2==mvD[cd][i][1]))
                    perft->perftres[6+(curr_depth-1)*7]++;                     
            
            } else
                perft->perftres[(curr_depth-1)*7]--;                
            undoMove(cd);
            
        }
    }    

    perft->perftres[(curr_depth-1)*7]+=mvSize[cd];
    
    if (perft!=NULL) {
        /*        


        // count checks
        Position tmp_pos3(*tmp_pos);
        if (opponent_target_bitboard(tmp_pos3,tmp_pos->kings[abs(color-1)],abs(color-1)))
            perft->perftres[2+(curr_depth-1)*7]++;                

        */
    }    
}

// save nodes-sum of a given ply in variable sum, used by Perft.cpp, "perft x" command
void Position::getChildsSpeed(int color, int curr_depth, int depth, long &sum) {
    cd = curr_depth-1;
    doBackup(cd);
    calc_moves(true);
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            move_figure(mvD[cd][i][0],mvD[cd][i][1],mvI[cd][i]);
            if (!opponent_target_bitboard2(kings[color])) {
                getChildsSpeed(abs(color-1),(curr_depth+1),depth, sum);
                cd--;
            } else {
                if (curr_depth==depth)
                    sum--;
            }
            undoMove(cd);
        }
    }
    if (curr_depth==depth)
        sum+=mvSize[cd];
}

void Position::sortMoves(unsigned short int sortedIndizes[], int cd) {
    unsigned short int cntMoves=mvSize[cd];
    short int moveValues[cntMoves];

    // calculate a value for every move
    for (int i=0;i<cntMoves;i++) {
        
        if ((mvD[cd][i][1]&all)==0L) {
            moveValues[i]=-2000;
            continue;
        }
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else /* if ((pawns[onMove]&mvD[cd][i][0])!=0L) */    fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else /* if ((pawns[notOnMove]&mvD[cd][i][1])!=0L) */     toValue=1;        

        moveValues[i]=toValue-fromValue;
        
    }
    
    for (int i=0;i<cntMoves;i++)
        sortedIndizes[i]=i;
    
    
    Heap::buildHeap(sortedIndizes,moveValues,cntMoves);
    
    /*
    long cntMoves=mvSize[cd];
    
    unsigned long mvD_new[cntMoves][2];
    short int mvI_new[cntMoves];    
    
    // sehr langsam: (nur schlag-züge bevorzugen)
    
    long cnt=0;
    for (long i=0;i<cntMoves;i++) {
        if ((mvD[cd][i][1]&all)!=0L) {
            mvD_new[cnt][0]=mvD[cd][i][0];
            mvD_new[cnt][1]=mvD[cd][i][1];
            mvI_new[cnt]=mvI[cd][i];
            cnt++;
        }
    }
    
    for (long i=0;i<cntMoves;i++) {
        if ((mvD[cd][i][1]&all)==0L) {
            mvD_new[cnt][0]=mvD[cd][i][0];
            mvD_new[cnt][1]=mvD[cd][i][1];
            mvI_new[cnt]=mvI[cd][i];
            cnt++;
        }
    }    
    
    for (long i=0;i<cntMoves;i++) {
        mvD[cd][i][0]=mvD_new[i][0];
        mvD[cd][i][1]=mvD_new[i][1];
        mvI[cd][i]=mvI_new[i];
    }
    
    
    // unsigned long mvD[15][356][2];
    // short int mvI[15][356];    
    
    */
    
}

int Position::alphaBetaMax(bool side2move, int alpha, int beta, int curr_depth, int depth, int &sum, string &move) {
    sum++;
    cd = curr_depth-1;
    if (curr_depth==depth)
        return evaluate(side2move);
    calc_moves(true);
    unsigned short int sortedIndizes[256];
    sortMoves(sortedIndizes,cd);
    doBackup(cd);
    
    int maxWert=alpha;
    
    // for (int i=0;i<mvSize[cd];i++) {
    for (int i=mvSize[cd]-1;i>=0;i--) {
        // move_figure(mvD[cd][i][0],mvD[cd][i][1],mvI[cd][i]);
        move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],mvI[cd][sortedIndizes[i]]);
        
        // legality-check
        if (curr_depth==1) {
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            if (catchKing) {
                undoMove(cd);
                continue;
            }
        }
        
        
        
        
        
        /*
        for (int i=0;i<26;i++) {
            int singleBit = (fullHash>>i)&1;
            cout << singleBit;
        }
        cout << endl;
        */

        // int hashScore=hashTable->getScoreAt(fullHash&67108863L,fullHash);

        int score;
        unsigned long hashKey = hashTable->getHashKey(this);
        

        int position=hashTable->contains(hashKey,cd);
        
        // if (cd%2==0) {
        
            if (position>-1) { // map contains hashKey
                score=hashTable->getScoreAt(hashKey, position);
            } else {
                score = -alphaBetaMax(!side2move,-beta, -maxWert, (curr_depth+1), depth, sum, move);
                hashTable->insert(hashKey,score,cd);
                cd--;
            }
/*
        }
        else {
            score = -alphaBetaMax(!side2move,-beta, -maxWert, (curr_depth+1), depth, sum, move);
            cd--;        
        }
*/
       
        undoMove(cd);

        if (score > maxWert) {
            maxWert = score;
            if (maxWert>=beta)
                break;
            if (curr_depth==1)
                move=getMoveString(sortedIndizes[i]);            
        }
    }
    return maxWert;
}

int Position::evaluate(bool side2move) {
    
    int materialScore0, materialScore1;
    
    materialScore0  =   10 *   popCount(pawns[0]) +
                        275 *   popCount(knights[0]) +
                        325 *   popCount(bishops[0]) +
                        465 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        2000 *  popCount(kings[0]);
    
    materialScore1  =   10 *   popCount(pawns[1]) +
                        275 *   popCount(knights[1]) +
                        325 *   popCount(bishops[1]) +
                        465 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        2000 *  popCount(kings[1]);    
    
    int moveScore0, moveScore1;
    int moveWeight=200;
    
    
    onMove=false;
    targetmap=0L;
    calc_moves(false); // true=with castling
    moveScore0      =   popCount(targetmap);

    int fightScore0 = popCount(targetmap&all);
    

    
    onMove=true;
    targetmap=0L;
    calc_moves(false); // true=with castling
    moveScore1      =   popCount(targetmap);
    
    int fightScore1 = popCount(targetmap&all);
    

    
    int fightWeight = 100;
    

    
    int score = (materialScore0-materialScore1)+(moveScore0-moveScore1)*moveWeight+(fightScore0-fightScore1)*fightWeight;
    
    
    if (side2move)
        score*=-1;
    
    
    return score;
}

void Position::doBackup(int depth) {
    myBackup[depth].aw=aw;
    myBackup[depth].ab=ab;
    myBackup[depth].all=all;
    myBackup[depth].kings[0]=kings[0];
    myBackup[depth].kings[1]=kings[1];
    myBackup[depth].pawns[0]=pawns[0];
    myBackup[depth].pawns[1]=pawns[1];
    myBackup[depth].bishops[0]=bishops[0];
    myBackup[depth].bishops[1]=bishops[1];
    myBackup[depth].knights[0]=knights[0];
    myBackup[depth].knights[1]=knights[1];
    myBackup[depth].queens[0]=queens[0];
    myBackup[depth].queens[1]=queens[1];
    myBackup[depth].rooks[0]=rooks[0];
    myBackup[depth].rooks[1]=rooks[1];
    myBackup[depth].targetmap=targetmap;
    myBackup[depth].ep=ep;
    myBackup[depth].white_castling_short=white_castling_short;
    myBackup[depth].white_castling_long=white_castling_long;
    myBackup[depth].black_castling_short=black_castling_short;
    myBackup[depth].black_castling_long=black_castling_long;
    myBackup[depth].onMove=onMove;
}

void Position::undoMove(int depth) {
    aw=myBackup[depth].aw;
    ab=myBackup[depth].ab;
    all=myBackup[depth].all;
    kings[0]=myBackup[depth].kings[0];
    kings[1]=myBackup[depth].kings[1];
    pawns[0]=myBackup[depth].pawns[0];
    pawns[1]=myBackup[depth].pawns[1];
    bishops[0]=myBackup[depth].bishops[0];
    bishops[1]=myBackup[depth].bishops[1];
    knights[0]=myBackup[depth].knights[0];
    knights[1]=myBackup[depth].knights[1];
    queens[0]=myBackup[depth].queens[0];
    queens[1]=myBackup[depth].queens[1];
    rooks[0]=myBackup[depth].rooks[0];
    rooks[1]=myBackup[depth].rooks[1];
    targetmap=myBackup[depth].targetmap;
    ep=myBackup[depth].ep;
    white_castling_short=myBackup[depth].white_castling_short;
    white_castling_long=myBackup[depth].white_castling_long;
    black_castling_short=myBackup[depth].black_castling_short;
    black_castling_long=myBackup[depth].black_castling_long;
    onMove=myBackup[depth].onMove;
}

bool Position::move_figure (const unsigned long &from, const unsigned long &to, const int &promotion) {
    bool capture=false;

    ep=0;
    
    // do castling
    // update king- and rook-position
    if ((from&kings[onMove])!=0L && (from&(BIN[4]|BIN[60]))!=0L && (to&(BIN[2]|BIN[6]|BIN[58]|BIN[62]))!=0L) {    
        
        if (to==BIN[2]) rochade(from,to,BIN[0],BIN[3]);
        if (to==BIN[6]) rochade(from,to,BIN[7],BIN[5]);
        if (to==BIN[58]) rochade(from,to,BIN[56],BIN[59]);
        if (to==BIN[62]) rochade(from,to,BIN[63],BIN[61]);

        // remove rochade possibilities
        if (to==BIN[58] || to==BIN[62]) {
            white_castling_short=false;
            white_castling_long=false;
        } else if (to==BIN[2] || to==BIN[6]) {
            black_castling_short=false;
            black_castling_long=false;
        }

        updateBitboardLongs();
        onMove=!onMove;
        return capture;
    }

    
    bool notOnMove=!onMove;
        
    // kill target figure (if there's one)
    if ((kings[notOnMove]&to)!=0L) kings[notOnMove]=kings[notOnMove] ^ to;
    else if ((queens[notOnMove]&to)!=0L) queens[notOnMove]=queens[notOnMove] ^ to;
    else if ((bishops[notOnMove]&to)!=0L) bishops[notOnMove]=bishops[notOnMove] ^ to;
    else if ((knights[notOnMove]&to)!=0L) knights[notOnMove]=knights[notOnMove] ^ to;
    else if ((rooks[notOnMove]&to)!=0L) {
        rooks[notOnMove]=rooks[notOnMove] ^ to;
        if (to==72057594037927936L) white_castling_long=false;
        else if (to==9223372036854775808UL) white_castling_short=false;
        else if (to==1L) black_castling_long=false;
        else if (to==128L) black_castling_short=false;        
    }
    else if ((pawns[notOnMove]&to)!=0L) pawns[notOnMove]=pawns[notOnMove] ^ to;
    
    if ((to & all)!=0L)
        capture=true;
    
    // ep-move
    if ((pawns[onMove]&from)!=0L && (all&to)==0L) {
        if (to==from>>9) pawns[notOnMove]=pawns[notOnMove] ^ (from>>1);
        if (to==from>>7) pawns[notOnMove]=pawns[notOnMove] ^ (from<<1);
        if (to==from<<9) pawns[notOnMove]=pawns[notOnMove] ^ (from<<1);
        if (to==from<<7) pawns[notOnMove]=pawns[notOnMove] ^ (from>>1);
    }
    
    // move figure
    if ((pawns[onMove]&from)!=0L) {
        pawns[onMove]^=(from|to);
        if (from==Position::BIN[48] && to==Position::BIN[32])   ep=9;
        else if (from==Position::BIN[49] && to==Position::BIN[33])   ep=10;
        else if (from==Position::BIN[50] && to==Position::BIN[34])   ep=11;
        else if (from==Position::BIN[51] && to==Position::BIN[35])   ep=12;
        else if (from==Position::BIN[52] && to==Position::BIN[36])   ep=13;
        else if (from==Position::BIN[53] && to==Position::BIN[37])   ep=14;
        else if (from==Position::BIN[54] && to==Position::BIN[38])   ep=15;
        else if (from==Position::BIN[55] && to==Position::BIN[39])   ep=16;
        
        else if (from==Position::BIN[8] && to==Position::BIN[24])    ep=1;
        else if (from==Position::BIN[9] && to==Position::BIN[25])    ep=2;
        else if (from==Position::BIN[10] && to==Position::BIN[26])   ep=3;
        else if (from==Position::BIN[11] && to==Position::BIN[27])   ep=4;
        else if (from==Position::BIN[12] && to==Position::BIN[28])   ep=5;
        else if (from==Position::BIN[13] && to==Position::BIN[29])   ep=6;
        else if (from==Position::BIN[14] && to==Position::BIN[30])   ep=7;
        else if (from==Position::BIN[15] && to==Position::BIN[31])   ep=8;        
            
    } else if ((queens[onMove]&from)!=0L) {
        queens[onMove]^=(from|to);
    } else if ((bishops[onMove]&from)!=0L) {
        bishops[onMove]^=(from|to);
    } else if ((knights[onMove]&from)!=0L) {
        knights[onMove]^=(from|to);
    } else if ((kings[onMove]&from)!=0L) {
        kings[onMove]^=(from|to);
        if (from==1152921504606846976L) { // white kings startposition
            white_castling_short=false;
            white_castling_long=false;
        }
        else if (from==16L) { // black kings startposition
            black_castling_short=false;
            black_castling_long=false;
        }
    } else if ((rooks[onMove]&from)!=0L) {
        rooks[onMove]^=(from|to);
        if (from==72057594037927936L) { // long white rook startposition
            white_castling_long=false;
        } else if (from==9223372036854775808UL) { // short white rook startposition
            white_castling_short=false;
        } else if (from==1L) { // long black rook startposition
            black_castling_long=false;
        } else if (from==128L) { // short black rook startposition
            black_castling_short=false;
        }                    
    }
    
    // promote figure
    if (promotion>0) {
        int color2=0;
        if (onMove) color2=1;
        promote(to, promotion, color2);
    }

    updateBitboardLongs();

    onMove=!onMove;
    return capture;
}

int Position::popCount(unsigned long x) { // too slow
   int count = 0;
   for (int i = 0; i < 64; i++, x >>= 1)
      count += (int)x & 1;
   return count;
}

void Position::printBoard() {
    
    cout    << "  a b c d e f g h" << endl
            << "8" << printBoardRow(0) << endl
            << "7" << printBoardRow(1)<< endl
            << "6" << printBoardRow(2)<< endl
            << "5" << printBoardRow(3)<< endl
            << "4" << printBoardRow(4)<< endl
            << "3" << printBoardRow(5)<< endl
            << "2" << printBoardRow(6)<< endl
            << "1" << printBoardRow(7)<< endl;            
}

string Position::printBoardRow(int row) {
    string ret;
    for (int i=0;i<8;i++) {
        int position=row*8+i;
        
        if (figOnPos(kings[0],position))
            ret.append(" K");
        else if (figOnPos(kings[1],position))
            ret.append(" k");
        else if (figOnPos(queens[0],position))
            ret.append(" Q");
        else if (figOnPos(queens[1],position))
            ret.append(" q");
        else if (figOnPos(knights[0],position))
            ret.append(" N");
        else if (figOnPos(knights[1],position))
            ret.append(" n");
        else if (figOnPos(bishops[0],position))
            ret.append(" B");
        else if (figOnPos(bishops[1],position))
            ret.append(" b");        
        else if (figOnPos(pawns[0],position))
            ret.append(" P");
        else if (figOnPos(pawns[1],position))
            ret.append(" p");        
        else if (figOnPos(rooks[0],position))
            ret.append(" R");        
        else if (figOnPos(rooks[1],position))
            ret.append(" r");                
        else
            ret.append(" .");
    }
    return ret;
}

bool Position::figOnPos(long figure, int position) {
    return ((figure & Position::BIN[position]) != 0L);
}

void Position::getInfo() {
    
    cout << getFen() << endl << endl;
    
    cout << "en passant: " << ep << endl << endl;
    
    
    cout << "white castling short: " << getBoolString(white_castling_short) << endl;
    cout << "white castling long:  " << getBoolString(white_castling_long) << endl;
    cout << "black castling short: " << getBoolString(black_castling_short) << endl;
    cout << "black castling long:  " << getBoolString(black_castling_long) << endl;
    
    cout << endl;
    
    cout << "on Move: ";
    if (onMove)
        cout << "black";
    else
        cout << "white";
    
    cout << endl;
}

string Position::getBoolString(bool value) {
    if (value)
        return "true";
    else
        return "false";
}

inline void Position::promote(const unsigned long &pos, const int &promotype, const int &color) {
    // remove pawn
    if ((pawns[0] & pos) != 0L) pawns[0]=pawns[0] ^ pos;
    if ((pawns[1] & pos) != 0L) pawns[1]=pawns[1] ^ pos;

    // add promoted figure
    if      (promotype==1 && color==0) queens[0]    = queens[0]     | pos;
    else if (promotype==2 && color==0) rooks[0]     = rooks[0]      | pos;
    else if (promotype==3 && color==0) bishops[0]   = bishops[0]    | pos;
    else if (promotype==4 && color==0) knights[0]   = knights[0]    | pos;
    else if (promotype==1 && color==1) queens[1]    = queens[1]     | pos;
    else if (promotype==2 && color==1) rooks[1]     = rooks[1]      | pos;
    else if (promotype==3 && color==1) bishops[1]   = bishops[1]    | pos;
    else if (promotype==4 && color==1) knights[1]   = knights[1]    | pos;
}

inline void Position::rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos) {
    // move king
    if ((kings[0] & old_king_pos)!=0L) {
        kings[0]=kings[0] ^ old_king_pos;
        kings[0]=kings[0] | new_king_pos;
    } else if ((kings[1] & old_king_pos)!=0L) {
        kings[1]=kings[1] ^ old_king_pos;
        kings[1]=kings[1] | new_king_pos;
    }

    // move rook
    if ((rooks[0] & old_rook_pos)!=0L) {
        rooks[0]=rooks[0] ^ old_rook_pos;
        rooks[0]=rooks[0] | new_rook_pos;
    } else if ((rooks[1] & old_rook_pos)!=0L) {
        rooks[1]=rooks[1] ^ old_rook_pos;
        rooks[1]=rooks[1] | new_rook_pos;
    }
}

inline void Position::updateBitboardLongs() {
    aw=kings[0] | queens[0] | rooks[0] | bishops[0] | knights[0] | pawns[0];
    ab=kings[1] | queens[1] | rooks[1] | bishops[1] | knights[1] | pawns[1];
    all=ab|aw;    
}

inline int Position::bitScanForward(unsigned long &bb) {
   const unsigned long debruijn64 = 0x07EDD5E59A4E28C2;
   return Position::index64[((bb & -bb) * debruijn64) >> 58];
}

void Position::calc_moves(const bool &check) {
    
    int color2;
    if (onMove)
        color2=1;
    else
        color2=0;
    
    
    mvSize[cd]=0;
    
    unsigned long bitboard;    
    unsigned long myFigures;
    unsigned long opFigures;
    
    if (onMove) {
        // all figures bitboard from moving / opponent color
        myFigures=ab;
        opFigures=aw;
        // pawns
        bitboard = pawns[color2];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_black(x);
            bitboard-=Position::BIN[x];
        }        
    } else {
        // all figures bitboard from moving / opponent color
        myFigures=aw;
        opFigures=ab;        
        // pawns
        bitboard = pawns[color2];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_white(x);
            bitboard-=Position::BIN[x];
        }
    }
    
    // rooks
    bitboard = rooks[color2];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // bishops
    bitboard = bishops[color2];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // queens
    bitboard = queens[color2];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);                        
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }         
    
    // knights
    bitboard = knights[color2];
    while(bitboard != 0L) {
        int x=bitScanForward(bitboard);
        split_moves_direct(x, Position::KN_TARGET3x, Position::KN_TARGET3x_FROM[x], Position::KN_TARGET3x_TO[x], myFigures);
        bitboard-=Position::BIN[x];
    }       
    
    // kings
    int x = bitScanForward(kings[color2]);
    split_moves_direct(x, Position::K_TARGET2x, Position::K_TARGET2x_FROM[x], Position::K_TARGET2x_TO[x], myFigures);
    if (check)
        addKingMoves(color2);

    // update targetmap
    for (int i=0;i<mvSize[cd];i++)
        targetmap=targetmap|mvD[cd][i][1];
}

inline void Position::split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & myFigures) == 0L) { // don't fight my color figures
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvI[cd][mvSize[cd]]=0;
            mvSize[cd]++;
        }
    }
}

inline void Position::split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & all) == 0L) {
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvI[cd][mvSize[cd]]=0;
            mvSize[cd]++;                
        } else {
            if ((target[i] & op) != 0L) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[z];
                mvD[cd][mvSize[cd]][1]=target[i];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                    
            }
            break;
        }
    }
}

inline void Position::addPawnSingles_black(const int &z) {
    unsigned long startpos_black=65280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (Position::TARGET_BLACK_V1[z] & all) == 0UL) {// one step forward: just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
    }
    if (    (Position::TARGET_BLACK_V3[z] & aw) != 0UL ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;        
    }
    if (    (Position::TARGET_BLACK_V4[z] & aw) != 0L ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
    }

    // search for promotions and add moves
    int size=mvSize[cd];
    for (int i=0;i<size;i++) {
        if (mvD[cd][i][0]==Position::BIN[z] && (mvD[cd][i][1] & 18374686479671623680UL) != 0L) {
            mvI[cd][i]=1;
            long froml=mvD[cd][i][0];
            long tol=mvD[cd][i][1];
            
                mvD[cd][mvSize[cd]][0]=froml;
                mvD[cd][mvSize[cd]][1]=tol;
                mvI[cd][mvSize[cd]]=2;
                mvSize[cd]++;
                
                mvD[cd][mvSize[cd]][0]=froml;
                mvD[cd][mvSize[cd]][1]=tol;
                mvI[cd][mvSize[cd]]=3;
                mvSize[cd]++;
                
                mvD[cd][mvSize[cd]][0]=froml;
                mvD[cd][mvSize[cd]][1]=tol;
                mvI[cd][mvSize[cd]]=4;
                mvSize[cd]++;
        }
    }
      
    if (    (bitboardpos & startpos_black) != 0L && (Position::TARGET_BLACK_V2[z] & all) == 0L && (Position::TARGET_BLACK_V1[z] & all) == 0L) { // two steps forward: both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V2[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;        
        
    }    
    
    // en passant moves
    if (ep>8 && ep<=16) { // possible ep for black
        if (ep==9 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[40];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
        }
        else if (ep==10 && bitboardpos==Position::BIN[32]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[32];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==10 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==15 && bitboardpos==Position::BIN[39]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[39];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==16 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[47];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
        }
    }    
    
}

inline void Position::addPawnSingles_white(const int &z) {
    unsigned long startpos_white=71776119061217280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (TARGET_WHITE_V1[z] & all) == 0L) { // just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V1[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
    }
    if (    (TARGET_WHITE_V3[z] & ab) != 0L ) { // just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V3[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
    }
    if (    (TARGET_WHITE_V4[z] & ab) != 0L ) {// just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V4[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
    }

    // search for promotions and add moves
    int size=mvSize[cd];
    for (int i=0;i<size;i++) {
        if (mvD[cd][i][0]==Position::BIN[z] && (mvD[cd][i][1] & 255UL) != 0L) {
            mvI[cd][i]=1;
            long froml=mvD[cd][i][0];
            long tol=mvD[cd][i][1];
            
                mvD[cd][mvSize[cd]][0]=froml;
                mvD[cd][mvSize[cd]][1]=tol;
                mvI[cd][mvSize[cd]]=2;
                mvSize[cd]++;
                
                mvD[cd][mvSize[cd]][0]=froml;
                mvD[cd][mvSize[cd]][1]=tol;
                mvI[cd][mvSize[cd]]=3;
                mvSize[cd]++;
                
                mvD[cd][mvSize[cd]][0]=froml;
                mvD[cd][mvSize[cd]][1]=tol;
                mvI[cd][mvSize[cd]]=4;
                mvSize[cd]++;
        }
    }    

    if (    (bitboardpos & startpos_white) != 0L && (TARGET_WHITE_V2[z] & all) == 0L && (TARGET_WHITE_V1[z] & all) == 0L) { // both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V2[z];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;
    }    
    
    // en passant moves
    if (ep>0 && ep<=8) { // possible ep for white
        if (ep==1 && bitboardpos==Position::BIN[25]  ) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[16];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;            
        }
        else if (ep==2 && bitboardpos==Position::BIN[24]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[24];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==2 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[25]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[31]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[31];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
        else if (ep==8 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[23];
                mvI[cd][mvSize[cd]]=0;
                mvSize[cd]++;                        
        }
    }    
    
}

inline void Position::addKingMoves(const int &color) {
   
    // white short, K
    if (color==0
            && white_castling_short // castling allowed
            && (all & 6917529027641081856L) == 0L // fields are free between king and rook
            && opponent_target_bitboard2(8070450532247928832L)==false // check if opponent targets an involved field
            && (pawns[1] & 69805794224242688L) == 0L) { // check if there's a pawn promotion move targeting an involved field

            mvD[cd][mvSize[cd]][0]=kings[color];
            mvD[cd][mvSize[cd]][1]=4611686018427387904L;
            mvI[cd][mvSize[cd]]=0;
            mvSize[cd]++;
    }

    // white long, Q
    if (color==0 && white_castling_long 
            && (all & 1008806316530991104L) == 0L 
            && opponent_target_bitboard2(2017612633061982208L)==false 
            && (pawns[1] & 17451448556060672L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[color];
            mvD[cd][mvSize[cd]][1]=288230376151711744L;
            mvI[cd][mvSize[cd]]=0;
            mvSize[cd]++;
    }

    // black short, k
    if (color==1 && black_castling_short 
            && (all & 96L) == 0L 
            && opponent_target_bitboard2(112L)==false 
            && (pawns[0] & 63488L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[color];
            mvD[cd][mvSize[cd]][1]=64L;
            mvI[cd][mvSize[cd]]=0;
            mvSize[cd]++;
    }

    // black long, q
    if (color==1 && black_castling_long 
            && (all & 14L) == 0L 
            && opponent_target_bitboard2(28L)==false 
            && (pawns[0] & 15872L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[color];
            mvD[cd][mvSize[cd]][1]=4L;
            mvI[cd][mvSize[cd]]=0;
            mvSize[cd]++;            
    }
}

inline bool Position::opponent_target_bitboard2(const unsigned long &bitboard) {
    bool value=false;
    cd++;
    doBackup(cd);
    targetmap=0L;
    onMove=!onMove;
    calc_moves(false);
    onMove=!onMove;
    if ((targetmap & bitboard) != 0L) value=true;
    undoMove(cd);
    cd--;
    return value;
}

string Position::getFen() {
    
    string fen;
    
    // 1. group: figures
    int spaces=0;
    for (int i=1;i<=64;i++) {
        if ((Position::BIN[i-1]&all)!=0L) {
            if (spaces>0) {
                fen+=Position::intToString(spaces);
                spaces=0;
            }
            if ((kings[0]&Position::BIN[i-1])!=0L) fen+="K";
            else if ((queens[0]&Position::BIN[i-1])!=0L) fen+="Q";
            else if ((bishops[0]&Position::BIN[i-1])!=0L) fen+="B";
            else if ((knights[0]&Position::BIN[i-1])!=0L) fen+="N";
            else if ((rooks[0]&Position::BIN[i-1])!=0L) fen+="R";
            else if ((pawns[0]&Position::BIN[i-1])!=0L) fen+="P";
            else if ((kings[1]&Position::BIN[i-1])!=0L) fen+="k";
            else if ((queens[1]&Position::BIN[i-1])!=0L) fen+="q";
            else if ((bishops[1]&Position::BIN[i-1])!=0L) fen+="b";
            else if ((knights[1]&Position::BIN[i-1])!=0L) fen+="n";
            else if ((rooks[1]&Position::BIN[i-1])!=0L) fen+="r";
            else if ((pawns[1]&Position::BIN[i-1])!=0L) fen+="p";
        } else
            spaces++;

        if (i%8==0 && i<64) {
            if (spaces>0) {
                fen+=Position::intToString(spaces);
                spaces=0;
            }
            fen+="/";
        }
        if (i==64 && spaces>0) {
                fen+=Position::intToString(spaces);
        }
            
    }

    // 2. group: color on move
    if (onMove)
        fen+=" b";
    else
        fen+=" w";

    // 3. group: castles
    if (white_castling_short==false && white_castling_long==false && black_castling_short==false && black_castling_long==false)
        fen+=" -";
    else {
        fen+=" ";
        if (white_castling_short)   fen+="K";
        if (white_castling_long)    fen+="Q";
        if (black_castling_short)   fen+="k";
        if (black_castling_long)    fen+="q";
    }

    // 4. group: en passant
    if (ep==0)  fen+=" -";
    else if(ep==1)  fen+=" a6";
    else if (ep==2) fen+=" b6";
    else if (ep==3) fen+=" c6";
    else if (ep==4) fen+=" d6";
    else if (ep==5) fen+=" e6";
    else if (ep==6) fen+=" f6";
    else if (ep==7) fen+=" g6";
    else if (ep==8) fen+=" h6";
    else if (ep==9)     fen+=" a3";
    else if (ep==10)    fen+=" b3";
    else if (ep==11)    fen+=" c3";
    else if (ep==12)    fen+=" d3";
    else if (ep==13)    fen+=" e3";
    else if (ep==14)    fen+=" f3";
    else if (ep==15)    fen+=" g3";
    else if (ep==16)    fen+=" h3";

    return fen;
}

string Position::getMoveString(int i) {
    string val;
    
    val+=get_hum(bitScanForward(mvD[cd][i][0]));
    val+=get_hum(bitScanForward(mvD[cd][i][1]));
    
    if (mvI[cd][i]==1)
        val+="q";
    else if (mvI[cd][i]==2)
        val+="r";
    else if (mvI[cd][i]==3)
        val+="b";
    else if (mvI[cd][i]==4)
        val+="n";
    
    return val;
}

string Position::get_hum(int k) {
    string str; // = new String();

    if (k%8==0) str+="a";
    if (k%8==1) str+="b";
    if (k%8==2) str+="c";
    if (k%8==3) str+="d";
    if (k%8==4) str+="e";
    if (k%8==5) str+="f";
    if (k%8==6) str+="g";
    if (k%8==7) str+="h";
    str+=Position::intToString(8-k/8);
    return str;
}