/* 
 * File:   MoveList.h
 * Author: peter
 *
 * Created on 16. Dezember 2012, 20:06
 */

#ifndef MOVELIST_H
#define	MOVELIST_H

using namespace std;

class MoveList {
public:
    
    /*****************************************************
     * Initialisiere eine GameHistory. Dabei wird davon  *
     * ausgegangen, dass ein Spiel nicht mehr als 1024   *
     * Halbzüge in Anspruch nimmt.                       *
     * Es werden zwei Array mit der Größe 1024 angelegt. *
     * Eines enthält die 64bit-HashKeys, das andere den  *
     * zu dieser Position entsprechenden besten Zug als  *
     * String.                                           *
     *****************************************************/
    MoveList() {
        size=1024;
        rootIndex=0;
        hashKeys = new unsigned long[size];
        bestmoves = new string[size];
        remainingDepth = new string[size];
    };
    
    /*********************************************************
     * Die Variable 'roodIndex' gibt die Anzahl der Halbzüge *
     * des Spiels an. Bei Initialisierung eines neuen Spiels *
     * muss daher der Index auf 0 gesetzt werden. Ein        *
     * Überschreiben der zuvor gespeicherten Werte ist daher *
     * nicht notwendig.                                      *
     *********************************************************/
    void reset() {
        rootIndex=0;
    }
    
    /***********************************************************
     * Fügt einen neuen HashKey und den dazugehörigen besten   *
     * Zug in das Array ein. Da keine dynamische Datenstruktur *
     * verwendet wird, wird hier einfach der rootIndex um den  *
     * Wert 1 erhöht.                                          *
     ***********************************************************/
    void addMove(unsigned long hashKey, string move) {
        hashKeys[rootIndex]=hashKey;
        bestmoves[rootIndex]=move;
        rootIndex++;
    }
    
    /****************************************************************
     * Während der NegaMax-Suche wird der aktuelle Suchpfad relativ *
     * zum aktuellen rootIndex in der Game-History gespeichert.     *
     ****************************************************************/
    void setMoveAt(int index, unsigned long hashKey, string move, int rmDepth) {
        index+=rootIndex;
        hashKeys[index]=hashKey;
        bestmoves[index]=move;
        remainingDepth[index]=int2String(rmDepth);
    }
    
    /*************************************************
     * Bei Programmende den Speicherplatz der beiden *
     * Arrays freigeben.                             *
     *************************************************/
    virtual ~MoveList() {
        delete []hashKeys;
        delete []bestmoves;
    };
    
    /*******************************************************
     * Gibt den besten Zug als String am übergebenen Index *
     * der GameHistory zurück.                             *
     *******************************************************/
    string getBestMoveAt(int index) {
        return bestmoves[index];
    }

    /******************************************************
     * Gibt den HashKey der Position am übergebenen Index *
     * der GameHistory zurück.                            *
     ******************************************************/
    unsigned long getHashKeyAt(int index) {
        return hashKeys[index];
    }
    
    /***************************************************************
     * Prüft ob die an der angebenen Tiefe (relativ zum rootIndex) *
     * gespeicherte Position bereits zuvor in der GameHistory      *
     * gespeichert ist. Da Wiederholungen nur jeweils nach vier    *
     * Halbzügen auftreten können, werden hier auch nur jeweils    *
     * die HashKeys aller vierfachen Positionen betrachtet, die    *
     * unterhalb der aktuellen Position gespeichert sind.          *
     ***************************************************************/
    bool isRepetition(int depth) {
        int pos = depth+rootIndex;
        for (int i=pos-4;i>=0;i-=4) {
            if (hashKeys[pos]==hashKeys[i])
                return true;
        }
        return false;
    }
    
    /*************************************************************
     * Gibt eine Folge der Hashkeys des aktuellen Spiels von der *
     * Startposition bis zur aktuellen Position aus.             *
     *************************************************************/
    void printGameList(int depth) {
        for (unsigned int i=0;i<=depth+rootIndex;i++)
            // cout << hashKeys[i] << " ";
            cout << bestmoves[i] << " ";
        cout << endl;
    }
    
    string getGameList(int depth) {
        string game;
        for (unsigned int i=0;i<=depth+rootIndex;i++)
            game.append(bestmoves[i]+" ("+remainingDepth[i]+"), ");
        return game;
    }
    
    /***************************************************
     * Gibt die Anzahl der Halbzüge des Spiels zurück. *
     ***************************************************/
    unsigned int getNumHalfMoves() {
        return rootIndex;
    }
    
private:
    unsigned int size;
    unsigned int rootIndex;
    unsigned long *hashKeys;
    string *bestmoves;
    string *remainingDepth;
    
};

#endif	/* MOVELIST_H */

