/* 
 * File:   main.cpp
 * Author: peter
 *
 * Created on 31. Jänner 2012, 18:18
 */

#include <cstdlib>
#include <iostream>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <iomanip>
#include <math.h>
#include <vector>
#include <ctime>
#include "Position.h"
#include "main.h"
#include "Perft.h"
#include "Egdb.h"
#include "globals.h"
#include "HashT.h"
#include "EpdSuite.h"
#include "MoveList.h"
#include "Timer.h"
#include "Logger.h"


using namespace std;


HashT hashTable;
Perft perft;
MoveList moveList;
Timer timer;
Logger logger;
Evalstat evalstat;
unsigned short int distances[64][64];


int main(int argc, char** argv) {
    
    cout << "argc=" << argc << endl;
    for (int i=0; i < argc; i++)
        cout << "argv[" << i << "]=" << argv[i] << endl;
    cout << endl;

    /*
    cout << "--------------------------------------------" << endl;
    cout << "--                                        --" << endl;
    cout << "--   Welcome to Magdalena                 --" << endl;
    cout << "--   Chess-Engine                         --" << endl;
    cout << "--                                        --" << endl;
    cout << "--   programmed by Peter Garscha          --" << endl;
    cout << "--   peter [at] rlsux [dot] com           --" << endl;
    cout << "--                                        --" << endl;
    cout << "--------------------------------------------" << endl << endl;
       */
    
    for (int i=0;i<64;i++) {
        for (int j=0;j<64;j++) {
            int row_i=i/8;
            int row_j=j/8;
            int col_i=i%8;
            int col_j=j%8;
            distances[i][j]=14-(abs(row_i-row_j)+abs(col_i-col_j))/2;
        }
    }
    
    hashTable.init(pow(2,25));
    Position *myPosition = new Position();
    
    
    char input[4096];
    string check;
    
    bool loop=true;
    while (loop) {
        check.clear();
        cin.getline(input,4096);
        check.append(input);
        string logStr=check;
        logger.log(logStr);        
        check = nicerCmd(check);
        // log(check);
        
        if (check=="uci") {
            cout << "id name Magdalena\n";
            cout << "id author Peter Garscha\n";
            cout << "uciok\n";
        }
        
        else if (check.size()==0) {}
        
        else if (check=="isready")
            cout << "readyok\n";
        
        
        else if (check=="stop") {}
        
        else if (check=="debug on")
            cout << "debugging on now" << endl;
        
        else if (check=="debug off")
            cout << "debugging off now" << endl;
        
        else if (check=="ucinewgame") {
            hashTable.clearHash();
            moveList.reset();
            cout << "ucinewgame ok" << endl;
        }
        
        else if (check=="perft")
            perft.doPerft((*myPosition).getFen(),5);

        else if (check=="perft all") {
            perft.doPerft("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1",4);
            perft.doPerft("r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq -",4);
            perft.doPerft("8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - -",4);
            perft.doPerft("r3k2r/Pppp1ppp/1b3nbN/nP6/BBP1P3/q4N2/Pp1P2PP/R2Q1RK1 w kq - 0 1",4);
            perft.doPerft("r2q1rk1/pP1p2pp/Q4n2/bbp1p3/Np6/1B3NBn/pPPP1PPP/R3K2R b KQ - 0 1",4);
        }
        
        else if (check=="perft list")
            perft.doPerftList();

        else if (check.substr(0,5)=="perft" && check.size()>5) {
            int depth=string2int(check.substr(6,check.size()));
            if (depth<8 && depth>0)
                perft.doPerft((*myPosition).getFen(),depth);
            else
                cout << "given depth is invalid for perft-test" << endl;
        }
        
        else if (check.substr(0,6)=="pseudo" && check.size()>6) {
            int depth=string2int(check.substr(7,check.size()));
            if (depth<8 && depth>0)
                perft.doPseudoPerft(*myPosition,depth);
            else
                cout << "given depth is invalid for pseudo-test" << endl;
        }
        
        else if (check.substr(0,4)=="eval" || check.substr(0,2)=="go") {
            
            /***********************************************
             * Das übergebene Command verarbeiten und die  *
             * einzelnen Teile im vector fields speichern. *
             ***********************************************/
            vector <string> fields;

            while(true) {
                int pos=check.find_first_of(" ");
                if (pos>=0) {
                    fields.push_back(check.substr(0,pos));
                    check=check.substr(pos+1);
                } else {
                    fields.push_back(check);
                    break;
                }
            }            
            
            unsigned int depth=0;
            unsigned int btime=0;
            unsigned int wtime=0;
            unsigned int movestogo=0;
            
            for (unsigned int i=0;i<fields.size();i++) {
                if (fields[i]=="depth" && (i+1)<fields.size()) {
                    unsigned int newDepth=string2int(fields[i+1]);
                    if (newDepth>=1)
                        depth=newDepth;
                    if (depth>10)
                        depth=10;
                    continue;
                } else if (fields[i]=="wtime" && (i+1)<fields.size()) {
                    wtime=string2int(fields[i+1]);
                    continue;
                } else if (fields[i]=="btime" && (i+1)<fields.size()) {
                    btime=string2int(fields[i+1]);
                    continue;
                } else if (fields[i]=="movestogo" && (i+1)<fields.size()) {
                    movestogo=string2int(fields[i+1]);
                    continue;
                }                
            }
            
            if (movestogo==0)
                movestogo=1;
                        
            /***************************************************
             * Suche vorbereiten, dh. TranspositionAge erhöhen *
             * TimeManagement festlegen, usw.                  *
             ***************************************************/
            hashTable.inkTranspositionAge();
            (*myPosition).sum=0;
            
            bool side2move=(*myPosition).onMove;
            
            int timeForMove=-1;
            if (side2move && btime)
                timeForMove=btime;
            else if (!side2move && wtime)
                timeForMove=wtime;

            
            
                    
            timer.resetTimer();
            timer.startTimer();
            
            timer.setTimeForMove(timeForMove,movestogo);
                                   
            if (depth==0 && timeForMove>0)
                depth=90;
            
            /********************************
             * Iteratives Deepening starten *
             ********************************/
            
            int score;
            int aspirationWindow=50;
            int infinity=100000;
            
            score=(*myPosition).negaMax(-infinity,infinity,0,1,false,false);

            string foundMove=(*myPosition).getMoveString((*myPosition).bestMoveIndex);

            unsigned long iterationTime=0;
            
            for (unsigned int currDepth=2;currDepth<=depth;currDepth++) {
                (*myPosition).sum=0;
                iterationTime=timer.getTime();
                
                
                int tempScore=(*myPosition).negaMax(score-aspirationWindow,score+aspirationWindow,0,currDepth,false,false);
                
                if (tempScore>score-aspirationWindow && tempScore<score+aspirationWindow)
                    score=tempScore;
                else
                    score=(*myPosition).negaMax(-infinity,infinity,0,currDepth,false,false);
                
                /**************************************************
                 * Iteratives Deepening abbrechen, falls maximale *
                 * Zeit für die Zugsuche verstrichen ist.         *
                 **************************************************/
                if (timer.isTimeout()) { // timeout während letzter Suche
                    break;
                } if (timer.checkTimeout()) { // hier sollte abgebrochen werden, wenn die restliche Zeit für eine erneute Suche nicht ausreicht
                    break;                    
                }
                
                foundMove=(*myPosition).getMoveString((*myPosition).bestMoveIndex);
                string a="depth: ";
                a.append(int2String(currDepth));
                cout << a << endl;


                /***********************************************
                 * Iteratives Deepening abbrechen, falls nicht *
                 * genug Zeit für eine neue Iteration.         *
                 ***********************************************/
                
                iterationTime=timer.getTime()-iterationTime;
                if ((timer.getTimeout()-timer.getTime())<(iterationTime*2))
                    break;
            }
            
            /************************************
             * Das Ergebnis der Suche ausgeben. *
             ************************************/
            timer.stopTimer();
            double seconds = timer.getStartEndDiffSeconds();
            if ((fields.size()>0 && fields[0]=="eval") || check.substr(0,4)=="eval") {
                // cout << "Move: " << (*myPosition).getMoveString((*myPosition).bestMoveIndex) << endl;
                cout << "Move: " << foundMove << endl;
                cout << "Score: " << score << endl;
                cout << "Positions: " << (*myPosition).sum << endl;
                cout.setf(ios::fixed, ios::floatfield);
                cout << "Time: " << setprecision(5) << seconds << " seconds" << endl;
            } else {
                // cout << "bestmove " << (*myPosition).getMoveString((*myPosition).bestMoveIndex) << endl;
                cout << "bestmove " << foundMove << endl;
            }
        }
        
        else if (check=="egdb") {
            Egdb egdb;
            egdb.build();
        }
        
        else if (check=="print")
            (*myPosition).printBoard();
        
        else if (check=="info")
            (*myPosition).getInfo();
               
        else if (check=="epdsuite") {
            EpdSuite::testCompleteSuite();
        }

        else if (check.substr(0,6)=="divide") {
            bool modus=false;
            if (check.substr(0,7)=="divide2")
                modus=true;
            
            size_t found;
            found=check.find(" ");
            
            string strDepth = check.substr(int(found)+1,check.size()-int(found));
            int depth = atoi(strDepth.c_str());
            
            if (depth>0) {
                long sum=0;
                long total=0;
                int color;
                if ((*myPosition).onMove==true)
                    color=1;
                else color=0;
         
                (*myPosition).getDivide(color,1,depth,modus,sum,total);
                cout << "Nodes: " << total << endl;
            } else {
                cout << "no valid depth given. must be a valid number greater than 0" << endl;
            }
        }
        
        else if (check=="new") {
            delete myPosition;
            myPosition=new Position();
        }
        
        else if (check=="fen")
            cout << endl << "fen: " << (*myPosition).getFen() << endl << endl;
        
        else if (check.substr(0,8)=="position") {
            
            moveList.reset();
            
            vector <string> fields;

            while(true) {
                int pos=check.find_first_of(" ");
                if (pos>=0) {
                    fields.push_back(check.substr(0,pos));
                    check=check.substr(pos+1);
                } else {
                    fields.push_back(check);
                    break;
                }
            }
            
            if (fields.size()>2 && fields[1]=="fen") {
                string newFen=fields[2];
                for (unsigned int i=3;i<fields.size();i++) {
                    if (fields[i]!="moves") {
                        newFen.append(" ");
                        newFen.append(fields[i]);
                    } else
                        break;
                }
                cout << newFen << endl;
                
                delete myPosition;
                myPosition=new Position(newFen);                
            }
            
            if (fields.size()>1 && fields[1]=="startpos") {
                delete myPosition;
                myPosition=new Position();
            }
            
            bool doMoves=false;
            for (unsigned int i=0;i<fields.size();i++) {
                if (!doMoves && fields[i]=="moves") {
                    doMoves=true;
                    continue;
                }
                if (doMoves) {

                    string moveArray[256];
                    int moveIndex[256];
                    int cntMoves = myPosition->getMMChilds(moveArray, moveIndex);
                    bool found=false;
                    int realMoveIndex;
                    for (int k=0;k<cntMoves;k++) {
//                        cout << moveArray[k] << "    " << ((myPosition->pawns[myPosition->onMove])&(myPosition->mvD[0][moveIndex[k]][0]))<<"     "<< (myPosition->mvD[0][moveIndex[k]][1]&myPosition->promotionRow[myPosition->onMove]) << "   " << moveIndex[k] << endl;
                        if (moveArray[k]==fields[i]) {
                            found=!found;
                            realMoveIndex=moveIndex[k]; // notwendig, weil nicht immer alle Züge gültig sind
                            break;
                        }
                    }
                    if (found) {
                        // unsigned long positionHashKey=hashTable.getHashKey(myPosition);
                        myPosition->move_figure(myPosition->mvD[0][realMoveIndex][0],myPosition->mvD[0][realMoveIndex][1],realMoveIndex%4);
                        string moveString=myPosition->getMoveString(realMoveIndex);
                        // moveList.addMove(positionHashKey,moveString);
                        moveList.addMove(myPosition->hashKey,moveString);
                    }
                    
                }               
            }
            
            
        }
        
        else if (check.substr(0,8)=="setboard") {
            if (check.size()>9) {
                string newFen = check.substr(9,check.size()-9);
                cout << newFen << endl;
                delete myPosition;
                myPosition=new Position(newFen);
            }
        }
        
        else if (check=="quit" || check=="exit") {
            cout << endl << "Goodbye!" << endl;
            loop=false;
        }
        
        else if (check=="help" || check=="?") {
            cout << "divide [n] ........... divide position with given depth" << endl;
            cout << "divide2 [n] .......... divide position with given depth" << endl;
            cout << "egdb ................. build end game database" << endl;
            cout << "eval ................. evaluate position" << endl;
            cout << "exit ................. exit engine" << endl;
            cout << "fen .................. print fen of current position" << endl;
            cout << "help ................. show this menu" << endl;
            cout << "info ................. print additional info about current position" << endl;
            cout << "new .................. reset position" << endl;
            cout << "perft [n] ............ standard performance test" << endl;
            cout << "                       (with current/given position)" << endl;
            cout << "perft all ............ performance test with 5 positions" << endl;
            cout << "perft list ........... test 126 positions to depth 6 (takes a while)" << endl;
            cout << "print ................ print board of current position" << endl;
            cout << "setboard [fen] ....... set current position to given fen" << endl;
            cout << "quit ................. exit engine" << endl;
            cout << "? .................... show this menu" << endl;
        }
        
        else {
            string moveArray[256];
            int moveIndex[256];
            int cntMoves = myPosition->getMMChilds(moveArray,moveIndex);
            if (cntMoves==0) cout << "no valid move available or checkmate" << endl;
            bool found=false;
            for (int i=0;i<cntMoves;i++) {
                if (moveArray[i]==check) {
                    found=!found;
                    break;
                }
            }
            if (found) cout << "valid move" << endl;
            else cout << "invalid move / unknown command: " << check << endl;
        }
    }
    
    delete myPosition;
    

   
    return 0;
}