/* 
 * File:   Position.cpp
 * Author: peter
 * 
 * Created on 31. Jänner 2012, 18:23
 */

#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "Position.h"
#include "Heap.h"
#include "HashT.h"
#include "globals.h"

using namespace std;

/************************************
 * Initialisiert eine Startposition *
 ************************************/
Position::Position() {
    ep=0;
    kings[1]=Position::BIN[4];
    queens[1]=Position::BIN[3];
    bishops[1]=Position::BIN[2] | Position::BIN[5];
    knights[1]=Position::BIN[1] | Position::BIN[6];
    rooks[1]=Position::BIN[0] | Position::BIN[7];
    pawns[1]=Position::BIN[8] | Position::BIN[9] | Position::BIN[10] | Position::BIN[11] | Position::BIN[12] | Position::BIN[13] | Position::BIN[14] | Position::BIN[15];
    kings[0]=Position::BIN[60];
    queens[0]=Position::BIN[59];
    bishops[0]=Position::BIN[58] | Position::BIN[61];
    knights[0]=Position::BIN[57] | Position::BIN[62];
    rooks[0]=Position::BIN[56] | Position::BIN[63];
    pawns[0]=Position::BIN[48] | Position::BIN[49] | Position::BIN[50] | Position::BIN[51] | Position::BIN[52] | Position::BIN[53] | Position::BIN[54] | Position::BIN[55];
    updateBitboardLongs();
    castlingRules=15;
    onMove=false;
    
    hashKey = hashTable.getHashKey(this);
}

/*****************************************************************
 * Initialisiert eine Position durch den übergebenen FEN-String. *
 *****************************************************************/
Position::Position(const string &fen) {
    
    castlingRules=0;
    
    ep=0;
//    targetmap=0L;

    aw=0L;
    ab=0L;
    all=0L;
    
    pawns[0]=0L;
    pawns[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    bishops[0]=0L;
    bishops[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    queens[0]=0L;
    queens[1]=0L;
    knights[0]=0L;
    knights[1]=0L;
    kings[0]=0L;
    kings[1]=0L;
    
    size_t found;
    found=fen.find(" ");
    string fengrp0 = fen.substr(0,int(found));
    string rest = fen.substr(int(found)+1,fen.size()-int(found));
    string fengrp1 = rest.substr(0,1);
    rest = rest.substr(2,rest.size()-2);
    found=rest.find(" ");
    string fengrp2 = rest.substr(0,int(found));
    rest = rest.substr(int(found)+1,rest.size()-int(found));
    string fengrp3 = rest;

    // 1. group: figures
    int pos=0;
    for (unsigned int i=0;i<fengrp0.length();i++) {
        if (fengrp0.at(i)=='/') continue;
        else if (fengrp0.at(i)=='1') pos=pos+1;
        else if (fengrp0.at(i)=='2') pos=pos+2;
        else if (fengrp0.at(i)=='3') pos=pos+3;
        else if (fengrp0.at(i)=='4') pos=pos+4;
        else if (fengrp0.at(i)=='5') pos=pos+5;
        else if (fengrp0.at(i)=='6') pos=pos+6;
        else if (fengrp0.at(i)=='7') pos=pos+7;
        else if (fengrp0.at(i)=='8') pos=pos+8;
        else if (fengrp0.at(i)=='k') { kings[1]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='q') { queens[1]=queens[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='n') { knights[1]=knights[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='b') { bishops[1]=bishops[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='r') { rooks[1]=rooks[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='p') { pawns[1]=pawns[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='K') { kings[0]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='Q') { queens[0]=queens[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='N') { knights[0]=knights[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='B') { bishops[0]=bishops[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='R') { rooks[0]=rooks[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='P') { pawns[0]=pawns[0]|Position::BIN[pos]; pos++;}
    }
    
    // 2. group: color on move, fengrp1
    if (fengrp1=="b")
        onMove=true;
    else
        onMove=false;

    // 3. group: castles
    if (fengrp2.at(0)!='-') {
        for (unsigned int i=0;i<fengrp2.size();i++) {
            if (fengrp2.at(i)=='K')   castlingRules^=1; // white_castling_short=true;
            if (fengrp2.at(i)=='Q')   castlingRules^=2; // white_castling_long=true;
            if (fengrp2.at(i)=='k')   castlingRules^=4; // black_castling_short=true;
            if (fengrp2.at(i)=='q')   castlingRules^=8; // black_castling_long=true;            
        }
    }
    
    // 4. group: en passant
    int tmp_ep=0;
    if (fengrp3.length()==2) {
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='6') tmp_ep=1;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='6') tmp_ep=2;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='6') tmp_ep=3;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='6') tmp_ep=4;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='6') tmp_ep=5;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='6') tmp_ep=6;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='6') tmp_ep=7;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='6') tmp_ep=8;
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='3') tmp_ep=9;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='3') tmp_ep=10;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='3') tmp_ep=11;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='3') tmp_ep=12;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='3') tmp_ep=13;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='3') tmp_ep=14;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='3') tmp_ep=15;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='3') tmp_ep=16;
    }
    ep=tmp_ep;
    
    updateBitboardLongs();
    
    hashKey = hashTable.getHashKey(this);
}

/**************************************************************
 * Wird ein UCI-Command oder ein Cmd des Benutzers            *
 * nicht verstanden, so wird diese Funktion aufgerufen,       *
 * um zu prüfen, ob es sich bei dem Cmd um einen zur Position *
 * gültigen Zug handelt (siehe main.cpp)                      *
 **************************************************************/
int Position::getMMChilds(string moveArray[], int moveIndex[]) {
    cd=0;
    doBackup();
    calc_moves();
    int cntMoves=0;
    for (int i=0;i<mvSize[cd];i++) {
        unsigned short int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);        
        if (!catchKing) {
            moveArray[cntMoves]=getMoveString(i);
            moveIndex[cntMoves]=i;
            cntMoves++;
        }
    }
    return cntMoves;
}

/**********************************************************************
 * Gibt zur gegebenen Position und Tiefe die Anzahl der Nodes         *
 * für jeden möglichen Zug an.                                        *
 * divide [n] ... Anzahl der Blätter/Nodes in Tiefe n                 *
 * divide2 [n] ... Summe aller Nodes bis inkl. Tiefe n                *
 * Diese Funktionen sind speziel zum Debugging des Zug-Generators     *
 * wertvoll. An dieser Stelle einen großen Dank an                    *
 * Roman Hartmann (http://www.rocechess.ch/roce.html),dessen Programm *
 * ich selbst verwendet hatte um meine Engine zu debuggen             *
 **********************************************************************/
void Position::getDivide(int color, int curr_depth, const int &depth, const bool &modus, long &sum, long &total) {
    cd = curr_depth-1;
    calc_moves();
    doBackup();
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            unsigned short int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);            
            if (curr_depth==1)
                sum=0;
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            if (!catchKing) {
                getDivide(abs(color-1), (curr_depth+1),depth, modus, sum, total);
                cd--;
                if (curr_depth==1)
                    cout << getMoveString(i) << " " << sum << "     " << getFen() << endl;                                
            } else {
                if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
                    sum--;
                    total--;
                }                
            }
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);            
        }
    }
    if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
        sum+=mvSize[cd];
        total+=mvSize[cd];
    }
}

/*************************************************************************
 * Führt einen Performance-Test zur aktuellen Position aus und           *
 * liefert ausführliche Daten zu Captures, En-Passants, Promotions, etc. *
 *************************************************************************/
void Position::getPerftInfo(int curr_depth, int depth, Perftres *perft) {
    
    cd = curr_depth-1;
    doBackup();
    calc_moves();
    
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            unsigned short int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
            
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            
            if (!catchKing) {
                getPerftInfo((curr_depth+1),depth, perft);
                cd--;

                // count captures
//                if (ret)
//                    perft->perftres[1+(curr_depth-1)*7]++;                    

                // count ep-moves
                // if (mvD[cd][i][2]!=0L) { // <-- funktioniert nicht mehr, kill / [2] nicht mehr vorhanden
                if (mvD[cd][i][1]!=0L) {                
                    perft->perftres[4+(curr_depth-1)*7]++;
                  //  perft->perftres[1+(curr_depth-1)*7]++;
                } 

                // count promotions
//                if ([cd*256+i]>0)
//                    perft->perftres[5+(curr_depth-1)*7]++;                                
                
                // count castlings
                if ((mvD[cd][i][1]&kings[!onMove])!=0L && (mvD[cd][i][0]<<2==mvD[cd][i][1] || mvD[cd][i][0]>>2==mvD[cd][i][1]))
                    perft->perftres[6+(curr_depth-1)*7]++;                     
            
            } else
                perft->perftres[(curr_depth-1)*7]--;                
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
        }
    }    

    perft->perftres[(curr_depth-1)*7]+=mvSize[cd];
    
    if (perft!=NULL) {
        /*        


        // count checks
        Position tmp_pos3(*tmp_pos);
        if (opponent_target_bitboard(tmp_pos3,tmp_pos->kings[abs(color-1)],abs(color-1)))
            perft->perftres[2+(curr_depth-1)*7]++;                

        */
    }    
}

/*****************************************************************************
 * Sortiert das Array mit den möglichen Zügen einer Position.                *
 * Züge mit Captures werden bevorzugt.                                       *
 * Wert eines Zuges ist abhängig welcher Figuren-Typ welchen gegnerischen    *
 * Figuren-Typ schlägt, dh. Bauer schlägt Dame ist ein wertvollerer Zug als  *
 * Dame schlägt Bauer, usw.                                                  *
 * Die Sortierung dient dazu, beim NegaMax früher bzw. bessere Cutoffs       *
 * zu bekommen, somit sollte sich die Zeit der Zugsuche deutlich reduzieren. *
 *****************************************************************************/
void Position::sortMoves(unsigned short int sortedIndizes[], int cd, int bestmoveHash) {
    unsigned short int cntMoves=mvSize[cd];
    short int moveValues[cntMoves];

    /*************************************************************
     * Jedem Zug einen Score zuweisen. Züge ohne Captures        *
     * werden mit dem Wert -2000 bewertet. Züge mit Captures mit *
     * einem Wert abhängig davon, welcher Figurentyp welchen     *
     * gegnerischen Figurentyp schlägt.                          *
     *************************************************************/
    for (int i=0;i<cntMoves;i++) {
        
        if ((mvD[cd][i][1]&all)==0L) { // all kann ersetzt werden
            moveValues[i]=-2000;
            continue;
        }
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else                                            fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else                                               toValue=1;        

        moveValues[i]=toValue-fromValue;
    }
    
    /**************************************************
     * Index des besten Zuges aus dem Hash markieren. *
     **************************************************/
    if (bestmoveHash>=0 && bestmoveHash<cntMoves)
        moveValues[bestmoveHash]=10000;
    
    /********************
     * Werte sortieren. *
     ********************/
    for (int i=0;i<cntMoves;i++)
        sortedIndizes[i]=i;
    
    Heap::buildHeap(sortedIndizes,moveValues,cntMoves);    
}

unsigned short int Position::sortMovesQuiesce(unsigned short int sortedIndizes[], int cd) {
    unsigned short int cntMoves=mvSize[cd];
    short int moveValues[cntMoves];

    // calculate a value for every move
    for (int i=0;i<cntMoves;i++) {
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else                                            fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else                                               toValue=1;        

        moveValues[i]=toValue-fromValue;
    }
    
    for (int i=0;i<cntMoves;i++)
        sortedIndizes[i]=i;
    
    Heap::buildHeap(sortedIndizes,moveValues,cntMoves);
    
    return 0;
}


void Position::getMoveScores(short int moveScores[], int cd) {
    for (int i=0;i<mvSize[cd];i++) {
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else                                            fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else                                               toValue=1;        

        moveScores[i]=toValue-fromValue;
    }    
}


/**************************************************************
 * NegaMax ist die Kernfunktion zur Suche nach dem besten Zug *
 * Aktuell implementiert ist dabei:                           *
 *   - eine einfache Zugsortierung                            *
 *   - Transposition/Hash-Table                               *
 *   - Quiescence Search                                      *
 **************************************************************/
int Position::negaMax(int alpha, int beta, int curr_depth, int depth, bool isPv, bool canNull) {

    sum++;
    
    cd = curr_depth;
    
    if (!timer.isTimeout() && (sum&1023)==0)
        timer.checkTimeout();    
    if (timer.isTimeout())
        return 0;    
    
    
    // unsigned long hashKey = hashTable.getHashKey(this);
    moveList.setMoveAt(curr_depth,hashKey,"",0);
    
    /*********************************************************************************
     * Überprüfe ob die aktuelle Position bereits im Spiel- bzw. Suchverlauf vorkam. *
     * Ggf. wird ein Score für ein Draw bzw. ein entsprechender Contempt-Faktor      *
     * zurück gegeben.                                                               *
     *********************************************************************************/

    if (cd>1 && moveList.isRepetition(cd)) {
        int contemptOpening=-50;
        int contemptMiddlegame=-25;
        int contempt=0;
        if (moveList.getNumHalfMoves()<20)
            contempt=contemptOpening;
        else if (!isEndgame())
            contempt=contemptMiddlegame;
        if (cd%2==1)
            contempt=-contempt;
        return contempt;
    }
    
    /****************************************************************************
     * Suche in der HashTable nach der aktuellen Position und prüfe ggf. ob der *
     * gespeicherte Score verwendet werden kann.                                *
     ****************************************************************************/
    int score2;
    int bestmoveHash=-1;
    
    if (cd>1 && hashTable.contains(hashKey,depth-cd,&score2,alpha,beta,&bestmoveHash)) // cd>1 oder cd>0 ???
        return score2;
    
    /****************************************
     * Überprüfung ob König im Schach steht *
     ****************************************/
    bool inCheck = false;
    inCheck=opponent_target_bitboard2(kings[onMove]);

    
    /*********************
     * Null-Move-Pruning *
     *********************/      
    int score=0;  

//    if (depth-curr_depth>3 && canNull && evaluate(onMove)>beta && !isEndgame() && !inCheck) {
//        
////        makeNullMove();
//        
//        int r=2;
//        if (depth-curr_depth>6)
//            r=3;
//        
//        score = -negaMax(-beta,-beta+1,(curr_depth+1),depth-r-1,false,false);
//        cd=curr_depth;
//        
////        undoNullMove();
//        
//        if (timer.isTimeout())
//            return 0;
//        if (score>=beta) {// save hash here
//            // logger.log("nullmove cutoff");
//            return beta;
//        }
//    }
    
    /********************************************************************************
     * Sofern wir in der maximalen Tiefe angekommen sind, soll die Quiescence-Suche *
     * aufgerufen werden.                                                           *
     ********************************************************************************/
    if (curr_depth==depth) {
        if (depth==1)
            return evaluate(onMove);
        else
            score = quiesce(alpha,beta,curr_depth); // hashScore hier speichern
        
//        hashTable.insert(hashKey,maxWert,nodeType,bmIndex,depth-cd);        
        
        return score;
    }

    /************************************************************************
     * Züge der aktuellen Position generieren und zwecks Pruning sortieren. *
     ************************************************************************/
    calc_moves();
    unsigned short int sortedIndizes[256];
    sortMoves(sortedIndizes,cd,bestmoveHash);
    doBackup();

    unsigned int nodeType=3;
    int maxWert=alpha;
    bool raisedAlpha=false;
    
    int i;
    int bmIndex=-1;

    for (i=mvSize[cd]-1;i>=0;i--) {
        unsigned short int ret = move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4);        
        
        /***********************************************************************************
         * Prüft ob der Zug legal ist. Das ist dann der Fall, wenn der gegnerische Spieler *
         * anschließend nicht den König schlagen kann. Da diese Überprüfung sehr aufwendig *
         * ist, wird sie nur in der untersten Ebene durchgeführt, damit die Engine nur     *
         * einen legalen Zug als besten Zug wählen kann.                                   *
         ***********************************************************************************/
//        if (curr_depth==0) {
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            if (catchKing) {
                undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);
                continue;
            }
//        }
    
        /*************************************
         * Check-Extension                   *
         * Wenn der König im Schach steht,   *
         * soll die Suchtiefe erhöht werden. *
         *************************************/
        unsigned short int ext=0;
        if (depth>1 && opponent_target_bitboard2(kings[onMove])) {
            ext=1;
        }

        /*****************************
         * Principal-Variation-Suche *
         *****************************/
        if (depth==1) {
            if (i==mvSize[cd]-1 || -negaMax(-maxWert-1,-maxWert,(curr_depth+1),depth+ext,false,true)>maxWert) // {
                score = -negaMax(-beta,-maxWert,(curr_depth+1),depth+ext,true,true);
        } else {
            // score = -negaMax(!side2move,-beta,-maxWert,(curr_depth+1),depth,sum,true);
            if (!isPv || (depth-curr_depth)<3)
                score = -negaMax(-beta,-maxWert,(curr_depth+1),depth+ext,false,true);
            else {
                if (!raisedAlpha)
                    score = -negaMax(-beta,-maxWert,(curr_depth+1),depth+ext,true,true);
                else {
                    if (-negaMax(-maxWert-1,-maxWert,(curr_depth+1),depth+ext,false,true)>maxWert)
                        score = -negaMax(-beta,-maxWert,(curr_depth+1),depth+ext,true,true);
                }
            }
        }

//        score = -negaMax(-beta,-maxWert,(curr_depth+1),depth,true,true);
        cd=curr_depth;
        
        /*****************************************************
         * Zug rückgängig machen und den Score des Zuges     *
         * ggf. speichern, bzw. überprüfen ob die restlichen *
         * Züge noch durchgeführt werden müssen.             *
         *****************************************************/
        undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);        

        if (score==0 && timer.isTimeout())
            return 0;
        
        if (score > maxWert) {
            raisedAlpha=true;
            bmIndex=sortedIndizes[i];
            maxWert = score;
            if (curr_depth==0)
                bestMoveIndex=sortedIndizes[i];
            if (maxWert>=beta) {
                nodeType=2;
                maxWert=beta;
                break;
            }
                       
            nodeType=1;
        }
    }

    /**************************************************
     * Den berechneten Score in der HashTable und den *
     * besten Zug in der GameHistory speichern.       *
     **************************************************/
    hashTable.insert(hashKey,maxWert,nodeType,bmIndex,depth-cd);
    moveList.setMoveAt(curr_depth,hashKey,getMoveString(bmIndex),(depth-curr_depth));
    return maxWert;
}

bool Position::isCheckMate() {
       
    calc_moves();
    doBackup();
    int i;
    cd = 0;
    for (i=mvSize[cd]-1;i>=0;i--) {
        unsigned short int ret = move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
        
        /***********************************************************************************
         * Prüft ob der Zug legal ist. Das ist dann der Fall, wenn der gegnerische Spieler *
         * anschließend nicht den König schlagen kann. Da diese Überprüfung sehr aufwendig *
         * ist, wird sie nur in der untersten Ebene durchgeführt, damit die Engine nur     *
         * einen legalen Zug als besten Zug wählen kann.                                   *
         ***********************************************************************************/
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        if (catchKing) {
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
            continue;
        }
        
        
        return false; // Zug gültig, dh. Spielfarbe kann mit diesem Zug dem Schach entweichen. Daher kann es nicht Schachmatt sein.
        
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
    }
    return true; // Alle prinzipiell möglichen Züge sind ungültig, dh. Spielfarbe kann nicht aus dem Schach fliehen. Daher Schachmatt.
}

list<string> Position::getMovePositions() {
    list<string> positions;    

    cd = 0;
    calc_moves();
    doBackup();
    int i;
    for (i=mvSize[cd]-1;i>=0;i--) {


        unsigned short int ret = move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
        
        /***********************************************************************************
         * Prüft ob der Zug legal ist. Das ist dann der Fall, wenn der gegnerische Spieler *
         * anschließend nicht den König schlagen kann. Da diese Überprüfung sehr aufwendig *
         * ist, wird sie nur in der untersten Ebene durchgeführt, damit die Engine nur     *
         * einen legalen Zug als besten Zug wählen kann.                                   *
         ***********************************************************************************/
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        if (catchKing) {
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
            continue;
        }

//        printBoard();
//        cout << bitScanForward(mvD[cd][i][0]) << "," << bitScanForward(mvD[cd][i][1]) << endl;
//        cout << getFenShort() << endl;
//        cout << pawns[!onMove] << endl;
//        cout << kings[!onMove] << endl;
//        cout << pawns[onMove] << endl;
//        cout << kings[onMove] << endl;
//        cout << endl;        

        
        
        positions.push_back(getFenShort());
        
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
    }    
    
//    abort();
    return positions;
}


/******************************************************
 * Gibt zurück, ob es sich bei der aktuellen Position *
 * um eine Position im Endspiel handelt oder nicht.   *
 * Ein Endspiel ist gegeben, wenn:                    *
 * - keine der Damen mehr im Spiel sind,              *
 * - einer der Spieler weniger als 8 Figuren hat      *
 ******************************************************/
bool Position::isEndgame() {
    if (queens[0]==0UL && queens[1]==0UL)
        return true;
    if (popCount(aw)<8 || popCount(ab)<8)
        return true;
    return false;
}

/************************************************************************
 * Die Quiescence-Suche wird an jedem Leaf-Node des NegaMax aufgerufen. *
 * Betrachtet werden hierbei nur noch jene Züge die Figuren schlagen.   *
 * Im Idealfall wird die Suche solange fortgesetzt, bis eine Position   *
 * erreicht wurde, in der keine Figur geschlagen werden kann oder nicht *
 * mehr tiefer evaluiert werden muss.                                   *
 * StandingPat wurde bereits implementiert. Es fehlt allerdings noch    *
 * die Implementierung der Static Exchange Evaluation bzw. auch         *
 * Delta-Pruning.                                                       *
 ************************************************************************/
int Position::quiesce(int alpha, int beta, int curr_depth) {
       
    cd = curr_depth;

//    sum++;
//    
//    cd = curr_depth;
//    
//    if (!timer.isTimeout() && (sum&1023)==0)
//        timer.checkTimeout();    
//    if (timer.isTimeout())
//        return 0;    
    
    
    int standPat = evaluate(onMove);
    if (standPat>=beta)
        return beta;
    if(standPat>alpha)
        alpha=standPat;

    calcCaptures();
    if (!mvSize[cd])
        return alpha;

    unsigned short int sortedIndizes[256];
    sortMovesQuiesce(sortedIndizes,cd);

//    unsigned short int moveScores[256];
//    getMoveScores(moveScores,cd);
    
    doBackup();
    for (int i=mvSize[cd]-1;i>=0;i--) {    
        
        if ((mvD[cd][sortedIndizes[i]][1]&kings[!onMove])!=0)
            return 100000;
        
        unsigned short int ret=move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4);
        int score = -quiesce(-beta,-alpha,(curr_depth+1));
        cd--;
        undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);
        if (score > alpha) {
            if (score>=beta) 
                return beta;
            alpha=score;
        }
    }
    return alpha;
}

/***********************************************************************
 * Die evaluate-Funktion bewertet eine gegebene Position               *
 * nach verschiedenen heuristischen Methoden für jeweils beide Spieler *
 * Die Werte der Spieler zu jeder Methode werden subtrahiert,          *
 * nach der Priorität gewichtet und zu einem Gesamt-Score              *
 * zusammengefasst.                                                    *
 ***********************************************************************/
int Position::evaluate(bool side2move) {
    
    /******************
     * Material Score *
     * ****************/
    
    int materialScore0, materialScore1;

    materialScore0  =   100 *   popCount(pawns[0]) +
                        275 *   popCount(knights[0]) +
                        325 *   popCount(bishops[0]) +
                        465 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        50000 *  popCount(kings[0]);    
    
    materialScore1  =   100 *   popCount(pawns[1]) +
                        275 *   popCount(knights[1]) +
                        325 *   popCount(bishops[1]) +
                        465 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        50000 *  popCount(kings[1]);        

    /********************************************************************
     * Anpassung des Material Scores abhängig von der Anzahl der Bauern *
     ********************************************************************/
    int knightAdj[9] = { -20, -16, -12, -8, -4,  0,  4,  8, 12};
    int rookAdj[9] =   {  15,  12,   9,  6,  3,  0, -3, -6, -9};
    
    materialScore0 += popCount(knights[0]) * knightAdj[popCount(pawns[0])];
    materialScore0 += popCount(rooks[0]) * rookAdj[popCount(pawns[0])];
    materialScore1 += popCount(knights[1]) * knightAdj[popCount(pawns[1])];
    materialScore1 += popCount(rooks[1]) * rookAdj[popCount(pawns[1])];
    
    /*****************************
     * Keine Bauern mehr am Feld *
     *****************************/
    
    if (popCount(pawns[0])==0)  materialScore0-=16;
    if (popCount(pawns[1])==0)  materialScore1-=16;
    
    /*********
     * Tempo *
     *********/
    
    if (side2move) materialScore1+=10;
    else           materialScore0+=10;     

    /************************************************
     * King Tropism                                 *
     * (Abstand der Figuren zum gegnerischen König) *
     ************************************************/

    unsigned long bitboard;
    
    int kingTropism0=0, kingTropism1=0;

    unsigned short int posKingWhite = bitScanForward(kings[0]);
    unsigned short int posKingBlack = bitScanForward(kings[1]);
    
    int diagNW[64] = {
        0,  1,  2,  3,  4,  5,  6,  7,
        1,  2,  3,  4,  5,  6,  7,  8,
        2,  3,  4,  5,  6,  7,  8,  9,
        3,  4,  5,  6,  7,  8,  9, 10,
        4,  5,  6,  7,  8,  9, 10, 11,
        5,  6,  7,  8,  9, 10, 11, 12,
        6,  7,  8,  9, 10, 11, 12, 13,
        7,  8,  9, 10, 11, 12, 13, 14
    };

    int diagNE[64] = {
         7,  6,  5,  4,  3,  2,  1,  0,
         8,  7,  6,  5,  4,  3,  2,  1,
         9,  8,  7,  6,  5,  4,  3,  2,
        10,  9,  8,  7,  6,  5,  4,  3,
        11, 10,  9,  8,  7,  6,  5,  4,
        12, 11, 10,  9,  8,  7,  6,  5,
        13, 12, 11, 10,  9,  8,  7,  6,
        14, 13, 12, 11, 10,  9,  8,  7
    };    
    
    int bonusDiaDistance[14] = {3, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    
    /*****************
     * weiße Figuren *
     *****************/

    bitboard = knights[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack];
        bitboard-=Position::BIN[x];
    }

    bitboard = rooks[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]/2;
        bitboard-=Position::BIN[x];
    }    

    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]*5/2;
        bitboard-=Position::BIN[x];
    }        
    
    bitboard = bishops[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]/2;
        kingTropism0+=bonusDiaDistance[abs(diagNE[x]-diagNE[posKingBlack])];
        kingTropism0+=bonusDiaDistance[abs(diagNW[x]-diagNW[posKingBlack])];
        bitboard-=Position::BIN[x];
    }        
    
    /********************
     * schwarze Figuren *
     ********************/
    
    bitboard = knights[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = rooks[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]/2;
        bitboard-=Position::BIN[x];
    }    
    
    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]*5/2;
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]/2;
        kingTropism1+=bonusDiaDistance[abs(diagNE[x]-diagNE[posKingWhite])];
        kingTropism1+=bonusDiaDistance[abs(diagNW[x]-diagNW[posKingWhite])];        
        bitboard-=Position::BIN[x];
    }    
    
    
    /******************
     * Pawn Structure *
     ******************/
    
    /********************************************
     * Doppelte Bauern in einer Linie (Penalty) *
     ********************************************/
    
    int pawnScore0=0, pawnScore1=0;
    
    int multiPawnPenalty[7] = {0, 0, 25, 60, 100, 100, 100};
    
    unsigned long cols[8] = {
        72340172838076673,
        144680345676153346,
        289360691352306692,
        578721382704613384,
        1157442765409226768,
        2314885530818453536,
        4629771061636907072,
        9259542123273814144UL
    };
    
    for (unsigned short int i=0;i<8;i++) {
        pawnScore0-=multiPawnPenalty[popCount(cols[i]&pawns[0])];
        pawnScore1-=multiPawnPenalty[popCount(cols[i]&pawns[1])];
    }
    
    /*****************************
     * Bauern im Zentrum (Bonus) *
     *****************************/
    
    unsigned long d4e4bb=103079215104;
    unsigned long d4c4bb=51539607552;
    unsigned long d4e3bb=17626545782784;
    unsigned long e4f4bb=206158430208;
    unsigned long e4d3bb=8864812498944;
    
    unsigned long d5e5bb=402653184;
    unsigned long d5c5bb=201326592;
    unsigned long d5e6bb=135266304;
    unsigned long e5f5bb=805306368;
    unsigned long e5d6bb=268959744;
    
    unsigned short int d4e4bonus=10; // d5e5
    unsigned short int d4c4bonus=5; // d5c5
    unsigned short int d4e3bonus=5; // d5e6
    unsigned short int e4f4bonus=5; // e5f5
    unsigned short int e4d3bonus=5; // e5d6
    
    if ((d4e4bb&pawns[0])==d4e4bb)        pawnScore0+=d4e4bonus;
    if ((d4c4bb&pawns[0])==d4c4bb)        pawnScore0+=d4c4bonus;
    if ((d4e3bb&pawns[0])==d4e3bb)        pawnScore0+=d4e3bonus;
    if ((e4f4bb&pawns[0])==e4f4bb)        pawnScore0+=e4f4bonus;
    if ((e4d3bb&pawns[0])==e4d3bb)        pawnScore0+=e4d3bonus;
    
    if ((d5e5bb&pawns[1])==d5e5bb)        pawnScore1+=d4e4bonus;
    if ((d5c5bb&pawns[1])==d5c5bb)        pawnScore1+=d4c4bonus;
    if ((d5e6bb&pawns[1])==d5e6bb)        pawnScore1+=d4e3bonus;
    if ((e5f5bb&pawns[1])==e5f5bb)        pawnScore1+=e4f4bonus;
    if ((e5d6bb&pawns[1])==e5d6bb)        pawnScore1+=e4d3bonus;    
    
    /****************
     * Passed Pawns *
     ****************/
    
    int passedPawnsBonus[2][64] = {{
          0,   0,   0,   0,   0,   0,   0,   0,
        100, 100, 100, 100, 100, 100, 100, 100,
         80,  80,  80,  80,  80,  80,  80,  80,
         60,  60,  60,  60,  60,  60,  60,  60,
         40,  40,  40,  40,  40,  40,  40,  40,
         20,  20,  20,  20,  20,  20,  20,  20,
         20,  20,  20,  20,  20,  20,  20,  20,
          0,   0,   0,   0,   0,   0,   0,   0
    },{
          0,   0,   0,   0,   0,   0,   0,   0,
         20,  20,  20,  20,  20,  20,  20,  20,
         20,  20,  20,  20,  20,  20,  20,  20,                  
         40,  40,  40,  40,  40,  40,  40,  40,
         60,  60,  60,  60,  60,  60,  60,  60,
         80,  80,  80,  80,  80,  80,  80,  80,
        100, 100, 100, 100, 100, 100, 100, 100,                  
          0,   0,   0,   0,   0,   0,   0,   0                  
    }};
    
    int passedPawnsProtectedBonus[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        125, 125, 125, 125, 125, 125, 125, 125,
        100, 100, 100, 100, 100, 100, 100, 100,
        75, 75, 75, 75, 75, 75, 75, 75,
        50, 50, 50, 50, 50, 50, 50, 50,
        25, 25, 25, 25, 25, 25, 25, 25,
        25, 25, 25, 25, 25, 25, 25, 25,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        25, 25, 25, 25, 25, 25, 25, 25,
        25, 25, 25, 25, 25, 25, 25, 25,
        50, 50, 50, 50, 50, 50, 50, 50,                
        75, 75, 75, 75, 75, 75, 75, 75,
        100, 100, 100, 100, 100, 100, 100, 100,
        125, 125, 125, 125, 125, 125, 125, 125,
        0, 0, 0, 0, 0, 0, 0, 0                
    }};
    
    unsigned long passedPawnsBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 4, 8, 16, 32, 64, 128,
        257, 514, 1028, 2056, 4112, 8224, 16448, 32896,                
        65793, 131586, 263172, 526344, 1052688, 2105376, 4210752, 8421504, 
        16843009, 33686018, 67372036, 134744072, 269488144, 538976288, 1077952576, 2155905152, 
        4311810305, 8623620610, 17247241220, 34494482440, 68988964880, 137977929760, 275955859520, 551911719040, 
        1103823438081, 2207646876162, 4415293752324, 8830587504648, 17661175009296, 35322350018592, 70644700037184, 141289400074368,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,                
        72340172838076416UL, 144680345676152832UL, 289360691352305664UL, 578721382704611328UL, 1157442765409222656UL, 2314885530818445312UL, 4629771061636890624UL, 9259542123273781248UL,
        72340172838010880UL, 144680345676021760UL, 289360691352043520UL, 578721382704087040UL, 1157442765408174080UL, 2314885530816348160UL, 4629771061632696320UL, 9259542123265392640UL,
        72340172821233664UL, 144680345642467328UL, 289360691284934656UL, 578721382569869312UL, 1157442765139738624UL, 2314885530279477248UL, 4629771060558954496UL, 9259542121117908992UL,
        72340168526266368UL, 144680337052532736UL, 289360674105065472UL, 578721348210130944UL, 1157442696420261888UL, 2314885392840523776UL, 4629770785681047552UL, 9259541571362095104UL,
        72339069014638592UL, 144678138029277184UL, 289356276058554368UL, 578712552117108736UL, 1157425104234217472UL, 2314850208468434944UL, 4629700416936869888UL, 9259400833873739776UL,
        72057594037927936UL, 144115188075855872UL, 288230376151711744UL, 576460752303423488UL, 1152921504606846976UL, 2305843009213693952UL, 4611686018427387904UL, 9223372036854775808UL,
        0, 0, 0, 0, 0, 0, 0, 0
    }};

    unsigned long passedPawnsBitboardsAttacks[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0,
        514,1285, 2570, 5140, 10280, 20560, 41120, 16448,
        131586, 328965, 657930, 1315860, 2631720, 5263440, 10526880, 4210752,
        33686018, 84215045, 168430090, 336860180, 673720360, 1347440720, 2694881440, 1077952576,
        8623620610, 21559051525, 43118103050, 86236206100, 172472412200, 344944824400, 689889648800, 275955859520,
        2207646876162, 5519117190405, 11038234380810, 22076468761620, 44152937523240, 88305875046480, 176611750092960, 70644700037184,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        144680345676152832UL, 361700864190382080UL, 723401728380764160UL, 1446803456761528320UL, 2893606913523056640UL, 5787213827046113280UL, 11574427654092226560UL, 4629771061636890624UL,
        144680345676021760UL, 361700864190054400UL, 723401728380108800UL, 1446803456760217600UL, 2893606913520435200UL, 5787213827040870400UL, 11574427654081740800UL, 4629771061632696320UL,
        144680345642467328UL, 361700864106168320UL, 723401728212336640UL, 1446803456424673280UL, 2893606912849346560UL, 5787213825698693120UL, 11574427651397386240UL, 4629771060558954496UL,
        144680337052532736UL, 361700842631331840UL, 723401685262663680UL, 1446803370525327360UL, 2893606741050654720UL, 5787213482101309440UL, 11574426964202618880UL, 4629770785681047552UL,
        144678138029277184UL, 361695345073192960UL, 723390690146385920UL, 1446781380292771840UL, 2893562760585543680UL, 5787125521171087360UL, 11574251042342174720UL, 4629700416936869888UL,
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0
    }};

    unsigned long passedPawnsProtectedBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        131584, 328960, 657920, 1315840, 2631680, 5263360, 10526720, 4210688,
        33685504, 84213760, 168427520, 336855040, 673710080, 1347420160, 2694840320, 1077936128,
        8623489024, 21558722560, 43117445120, 86234890240, 172469780480, 344939560960, 689879121920, 275951648768,
        2207613190144, 5519032975360, 11038065950720, 22076131901440, 44152263802880, 88304527605760, 176609055211520, 70643622084608,
        565148976676864, 1412872441692160, 2825744883384320, 5651489766768640, 11302979533537280, 22605959067074560, 45211918134149120, 18084767253659648,
        562949953421312, 1407374883553280, 2814749767106560, 5629499534213120, 11258999068426240, 22517998136852480, 45035996273704960, 18014398509481984,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        512, 1280, 2560, 5120, 10240, 20480, 40960, 16384,
        131584, 328960, 657920, 1315840, 2631680, 5263360, 10526720, 4210688,
        33685504, 84213760, 168427520, 336855040, 673710080, 1347420160, 2694840320, 1077936128,
        8623489024, 21558722560, 43117445120, 86234890240, 172469780480, 344939560960, 689879121920, 275951648768,
        2207613190144, 5519032975360, 11038065950720, 22076131901440, 44152263802880, 88304527605760, 176609055211520, 70643622084608,
        565148976676864, 1412872441692160, 2825744883384320, 5651489766768640, 11302979533537280, 22605959067074560, 45211918134149120, 18084767253659648,
        0, 0, 0, 0, 0, 0, 0, 0
    }};
       
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (passedPawnsBitboards[0][x]&pawns[0])==0 && (passedPawnsBitboards[0][x]&pawns[1])==0 // Felder vor Bauern sind frei
                && (passedPawnsBitboardsAttacks[0][x]&pawns[1])==0 // Felder vor Bauern werden nicht vom Gegner durch Bauern bedroht
                ) {
            
            if ((passedPawnsProtectedBitboards[0][x]&pawns[0])!=0)
                pawnScore0+=passedPawnsProtectedBonus[0][x];
            else
                pawnScore0+=passedPawnsBonus[0][x];
        
        }
        
        bitboard-=Position::BIN[x];
    }
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (passedPawnsBitboards[1][x]&pawns[0])==0 && (passedPawnsBitboards[1][x]&pawns[1])==0 // Felder vor Bauern sind frei
                && (passedPawnsBitboardsAttacks[1][x]&pawns[0])==0 // Felder vor Bauern werden nicht vom Gegner durch Bauern bedroht                
                ) {
            
            if ((passedPawnsProtectedBitboards[1][x]&pawns[1])!=0)
                pawnScore1+=passedPawnsProtectedBonus[1][x];
            else
                pawnScore1+=passedPawnsBonus[1][x];
            
        }
        
        bitboard-=Position::BIN[x];
    }    
    
    /**************
     * Weak Pawns *
     **************/
    
    int weakPawnsPenalty[2][64] = {{
          0,    0,   0,   0,    0,   0,   0,   0,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
           0,   0,   0,    0,   0,   0,   0,   0
    },{
           0,   0,   0,    0,   0,   0,   0,   0,
          -8, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
          0,    0,   0,   0,    0,   0,   0,   0,                   
    }};
    
    unsigned long weakPawnsBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        144680345676152832UL, 361700864190382080UL, 723401728380764160UL, 1446803456761528320UL, 2893606913523056640UL, 5787213827046113280UL, 11574427654092226560UL, 4629771061636890624UL,
        144680345676021760UL, 361700864190054400UL, 723401728380108800UL, 1446803456760217600UL, 2893606913520435200UL, 5787213827040870400UL, 11574427654081740800UL, 4629771061632696320UL,
        144680345642467328UL, 361700864106168320UL, 723401728212336640UL, 1446803456424673280UL, 2893606912849346560UL, 5787213825698693120UL, 11574427651397386240UL, 4629771060558954496UL,
        144680337052532736UL, 361700842631331840UL, 723401685262663680UL, 1446803370525327360UL, 2893606741050654720UL, 5787213482101309440UL, 11574426964202618880UL, 4629770785681047552UL,
        144678138029277184UL, 361695345073192960UL, 723390690146385920UL, 1446781380292771840UL, 2893562760585543680UL, 5787125521171087360UL, 11574251042342174720UL, 4629700416936869888UL,
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0,
        514,1285, 2570, 5140, 10280, 20560, 41120, 16448,
        131586, 328965, 657930, 1315860, 2631720, 5263440, 10526880, 4210752,
        33686018, 84215045, 168430090, 336860180, 673720360, 1347440720, 2694881440, 1077952576,
        8623620610, 21559051525, 43118103050, 86236206100, 172472412200, 344944824400, 689889648800, 275955859520,
        2207646876162, 5519117190405, 11038234380810, 22076468761620, 44152937523240, 88305875046480, 176611750092960, 70644700037184,
        0, 0, 0, 0, 0, 0, 0, 0
    }};
    
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (weakPawnsBitboards[0][x]&pawns[0])==0 )
                pawnScore0+=weakPawnsPenalty[0][x];
        
        bitboard-=Position::BIN[x];
    }    
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (weakPawnsBitboards[1][x]&pawns[1])==0 )
                pawnScore1+=weakPawnsPenalty[1][x];
        
        bitboard-=Position::BIN[x];
    }    
        
    
    
    
    
    
    
    
    
    
    
    /********************
     * Piece Evaluation *
     ********************/
    
    int pieceScore0=0, pieceScore1=0;
    
    /************
     * Springer *
     ************/
    
    int knightA8Penalty=150;
    int knightH8Penalty=150;
    int knightA7Penalty=100;
    int knightH7Penalty=100;
    int knightC3Penalty=5;
    
    // A8
    if ((knights[0]&BIN[0])!=0) {
        if ((pawns[1]&BIN[8])!=0 || (pawns[1]&BIN[10])!=0)
            pieceScore0-=knightA8Penalty;
    }
    if ((knights[1]&BIN[56])!=0) {
        if ((pawns[0]&BIN[48])!=0 || (pawns[0]&BIN[50])!=0)
            pieceScore1-=knightA8Penalty;
    }
    
    // H8
    if ((knights[0]&BIN[7])!=0) {
        if ((pawns[1]&BIN[13])!=0 || (pawns[1]&BIN[15])!=0)
            pieceScore0-=knightH8Penalty;
    }
    if ((knights[1]&BIN[63])!=0) {
        if ((pawns[0]&BIN[55])!=0 || (pawns[0]&BIN[53])!=0)
            pieceScore1-=knightH8Penalty;
    }
    
    // A7
    if ((knights[0]&BIN[8])!=0) {
        if ((pawns[1]&BIN[16])!=0 && (pawns[1]&BIN[9])!=0)
            pieceScore0-=knightA7Penalty;
    }
    if ((knights[1]&BIN[48])!=0) {
        if ((pawns[0]&BIN[40])!=0 && (pawns[0]&BIN[49])!=0)
            pieceScore1-=knightA7Penalty;
    }
    
    // H7
    if ((knights[0]&BIN[15])!=0) {
        if ((pawns[1]&BIN[23])!=0 && (pawns[1]&BIN[14])!=0)
            pieceScore0-=knightH7Penalty;
    }
    if ((knights[1]&BIN[55])!=0) {
        if ((pawns[0]&BIN[47])!=0 && (pawns[0]&BIN[54])!=0)
            pieceScore1-=knightH7Penalty;
    }
    
    // C3
    if ((knights[0]&BIN[42])!=0) {
        if ((pawns[0]&BIN[50])!=0 && (pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[36])==0)
            pieceScore0-=knightC3Penalty;            
    }
    if ((knights[1]&BIN[18])!=0) {
        if ((pawns[1]&BIN[10])!=0 && (pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[28])==0)
            pieceScore1-=knightC3Penalty;
    }
        
    /**********
     * Läufer *
     **********/
    
    int bishopA7Penalty=150;
    int bishopH7Penalty=150;
    int bishopB8Penalty=150;
    int bishopG8Penalty=150;
    int bishopA6Penalty=50;
    int bishopH6Penalty=50;
    int bishopF1Bonus=20;
    int bishopC1Bonus=20;

    // A7
    if ((rooks[0]&BIN[8])!=0) {
        if ((pawns[1]&BIN[17])!=0)
            pieceScore0-=bishopA7Penalty;
    }
    if ((rooks[1]&BIN[48])!=0) {
        if ((pawns[0]&BIN[41])!=0)
            pieceScore1-=bishopA7Penalty;
    }
    
    // H7
    if ((rooks[0]&BIN[15])!=0) {
        if ((pawns[1]&BIN[22])!=0)
            pieceScore0-=bishopH7Penalty;
    }
    if ((rooks[1]&BIN[55])!=0) {
        if ((pawns[0]&BIN[46])!=0)
            pieceScore1-=bishopH7Penalty;
    }
    
    // B8
    if ((rooks[0]&BIN[1])!=0) {
        if ((pawns[1]&BIN[10])!=0)
            pieceScore0-=bishopB8Penalty;
    }
    if ((rooks[1]&BIN[57])!=0) {
        if ((pawns[0]&BIN[50])!=0)
            pieceScore1-=bishopB8Penalty;
    }
    
    // G8
    if ((rooks[0]&BIN[6])!=0) {
        if ((pawns[1]&BIN[13])!=0)
            pieceScore0-=bishopG8Penalty;
    }
    if ((rooks[1]&BIN[62])!=0) {
        if ((pawns[0]&BIN[53])!=0)
            pieceScore1-=bishopG8Penalty;
    }
    
    // A6
    if ((rooks[0]&BIN[16])!=0) {
        if ((pawns[1]&BIN[25])!=0)
            pieceScore0-=bishopA6Penalty;
    }
    if ((rooks[1]&BIN[40])!=0) {
        if ((pawns[0]&BIN[33])!=0)
            pieceScore1-=bishopA6Penalty;
    }
    
    // H6
    if ((rooks[0]&BIN[23])!=0) {
        if ((pawns[1]&BIN[30])!=0)
            pieceScore0-=bishopH6Penalty;
    }
    if ((rooks[1]&BIN[47])!=0) {
        if ((pawns[0]&BIN[38])!=0)
            pieceScore1-=bishopH6Penalty;
    }
    
    // F1
    if ((rooks[0]&BIN[61])!=0) {
        if ((kings[0]&BIN[62])!=0)
            pieceScore0+=bishopF1Bonus;
    }
    if ((rooks[1]&BIN[5])!=0) {
        if ((kings[1]&BIN[6])!=0)
            pieceScore1+=bishopF1Bonus;
    }
    
    // C1
    if ((rooks[0]&BIN[58])!=0) {
        if ((kings[0]&BIN[57])!=0)
            pieceScore0+=bishopC1Bonus;
    }
    if ((rooks[1]&BIN[2])!=0) {
        if ((kings[1]&BIN[1])!=0)
            pieceScore1+=bishopC1Bonus;
    }    
        
    /********************************************
     * Turm                                     *
     * - Bonus für offene und halboffene Linien *
     ********************************************/

    int rookOpenBonus=10;
    int rookHalfOpenBonus=5;
    
    for (unsigned short int i=0;i<8;i++) {
        if ((rooks[0]&cols[i])!=0 && (cols[i]&pawns[0])==0) {
            if ((cols[i]&pawns[1])==0)
                pieceScore0+=rookOpenBonus;
            else
                pieceScore0+=rookHalfOpenBonus;
        }
    }

    for (unsigned short int i=0;i<8;i++) {
        if ((rooks[1]&cols[i])!=0 && (cols[i]&pawns[1])==0) {
            if ((cols[i]&pawns[0])==0)
                pieceScore1+=rookOpenBonus;
            else
                pieceScore1+=rookHalfOpenBonus;
        }
    }    

    /**************************************
     * Dame                               *
     * - Penalty für zu frühe Entwicklung *
     **************************************/

    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if ((BIN[x]&281474976710655)!=0) {
            if ((knights[0]&BIN[57])!=0)  pieceScore0-=2;
            if ((knights[0]&BIN[62])!=0)  pieceScore0-=2;
            if ((bishops[0]&BIN[58])!=0)  pieceScore0-=2;
            if ((bishops[0]&BIN[61])!=0)  pieceScore0-=2;
        }

        bitboard-=Position::BIN[x];
    }
    
    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if ((BIN[x]&18446744073709486080UL)!=0) {
            if ((knights[1]&BIN[1])!=0)  pieceScore1-=2;
            if ((knights[1]&BIN[6])!=0)  pieceScore1-=2;
            if ((bishops[1]&BIN[2])!=0)  pieceScore1-=2;
            if ((bishops[1]&BIN[5])!=0)  pieceScore1-=2;
        }

        bitboard-=Position::BIN[x];
    }    
    
    
    /*********
     * König *
     *********/
    
    int shield1Bonus=10;
    int shield2Bonus=5;
    int noShieldPenalty=10;

    unsigned short int kCol;    
    
    /****************
     * Weißer König *
     ****************/
    
    for (kCol=0;kCol<8;kCol++) {
        if ((kings[0]&cols[kCol])!=0)
            break;
    }
    
    // king side
    if (kCol>4) {
        
        if ((pawns[0]&BIN[53])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[45])!=0)   pieceScore0+=shield2Bonus;
        
        if ((pawns[0]&BIN[54])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[46])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[6])==0)  pieceScore0-=noShieldPenalty;
        
        if ((pawns[0]&BIN[55])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[47])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[7])==0)  pieceScore0+=noShieldPenalty;
        
    }
    
    // queenside
    if (kCol<3) {
        
        if ((pawns[0]&BIN[48])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[40])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[0])==0)  pieceScore0-=noShieldPenalty;
        
        if ((pawns[0]&BIN[49])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[41])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[1])==0)  pieceScore0+=noShieldPenalty;        

        if ((pawns[0]&BIN[50])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[42])!=0)   pieceScore0+=shield2Bonus;        
        
    }
    
    /*******************
     * Schwarzer König *
     *******************/
    
    for (kCol=0;kCol<8;kCol++) {
        if ((kings[1]&cols[kCol])!=0)
            break;
    }
    
    // king side
    if (kCol>4) {
        
        if ((pawns[1]&BIN[13])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[21])!=0)   pieceScore1+=shield2Bonus;
        
        if ((pawns[1]&BIN[14])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[22])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[6])==0)  pieceScore1-=noShieldPenalty;
        
        if ((pawns[1]&BIN[15])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[23])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[7])==0)  pieceScore1+=noShieldPenalty;
        
    }
    
    // queenside
    if (kCol<3) {
        
        if ((pawns[1]&BIN[8])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[16])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[0])==0)  pieceScore1-=noShieldPenalty;
        
        if ((pawns[1]&BIN[9])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[17])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[1])==0)  pieceScore1+=noShieldPenalty;        

        if ((pawns[1]&BIN[10])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[18])!=0)   pieceScore1+=shield2Bonus;
        
    }

    /**************************
     * Diverse Pattern-Matches *
     **************************/
    
    int patternScore0=0, patternScore1=0;
    
    /**********************************
     * Blockierte Bauern in der Mitte *
     **********************************/
    
    int centerPawnPenalty=24;
    
    // D2
    if ((pawns[0]&BIN[51])!=0 && (all&BIN[43])!=0)
        patternScore0-=centerPawnPenalty;
    
    // E2
    if ((pawns[0]&BIN[52])!=0 && (all&BIN[44])!=0)
        patternScore0-=centerPawnPenalty;
    
    // D7
    if ((pawns[1]&BIN[11])!=0 && (all&BIN[19])!=0)
        patternScore1-=centerPawnPenalty;
    
    // E7
    if ((pawns[1]&BIN[12])!=0 && (all&BIN[20])!=0)
        patternScore1-=centerPawnPenalty;
    
    /********************
     * Blockierte Türme *
     ********************/
    
    int kingBlockRookPenalty=24;
    
    if(         ((kings[0]&BIN[61])!=0 || (kings[0]&BIN[62])!=0)
            &&  ((rooks[0]&BIN[63])!=0 || (rooks[0]&BIN[62])!=0))
        patternScore0-=kingBlockRookPenalty;
    
    if(         ((kings[0]&BIN[58])!=0 || (kings[0]&BIN[57])!=0)
            &&  ((rooks[0]&BIN[56])!=0 || (rooks[0]&BIN[57])!=0))
        patternScore0-=kingBlockRookPenalty;
    
    if(         ((kings[1]&BIN[5])!=0 || (kings[1]&BIN[6])!=0)
            &&  ((rooks[1]&BIN[7])!=0 || (rooks[1]&BIN[6])!=0))
        patternScore1-=kingBlockRookPenalty;
    
    if(         ((kings[1]&BIN[2])!=0 || (kings[1]&BIN[1])!=0)
            &&  ((rooks[1]&BIN[0])!=0 || (rooks[1]&BIN[1])!=0))
        patternScore1-=kingBlockRookPenalty;    
    
    /****************
     * Slave Fehler *
     ****************/
    
    int slaveMistakePenalty=10;
    
    if ((pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[26])!=0 &&
            (pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[18])!=0)
        patternScore0-=slaveMistakePenalty;
    
    if ((pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[34])!=0 &&
            (pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[42])!=0)
        patternScore1-=slaveMistakePenalty;
    
    /**************
     * Fianchetto *
     **************/
    
    int fianchettoBonus=4;
    int fianchettoPenalty=4;
    
    if ((pawns[0]&BIN[46])!=0) {
        if ((bishops[0]&BIN[54])!=0)
            patternScore0+=fianchettoBonus;
        else {
            if (    (bishops[0]&BIN[45])==0 && 
                    (bishops[0]&BIN[47])==0 && 
                    (bishops[0]&BIN[63])==0)
                patternScore0-=fianchettoPenalty;
        }
    }

    if ((pawns[0]&BIN[41])!=0) {
        if ((bishops[0]&BIN[49])!=0)
            patternScore0+=fianchettoBonus;
        else {
            if (    (bishops[0]&BIN[42])==0 && 
                    (bishops[0]&BIN[40])==0 && 
                    (bishops[0]&BIN[56])==0)
                patternScore0-=fianchettoPenalty;
        }
    }
    
    if ((pawns[1]&BIN[22])!=0) {
        if ((bishops[1]&BIN[14])!=0)
            patternScore1+=fianchettoBonus;
        else {
            if (    (bishops[1]&BIN[21])==0 && 
                    (bishops[1]&BIN[23])==0 && 
                    (bishops[1]&BIN[7])==0)
                patternScore1-=fianchettoPenalty;
        }
    }
    
    if ((pawns[1]&BIN[17])!=0) {
        if ((bishops[1]&BIN[9])!=0)
            patternScore1+=fianchettoBonus;
        else {
            if (    (bishops[1]&BIN[18])==0 && 
                    (bishops[1]&BIN[16])==0 && 
                    (bishops[1]&BIN[0])==0)
                patternScore1-=fianchettoPenalty;
        }
    }    
    
    /************************************************************************
     * berechne targetmaps und mobility score                               *
     * targetmaps werden nachfolgend zb. für den capture score benötigt     *
     * zur berechnung des mobility score wird die gesamte zugliste benötigt *
     * daher werden diese beiden elemente hier zusammen berechnet           *
     * **********************************************************************/

//    long targetmap0, targetmap1;
//    int mobilityScore0 = 0, mobilityScore1 = 0;
//    int moveSize0, moveSize1;
//    
//    /***************
//     * weiß am zug *
//     * *************/
//
//    onMove=false;
////    targetmap=0L;
//    calc_moves(false); // true=with castling
//    moveSize0=mvSize[cd];
////    targetmap0=targetmap;
//    
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[0])!=0L) mobilityScore0+=20;
//        // if ((mvD[cd][i][0]&queens[0])!=0L) mobilityScore0+=9;
//        if ((mvD[cd][i][0]&rooks[0])!=0L) mobilityScore0+=5;
//        else if ((mvD[cd][i][0]&bishops[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&knights[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&pawns[0])!=0L) mobilityScore0+=1;
//    }
//
//    /******************
//     * schwarz am zug *
//     * ****************/    
//    
//    onMove=true;
////    targetmap=0L;
//    calc_moves(false); // true=with castling
//    moveSize1=mvSize[cd];
////    targetmap1=targetmap;    
//
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[1])!=0L) mobilityScore1+=20;
//        // if ((mvD[cd][i][0]&queens[1])!=0L) mobilityScore1+=9;
//        if ((mvD[cd][i][0]&rooks[1])!=0L) mobilityScore1+=5;
//        else if ((mvD[cd][i][0]&bishops[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&knights[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&pawns[1])!=0L) mobilityScore1+=1;
//    }    
//
//     mobilityScore0/=moveSize0;
//     mobilityScore1/=moveSize1;
        
    
    /*************************
     * Anzahl möglicher Züge *
     *************************/
    
//    int legalMoves0, legalMoves1;
//    int legalMovesWeight=70;
    
//    legalMoves0 = popCount(targetmap0);
//    legalMoves1 = popCount(targetmap1);
    
    
    /****************************************************************
     * berechne capture score                                       *
     * in abhängigkeit der figur-typen die geschlagen werden können *
     * **************************************************************/    
    
    // sollte captureScore bei idealer quiescence-suche nicht 0 sein?
    
//    int captureScore0 = 0, captureScore1 = 0;
//    int captureWeight = 80;
    
//    if ((targetmap0&kings[1])!=0)       captureScore0+=20;
//    if ((targetmap0&queens[1])!=0)      captureScore0+=9*popCount(targetmap0&queens[1]);
//    if ((targetmap0&rooks[1])!=0)       captureScore0+=5*popCount(targetmap0&rooks[1]);
//    if ((targetmap0&bishops[1])!=0)     captureScore0+=3*popCount(targetmap0&bishops[1]);
//    if ((targetmap0&knights[1])!=0)     captureScore0+=3*popCount(targetmap0&knights[1]);
//    if ((targetmap0&pawns[1])!=0)       captureScore0+=popCount(targetmap0&pawns[1]);
//
//    if ((targetmap1&kings[0])!=0)       captureScore1+=20;
//    if ((targetmap1&queens[0])!=0)      captureScore1+=9*popCount(targetmap1&queens[0]);
//    if ((targetmap1&rooks[0])!=0)       captureScore1+=5*popCount(targetmap1&rooks[0]);
//    if ((targetmap1&bishops[0])!=0)     captureScore1+=3*popCount(targetmap1&bishops[0]);
//    if ((targetmap1&knights[0])!=0)     captureScore1+=3*popCount(targetmap1&knights[0]);
//    if ((targetmap1&pawns[0])!=0)       captureScore1+=popCount(targetmap1&pawns[0]);    
//    
        
    
    
    
    /*********************************************************************
     * Piece-Square Tabellen                                             *
     * Berechnet Punkt-Gewinn/Abzug für Figuren an bestimmten Positionen *
     *********************************************************************/
    
    int pieceSquare0=0, pieceSquare1=0;
    
    int pawnPieceSquare[2][64] = {{        
         0,  0,  0,  0,  0,  0,  0,  0,
        50, 50, 50, 50, 50, 50, 50, 50,
        10, 10, 20, 30, 30, 20, 10, 10,
         5,  5, 10, 25, 25, 10,  5,  5,
         0,  0,  0, 20, 20,  0,  0,  0,
         5, -5,-10,  0,  0,-10, -5,  5,
         5, 10, 10,-20,-20, 10, 10,  5,
         0,  0,  0,  0,  0,  0,  0,  0        
    },{
         0,  0,  0,  0,  0,  0,  0,  0,        
         5, 10, 10,-20,-20, 10, 10,  5,
         5, -5,-10,  0,  0,-10, -5,  5,         
         0,  0,  0, 20, 20,  0,  0,  0,
         5,  5, 10, 25, 25, 10,  5,  5,
        10, 10, 20, 30, 30, 20, 10, 10,         
        50, 50, 50, 50, 50, 50, 50, 50,         
         0,  0,  0,  0,  0,  0,  0,  0         
    }};

    int knightPieceSquare[2][64] = {{
        -50,-40,-30,-30,-30,-30,-40,-50,
        -40,-20,  0,  0,  0,  0,-20,-40,
        -30,  0, 10, 15, 15, 10,  0,-30,
        -30,  5, 15, 20, 20, 15,  5,-30,
        -30,  0, 15, 20, 20, 15,  0,-30,
        -30,  5, 10, 15, 15, 10,  5,-30,
        -40,-20,  0,  5,  5,  0,-20,-40,
        -50,-40,-30,-30,-30,-30,-40,-50        
    },{
        -50,-40,-30,-30,-30,-30,-40,-50,
        -40,-20,  0,  5,  5,  0,-20,-40,
        -30,  5, 10, 15, 15, 10,  5,-30,
        -30,  0, 15, 20, 20, 15,  0,-30,
        -30,  5, 15, 20, 20, 15,  5,-30,
        -30,  0, 10, 15, 15, 10,  0,-30,
        -40,-20,  0,  0,  0,  0,-20,-40,
        -50,-40,-30,-30,-30,-30,-40,-50        
    }};

    int bishopPieceSquare[2][64] = {{
        -20,-10,-10,-10,-10,-10,-10,-20,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -10,  0,  5, 10, 10,  5,  0,-10,
        -10,  5,  5, 10, 10,  5,  5,-10,
        -10,  0, 10, 10, 10, 10,  0,-10,
        -10, 10, 10, 10, 10, 10, 10,-10,
        -10,  5,  0,  0,  0,  0,  5,-10,
        -20,-10,-10,-10,-10,-10,-10,-20        
    },{
        -20,-10,-10,-10,-10,-10,-10,-20,
        -10,  5,  0,  0,  0,  0,  5,-10,
        -10, 10, 10, 10, 10, 10, 10,-10,
        -10,  0, 10, 10, 10, 10,  0,-10,
        -10,  5,  5, 10, 10,  5,  5,-10,
        -10,  0,  5, 10, 10,  5,  0,-10,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -20,-10,-10,-10,-10,-10,-10,-20        
    }};
    
    int rookPieceSquare[2][64] = {{
          0,  0,  0,  0,  0,  0,  0,  0,
          5, 10, 10, 10, 10, 10, 10,  5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
          0,  0,  0,  5,  5,  0,  0,  0                
    },{
          0,  0,  0,  5,  5,  0,  0,  0        
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
          5, 10, 10, 10, 10, 10, 10,  5,
          0,  0,  0,  0,  0,  0,  0,  0,
    }}; 
    
    int queenPieceSquare[2][64] = {{    
        -20,-10,-10, -5, -5,-10,-10,-20,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -10,  0,  5,  5,  5,  5,  0,-10,
         -5,  0,  5,  5,  5,  5,  0, -5,
          0,  0,  5,  5,  5,  5,  0, -5,
        -10,  5,  5,  5,  5,  5,  0,-10,
        -10,  0,  5,  0,  0,  0,  0,-10,
        -20,-10,-10, -5, -5,-10,-10,-20        
    },{
        -20,-10,-10, -5, -5,-10,-10,-20,
        -10,  0,  5,  0,  0,  0,  0,-10,
        -10,  5,  5,  5,  5,  5,  0,-10,
          0,  0,  5,  5,  5,  5,  0, -5,
         -5,  0,  5,  5,  5,  5,  0, -5,        
        -10,  0,  5,  5,  5,  5,  0,-10,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -20,-10,-10, -5, -5,-10,-10,-20        
    }};
    
    int kingPieceSquare[2][64] = {{
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -20,-30,-30,-40,-40,-30,-30,-20,
        -10,-20,-20,-20,-20,-20,-20,-10,
         20, 20,  0,  0,  0,  0, 20, 20,
         20, 30, 10,  0,  0, 10, 30, 20        
    },{
         20, 30, 10,  0,  0, 10, 30, 20,
         20, 20,  0,  0,  0,  0, 20, 20,         
        -10,-20,-20,-20,-20,-20,-20,-10,
        -20,-30,-30,-40,-40,-30,-30,-20,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30         
    }};
    
    int kingPieceSquareEndgame[2][64] = {{
        -50,-40,-30,-20,-20,-30,-40,-50,
        -30,-20,-10,  0,  0,-10,-20,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-30,  0,  0,  0,  0,-30,-30,
        -50,-30,-30,-30,-30,-30,-30,-50
    },{
        -50,-30,-30,-30,-30,-30,-30,-50,
        -30,-30,  0,  0,  0,  0,-30,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-20,-10,  0,  0,-10,-20,-30,
        -50,-40,-30,-20,-20,-30,-40,-50                
    }};
    
    /********
     * weiß *
     ********/
    
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=pawnPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=knightPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=bishopPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = rooks[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=rookPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=queenPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }    

    if (isEndgame()) {
        bitboard = kings[0];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare0+=kingPieceSquareEndgame[0][x];
            bitboard-=Position::BIN[x];
        }
    } else {
        bitboard = kings[0];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare0+=kingPieceSquare[0][x];
            bitboard-=Position::BIN[x];
        }
    }
    
    /***********
     * schwarz *
     ***********/
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=pawnPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=knightPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=bishopPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = rooks[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=rookPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=queenPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }        

    if (isEndgame()) {
        bitboard = kings[1];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare1+=kingPieceSquareEndgame[1][x];
            bitboard-=Position::BIN[x];
        }
    } else {
        bitboard = kings[1];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare1+=kingPieceSquare[1][x];
            bitboard-=Position::BIN[x];
        }        
    }
    
    /*****************************************************************************************************************
     * berechne gesamt score                                                                                         *
     * die gewichtung der einzelnen werte sollte später automatisch durch ein skript ermittelt werden.               *
     * mit entsprechend vielen tournaments sollte so eine verbesserung bzw. verschlechterung der ELO erkennbar sein. *
     * die dadurch gewonnen daten können auch für weitere analysen verwendet werden um so zu ermitteln,              *
     * ob es werte gibt, die eigentlich gar nicht zwingend berechnet werden müssten, da sie ohnehin                  *
     * mit anderen werten korrelieren                                                                                *
     *****************************************************************************************************************/    
    
    int score = (materialScore0 - materialScore1) +
//                (mobilityScore0 - mobilityScore1) +
//                (kingTropism0 - kingTropism1) +
//                (patternScore0 - patternScore1) +
//                (pawnScore0 - pawnScore1);
//                (pieceScore0 - pieceScore1) +
                (pieceSquare0 - pieceSquare1);
                                                   
   // cout << (legalMoves0 - legalMoves1) * legalMovesWeight << endl;
    
    /******************
     * we're ready :) *
     * ****************/        
    
    if (side2move)
        score*=-1;
    
//    evalstat.logFenScore(getFen(),      score, 
//                                        (materialScore0 - materialScore1),
//                                        materialScore0, materialScore1,
//                                        (mobilityScore0 - mobilityScore1) * mobilityWeight,                                        
//                                        mobilityScore0, mobilityScore1,
//                                        (legalMoves0 - legalMoves1) * legalMovesWeight,
//                                        legalMoves0, legalMoves1,
//                                        (captureScore0 - captureScore1) * captureWeight,
//                                        captureScore0, captureScore1,
//                                        (pieceSquare0 - pieceSquare1) * pieceSquareWeight,
//                                        pieceSquare0, pieceSquare1);
        
    return score;
}

/************************************
 * Speichert die aktuelle Position. *
 ************************************/
void Position::doBackup() {
    myBackup[cd].ep=ep;
    myBackup[cd].castlingRules=castlingRules;
}

/************************************
 * Lädt eine gespeicherte Position. *
 ************************************/
void Position::undoMove(const unsigned long &from, const unsigned long &to, const int &promotion, const unsigned short int &capture) {

    // richtigen aktuellen/alten hashkey setzen
//    hashKey = hashTable.getHashKey(this);    

    // castling-rules aus hashKey entfernen
    short int cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];                    
    
    bool notOnMove=onMove;
    onMove=!onMove;
    
    hashKey^=hashTable.side2move;
    
    // backmove figures
    if ((pawns[onMove]&to)!=0L) {
        pawns[onMove]^=(from|to);
        
        hashKey^=hashTable.phash[onMove][bitScanForward(from)];
        hashKey^=hashTable.phash[onMove][bitScanForward(to)];
        
        // ep-move
        if (capture==CAPTURE_NONE) {
            if (to==from>>9) { // kann eigentlich zu if-elses gemacht werden
                pawns[notOnMove]^=(from>>1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
            }
            if (to==from>>7) {
                pawns[notOnMove]^=(from<<1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
            }
            if (to==from<<9) {
                pawns[notOnMove]^=(from<<1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
            }
            if (to==from<<7) {
                pawns[notOnMove]^=(from>>1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
            }
        }
    } else if ((knights[onMove]&to)!=0L) {
        knights[onMove]^=(from|to);
        hashKey^=hashTable.nhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.nhash[onMove][bitScanForward(to)];        
    } else if ((bishops[onMove]&to)!=0L) {
        bishops[onMove]^=(from|to);
        hashKey^=hashTable.bhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.bhash[onMove][bitScanForward(to)];        
    } else if ((rooks[onMove]&to)!=0L) {
        rooks[onMove]^=(from|to);
        hashKey^=hashTable.rhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.rhash[onMove][bitScanForward(to)];        
    } else if ((queens[onMove]&to)!=0L) {
        queens[onMove]^=(from|to);
        hashKey^=hashTable.qhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.qhash[onMove][bitScanForward(to)];        
    } else if ((kings[onMove]&to)!=0L) {
        if ((from&1152921504606846992UL)!=0L && (to&4899916394579099716UL)!=0L) {
            if (to==BIN[2]) {
                rochade(from,to,BIN[0],BIN[3]);
                hashKey^=hashTable.rhash[onMove][0];
                hashKey^=hashTable.rhash[onMove][3];            
            } else if (to==BIN[6]) {
                rochade(from,to,BIN[7],BIN[5]);
                hashKey^=hashTable.rhash[onMove][7];
                hashKey^=hashTable.rhash[onMove][5];                            
            } else if (to==BIN[58]) {
                rochade(from,to,BIN[56],BIN[59]);
                hashKey^=hashTable.rhash[onMove][56];
                hashKey^=hashTable.rhash[onMove][59];                            
            } else if (to==BIN[62]) {
                rochade(from,to,BIN[63],BIN[61]);
                hashKey^=hashTable.rhash[onMove][63];
                hashKey^=hashTable.rhash[onMove][61];
            }
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];            
        } else {
            kings[onMove]^=(from|to);            
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];            
        }
    }
    
    // captures
    if ((capture&CAPTURE_MASK)==CAPTURE_PAWN) {
        pawns[notOnMove]|=to;
        hashKey^=hashTable.phash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_KNIGHT) {
        knights[notOnMove]|=to;
        hashKey^=hashTable.nhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_BISHOP) {
        bishops[notOnMove]|=to;
        hashKey^=hashTable.bhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_ROOK) {
        rooks[notOnMove]|=to;
        hashKey^=hashTable.rhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_QUEEN) {
        queens[notOnMove]|=to;
        hashKey^=hashTable.qhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_KING) {
        kings[notOnMove]|=to;
        hashKey^=hashTable.khash[notOnMove][bitScanForward(to)];
    }
    
    // recover promotion
    if ((capture&32768)!=0) { // isPromotionMask gesetzt?
        pawns[onMove]^=from;
        hashKey^=hashTable.phash[onMove][bitScanForward(from)];
        if (promotion==0) {
            queens[onMove]^=from;
            hashKey^=hashTable.qhash[onMove][bitScanForward(from)];
        } else if (promotion==1) {
            rooks[onMove]^=from;
            hashKey^=hashTable.rhash[onMove][bitScanForward(from)];
        } else if (promotion==2) {
            bishops[onMove]^=from;
            hashKey^=hashTable.bhash[onMove][bitScanForward(from)];
        } else if (promotion==3) {
            knights[onMove]^=from;
            hashKey^=hashTable.nhash[onMove][bitScanForward(from)];
        }
    }

    
    if (ep>0)
        hashKey^=hashTable.files[ep%8];    
    
    // recover castling-rules & ep-flag
    ep=myBackup[cd].ep;
    castlingRules=myBackup[cd].castlingRules;        
    updateBitboardLongs();
    
    if (ep>0)
        hashKey^=hashTable.files[ep%8];    
    

    
    cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];            
    
    
    
//    if (hashKey!=hashTable.getHashKey(this)) {
////        cout << oldboard << endl;
//        printBoard();
//        getInfo();
//        cout << "from: " << bitScanForward(from) << endl;
//        cout << "to: " << bitScanForward(to) << endl;
//        cout << "key1: " << hashKey << endl;
//        cout << "key2: " << hashTable.getHashKey(this) << endl;
//        cout << "hashkey error" << endl << endl;
//        abort();        
//    }
    
    
    // den sicher richtigen hashKey setzen
//    hashKey = hashTable.getHashKey(this);
    
}

/************************************************************
 * Führt einen Zug aus und updated entsprechend den Angaben *
 * des Zuges die notwendigen Bitboards.                     *
 ************************************************************/
unsigned short int Position::move_figure (const unsigned long &from, const unsigned long &to, const int &promotion) {

    // aufräumen
    unsigned short int capture=CAPTURE_NONE;
    bool notOnMove=!onMove;
    if (ep>0) {
        hashKey^=hashTable.files[ep%8];
        ep=0;
    }

    // richtigen aktuellen/alten hashkey setzen
//    hashKey = hashTable.getHashKey(this);
    
    // schachbrett-bild speichern
//    string oldboard = getBoardString();
    
    // castling-rules aus hashKey entfernen
    short int cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];                
    
    // kill target figure (if there's one)
    if ((to & all)!=0L) { // all kann ersetzt werden
        if ((pawns[notOnMove]&to)!=0L) {
            pawns[notOnMove]^=to;
            hashKey^=hashTable.phash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_PAWN;
        } else if ((knights[notOnMove]&to)!=0L) {
            knights[notOnMove]^=to;
            hashKey^=hashTable.nhash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_KNIGHT;
        } else if ((bishops[notOnMove]&to)!=0L) {
            bishops[notOnMove]^=to;
            hashKey^=hashTable.bhash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_BISHOP;
        } else if ((queens[notOnMove]&to)!=0L) {
            queens[notOnMove]^=to;
            hashKey^=hashTable.qhash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_QUEEN;
        } else if ((rooks[notOnMove]&to)!=0L) {
            rooks[notOnMove]^=to;
            hashKey^=hashTable.rhash[notOnMove][bitScanForward(to)];
            // Castling Möglichkeit entfernen
            if (to==72057594037927936L) castlingRules&=13; // white_castling_long=false;
            else if (to==9223372036854775808UL) castlingRules&=14; // white_castling_short=false;
            else if (to==1L) castlingRules&=7; // black_castling_long=false;
            else if (to==128L) castlingRules&=11; // black_castling_short=false;                
            capture=CAPTURE_ROOK;
        }
        else if ((kings[notOnMove]&to)!=0L) {
            kings[notOnMove]^=to;
            hashKey^=hashTable.khash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_KING;
        }
    }

    // move figure
    if ((pawns[onMove]&from)!=0L) {
                
        // promote figure
        if ((to&promotionRow[onMove])!=0L) {            
            pawns[onMove] ^= from; // remove pawn
            hashKey^=hashTable.phash[onMove][bitScanForward(from)];
            if (promotion==0) { // add promoted figure
                queens[onMove] |= to;
                hashKey^=hashTable.qhash[onMove][bitScanForward(to)];
            } else if (promotion==1) {
                rooks[onMove] |= to;
                hashKey^=hashTable.rhash[onMove][bitScanForward(to)];
            } else if (promotion==2) {
                bishops[onMove] |= to;
                hashKey^=hashTable.bhash[onMove][bitScanForward(to)];
            } else if (promotion==3) {
                knights[onMove] |= to;
                hashKey^=hashTable.nhash[onMove][bitScanForward(to)];
            }
            capture|=32768; // set isPromotionMask
        } else {
            // ep-move
            if ((all&to)==0L) { // all kann ersetzt werden
                if (to==from>>9) { // hier könnten eigentlich if-elses gemacht werden
                    pawns[notOnMove]^=(from>>1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
                }
                if (to==from>>7) {
                    pawns[notOnMove]^=(from<<1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
                }
                if (to==from<<9) {
                    pawns[notOnMove]^=(from<<1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
                }
                if (to==from<<7) {
                    pawns[notOnMove]^=(from>>1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
                }
            }            
            pawns[onMove]^=(from|to);

            hashKey^=hashTable.phash[onMove][bitScanForward(from)];
            hashKey^=hashTable.phash[onMove][bitScanForward(to)];
            
            if (from==Position::BIN[48] && to==Position::BIN[32])   ep=9;
            else if (from==Position::BIN[49] && to==Position::BIN[33])   ep=10;
            else if (from==Position::BIN[50] && to==Position::BIN[34])   ep=11;
            else if (from==Position::BIN[51] && to==Position::BIN[35])   ep=12;
            else if (from==Position::BIN[52] && to==Position::BIN[36])   ep=13;
            else if (from==Position::BIN[53] && to==Position::BIN[37])   ep=14;
            else if (from==Position::BIN[54] && to==Position::BIN[38])   ep=15;
            else if (from==Position::BIN[55] && to==Position::BIN[39])   ep=16;
            else if (from==Position::BIN[8] && to==Position::BIN[24])    ep=1;
            else if (from==Position::BIN[9] && to==Position::BIN[25])    ep=2;
            else if (from==Position::BIN[10] && to==Position::BIN[26])   ep=3;
            else if (from==Position::BIN[11] && to==Position::BIN[27])   ep=4;
            else if (from==Position::BIN[12] && to==Position::BIN[28])   ep=5;
            else if (from==Position::BIN[13] && to==Position::BIN[29])   ep=6;
            else if (from==Position::BIN[14] && to==Position::BIN[30])   ep=7;
            else if (from==Position::BIN[15] && to==Position::BIN[31])   ep=8;
        }
    } else if ((queens[onMove]&from)!=0L) {
        queens[onMove]^=(from|to);
        hashKey^=hashTable.qhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.qhash[onMove][bitScanForward(to)];
    } else if ((bishops[onMove]&from)!=0L) {
        bishops[onMove]^=(from|to);
        hashKey^=hashTable.bhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.bhash[onMove][bitScanForward(to)];
    } else if ((knights[onMove]&from)!=0L) {
        knights[onMove]^=(from|to);
        hashKey^=hashTable.nhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.nhash[onMove][bitScanForward(to)];
    } else if ((kings[onMove]&from)!=0L) {
        
        // do castling
        // update king- and rook-position
        if ((from&1152921504606846992UL)!=0L && (to&4899916394579099716UL)!=0L) {
        
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];
            if (to==BIN[2]) {
                rochade(from,to,BIN[0],BIN[3]);
                hashKey^=hashTable.rhash[onMove][0];
                hashKey^=hashTable.rhash[onMove][3];
            } else if (to==BIN[6]) {
                rochade(from,to,BIN[7],BIN[5]);
                hashKey^=hashTable.rhash[onMove][7];
                hashKey^=hashTable.rhash[onMove][5];
            } else if (to==BIN[58]) {
                rochade(from,to,BIN[56],BIN[59]);
                hashKey^=hashTable.rhash[onMove][56];
                hashKey^=hashTable.rhash[onMove][59];
            } else if (to==BIN[62]) {
                rochade(from,to,BIN[63],BIN[61]);
                hashKey^=hashTable.rhash[onMove][63];
                hashKey^=hashTable.rhash[onMove][61];
            }

            // remove rochade possibilities
            castlingRules&=castlingR1[onMove];

//            updateBitboardLongs();
//            onMove=!onMove;
//            return capture;
        } else { // no castling        
            kings[onMove]^=(from|to);
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];
            if (from==1152921504606846976L) castlingRules&=12; // castling white entfernen
            else if (from==16L) castlingRules&=3; // castling black entfernen
        }
    } else if ((rooks[onMove]&from)!=0L) {
        rooks[onMove]^=(from|to);
        hashKey^=hashTable.rhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.rhash[onMove][bitScanForward(to)];
        if (from==72057594037927936L) castlingRules&=13; // white_castling_long
        else if (from==9223372036854775808UL) castlingRules&=14; // white_castling_short
        else if (from==1L) castlingRules&=7; // black_castling_long
        else if (from==128L) castlingRules&=11; // black_castling_short
    }
    
    // bitboars updaten und spieler am zug wechseln
    updateBitboardLongs();
    onMove=!onMove;
    
    // castling-rules zu hashkey hinzufügen
    cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];            
    
    // spieler am zug wechseln
    hashKey^=hashTable.side2move;

    // ggf. en-passant hash setzen
    if (ep>0)
        hashKey^=hashTable.files[ep%8];    

    // überprüfung ob hashkey korrekt berechnet wurde
//    if (hashKey!=hashTable.getHashKey(this)) {
////        cout << oldboard << endl;
//        printBoard();
//        getInfo();
//        cout << "from: " << bitScanForward(from) << endl;
//        cout << "to: " << bitScanForward(to) << endl;
//        cout << "key1: " << hashKey << endl;
//        cout << "key2: " << hashTable.getHashKey(this) << endl;
//        cout << "hashkey error" << endl << endl;
//        abort();
//    }
    
    // den sicher richtigen hashkey setzen
//    hashKey=hashTable.getHashKey(this);
    
    return capture;
}

/**********************************************************
 * Ausgabe-Funktion für das "print"-Command. Zeichnet den *
 * aktuellen Zustand des Schachbretts.                    *
 **********************************************************/
void Position::printBoard() {
    cout << getBoardString();
}

string Position::getBoardString() {
    string ret="";
    ret.append("  a b c d e f g h\n");
    for (int k=0;k<8;k++) {
        ret.append(int2String(8-k));
        for (int i=0;i<8;i++) {
            int position=k*8+i;
            if (figOnPos(kings[0],position))            ret.append(" K");
            else if (figOnPos(kings[1],position))       ret.append(" k");
            else if (figOnPos(queens[0],position))      ret.append(" Q");
            else if (figOnPos(queens[1],position))      ret.append(" q");
            else if (figOnPos(knights[0],position))     ret.append(" N");
            else if (figOnPos(knights[1],position))     ret.append(" n");
            else if (figOnPos(bishops[0],position))     ret.append(" B");
            else if (figOnPos(bishops[1],position))     ret.append(" b");        
            else if (figOnPos(pawns[0],position))       ret.append(" P");
            else if (figOnPos(pawns[1],position))       ret.append(" p");        
            else if (figOnPos(rooks[0],position))       ret.append(" R");        
            else if (figOnPos(rooks[1],position))       ret.append(" r");
            else                                        ret.append(" .");
        }
        ret.append("\n");
    }    
    return ret;
}

/*******************************************************
 * Ausgabe-Funktion für das "info"-Command. Diverse    *
 * Angaben zur aktuellen Position: CastlingRules, etc. *
 *******************************************************/
void Position::getInfo() {
    
    cout << getFen() << endl << endl;
    cout << "en passant: " << ep << endl << endl;
    cout << "white castling short: " << getBoolString(castlingRules&1) << endl;
    cout << "white castling long:  " << getBoolString(castlingRules&2) << endl;
    cout << "black castling short: " << getBoolString(castlingRules&4) << endl;
    cout << "black castling long:  " << getBoolString(castlingRules&8) << endl;    
    cout << endl;
    
    cout << "on Move: ";
    if (onMove)
        cout << "black";
    else
        cout << "white";
    
    cout << endl;
}

/***********************************************
 * Eine Rochade durchführen, dh. Bitboards von *
 * König und Turm entsprechend updaten.        *
 ***********************************************/
inline void Position::rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos) {
    kings[onMove]^=(old_king_pos|new_king_pos);
    rooks[onMove]^=(old_rook_pos|new_rook_pos);
}

/****************************************************
 * Updated die 3 Bitboards mit allen Figuren (all), *
 * weißen Figuren (aw) und schwarzen Figuren (ab).  *
 ****************************************************/
inline void Position::updateBitboardLongs() {
    aw=kings[0] | queens[0] | rooks[0] | bishops[0] | knights[0] | pawns[0];
    ab=kings[1] | queens[1] | rooks[1] | bishops[1] | knights[1] | pawns[1];
    all=ab|aw;    // nicht notwendig falls all nicht mehr notwendig ;-)
}

/*******************************************************************
 * Hauptfunktion zur Generierung aller möglichen Züge.             *
 * Dabei wird nicht überprüft, welche Züge auch tatsächlich gültig *
 * sind, da diese Berechnung viel Zeit in Anspruch nimmt und diese *
 * Beachtung bei der Zugsuche nicht immer notwendig ist.           *
 *******************************************************************/
void Position::calc_moves() {
    
    mvSize[cd]=0;
    
    unsigned long bitboard;    
    unsigned long myFigures;
    unsigned long opFigures;
    
    if (onMove) {
        // all figures bitboard from moving / opponent color
        myFigures=ab;
        opFigures=aw;
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_black(x);
            bitboard-=Position::BIN[x];
        }        
    } else {
        // all figures bitboard from moving / opponent color
        myFigures=aw;
        opFigures=ab;        
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_white(x);
            bitboard-=Position::BIN[x];
        }
    }
    
    // rooks
    bitboard = rooks[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // bishops
    bitboard = bishops[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // queens
    bitboard = queens[onMove];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);                        
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }         
    
    // knights
    bitboard = knights[onMove];
    while(bitboard != 0L) {
        int x=bitScanForward(bitboard);
        split_moves_direct(x, Position::KN_TARGET3x, Position::KN_TARGET3x_FROM[x], Position::KN_TARGET3x_TO[x], myFigures);
        bitboard-=Position::BIN[x];
    }       
    
    /*********
     * König *
     *********/
    int x = bitScanForward(kings[onMove]);
    split_moves_direct(x, Position::K_TARGET2x, Position::K_TARGET2x_FROM[x], Position::K_TARGET2x_TO[x], myFigures);

    /***********************
     * König/Turm Rochaden *
     ***********************/
    // white short, K
    if (!onMove
            && (castlingRules&1) // castling allowed
            && (all & 6917529027641081856L) == 0L // fields are free between king and rook; all kann ersetzt werden
            && opponent_target_bitboard2(1152921504606846976)==false
            && opponent_target_bitboard2(2305843009213693952)==false
            && opponent_target_bitboard2(4611686018427387904)==false
            && (pawns[1] & 69805794224242688L) == 0L) { // check if there's a pawn promotion move targeting an involved field

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=4611686018427387904L;
            mvSize[cd]++;
    }

    // white long, Q
    if (!onMove
            && (castlingRules&2)
            && (all & 1008806316530991104L) == 0L // all kann ersetzt werden
            && opponent_target_bitboard2(288230376151711744)==false 
            && opponent_target_bitboard2(576460752303423488)==false 
            && opponent_target_bitboard2(1152921504606846976)==false 
            && (pawns[1] & 17451448556060672L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=288230376151711744L;
            mvSize[cd]++;
    }

    // black short, k
    if (onMove
            && (castlingRules&4)
            && (all & 96L) == 0L // all kann ersetzt werden
            && opponent_target_bitboard2(16)==false 
            && opponent_target_bitboard2(32)==false 
            && opponent_target_bitboard2(64)==false 
            && (pawns[0] & 63488L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=64L;
            mvSize[cd]++;
    }

    // black long, q
    if (onMove
            && (castlingRules&8)
            && (all & 14L) == 0L // all kann ersetzt werden
            && opponent_target_bitboard2(4)==false 
            && opponent_target_bitboard2(8)==false 
            && opponent_target_bitboard2(16)==false 
            && (pawns[0] & 15872L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=4L;
            mvSize[cd]++;            
    }        
        
}

/************************************************
 * Funktion zur Generierung aller Capture-Züge. *
 * Wird in der Quiescence-Suche verwendet.      *
 ************************************************/
void Position::calcCaptures() {
    
    mvSize[cd]=0;
    
    unsigned long bitboard;    
    unsigned long opFigures;
    
    if (onMove) {
        // all figures bitboard from moving / opponent color
        opFigures=aw;
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_black_captures(x);
            bitboard-=Position::BIN[x];
        }        
    } else {
        // all figures bitboard from moving / opponent color
        opFigures=ab;        
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_white_captures(x);
            bitboard-=Position::BIN[x];
        }
    }
    
    /********
     * Turm *
     ********/
    bitboard = rooks[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves_captures__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    /**********
     * Läufer *
     **********/
    bitboard = bishops[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves_captures__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    /********
     * Dame *
     ********/
    bitboard = queens[onMove];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves_captures__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);                        
        split_moves_captures__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }

    /************
     * Springer *
     ************/
    bitboard = knights[onMove];
    while(bitboard != 0L) {
        int x=bitScanForward(bitboard);
        split_moves_direct_captures(x, Position::KN_TARGET3x, Position::KN_TARGET3x_FROM[x], Position::KN_TARGET3x_TO[x], opFigures);
        bitboard-=Position::BIN[x];
    }

    /*********
     * König *
     *********/
    int x = bitScanForward(kings[onMove]);
    split_moves_direct_captures(x, Position::K_TARGET2x, Position::K_TARGET2x_FROM[x], Position::K_TARGET2x_TO[x], opFigures);    
}

/****************************************************************
 * Generierte Capture-Züge für König und Springer aus aktueller *
 * Position und vorgefertigter Bitboards.                       *
 ****************************************************************/
inline void Position::split_moves_direct_captures(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & op) != 0L) { // only capture opponent color figures
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;
        }
    }
}

/****************************************************************
 * Generiert Züge für König und Springer aus aktueller Position *
 * und vorgefertigter Bitboards.                                *
 ****************************************************************/
inline void Position::split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & myFigures) == 0L) { // don't fight my color figures
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;
        }
    }
}

/*******************************************************************
 * Generiert Züge für Läufer, Turm und Dame aus aktueller Position *
 * und vorgefertigter Bitboards.                                   *
 *******************************************************************/
inline void Position::split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & all) == 0L) { // all kann ersetzt werden
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;                
        } else {
            if ((target[i] & op) != 0L) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[z];
                mvD[cd][mvSize[cd]][1]=target[i];
                mvSize[cd]++;                    
            }
            break;
        }
    }
}

/******************************************************************
 * Generiert Capture-Züge für Läufer, Turm und Dame aus aktueller *
 * Position und vorgefertigter Bitboards.                         *
 ******************************************************************/
inline void Position::split_moves_captures__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & all) == 0L) // all kann ersetzt werden
            continue;
        else {
            if ((target[i] & op) != 0L) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[z];
                mvD[cd][mvSize[cd]][1]=target[i];
                mvSize[cd]++;                    
            }
            break;
        }
    }
}

/**********************************************************
 * Die Züge der schwarzen Bauern zur Zugliste hinzufügen. *
 **********************************************************/
inline void Position::addPawnSingles_black(const int &z) {
    unsigned long startpos_black=65280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (Position::TARGET_BLACK_V1[z] & all) == 0UL) {// one step forward: just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                mvSize[cd]++;
                if ((Position::TARGET_BLACK_V1[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (Position::TARGET_BLACK_V3[z] & aw) != 0UL ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                mvSize[cd]++;        
                if ((Position::TARGET_BLACK_V3[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (Position::TARGET_BLACK_V4[z] & aw) != 0L ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_BLACK_V4[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;                    
                }
    }

    if (    (bitboardpos & startpos_black) != 0L && (Position::TARGET_BLACK_V2[z] & all) == 0L && (Position::TARGET_BLACK_V1[z] & all) == 0L) { // two steps forward: both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V2[z];
                mvSize[cd]++;        
    }    
    
    // en passant moves
    if (ep>8 && ep<=16) { // possible ep for black
        if (ep==9 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[40];
                mvSize[cd]++;
        }
        else if (ep==10 && bitboardpos==Position::BIN[32]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[32];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==10 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[39]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[39];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;
        }
        else if (ep==16 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[47];
                mvSize[cd]++;
        }
    }    
    
}

/*******************************************************
 * Die Züge der weißen Bauern zur Zugliste hinzufügen. *
 *******************************************************/
inline void Position::addPawnSingles_white(const int &z) {
    unsigned long startpos_white=71776119061217280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (TARGET_WHITE_V1[z] & all) == 0L) { // just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V1[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V1[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V1[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (TARGET_WHITE_V3[z] & ab) != 0L ) { // just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V3[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V3[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;                    
                }                
    }
    if (    (TARGET_WHITE_V4[z] & ab) != 0L ) {// just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V4[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;                    
                }                
    }
    
    if (    (bitboardpos & startpos_white) != 0L && (TARGET_WHITE_V2[z] & all) == 0L && (TARGET_WHITE_V1[z] & all) == 0L) { // both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V2[z];
                mvSize[cd]++;
    }   
    
    // en passant moves
    if (ep>0 && ep<=8) { // possible ep for white
        if (ep==1 && bitboardpos==Position::BIN[25]  ) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[16];
                mvSize[cd]++;            
        }
        else if (ep==2 && bitboardpos==Position::BIN[24]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[24];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==2 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[25]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[31]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[31];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==8 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[23];
                mvSize[cd]++;                        
        }
    }    
}

/******************************************************************
 * Die Capture-Züge der schwarzen Bauern zur Zugliste hinzufügen. *
 ******************************************************************/
inline void Position::addPawnSingles_black_captures(const int &z) {
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (Position::TARGET_BLACK_V3[z] & aw) != 0UL ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                mvSize[cd]++;        
                if ((Position::TARGET_BLACK_V3[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (Position::TARGET_BLACK_V4[z] & aw) != 0L ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_BLACK_V4[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;                    
                }
    }

/*
    // en passant moves
    if (ep>8 && ep<=16) { // possible ep for black
        if (ep==9 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[40];
                mvSize[cd]++;
        }
        else if (ep==10 && bitboardpos==Position::BIN[32]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[32];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==10 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[39]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[39];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;
        }
        else if (ep==16 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[47];
                mvSize[cd]++;
        }
    }    
*/    
}

/***************************************************************
 * Die Capture-Züge der weißen Bauern zur Zugliste hinzufügen. *
 ***************************************************************/
inline void Position::addPawnSingles_white_captures(const int &z) {
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (TARGET_WHITE_V3[z] & ab) != 0L ) { // just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V3[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V3[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;                    
                }                
    }
    if (    (TARGET_WHITE_V4[z] & ab) != 0L ) {// just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V4[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;                    
                }                
    }
    
/*
    // en passant moves
    if (ep>0 && ep<=8) { // possible ep for white
        if (ep==1 && bitboardpos==Position::BIN[25]  ) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[16];
                mvSize[cd]++;            
        }
        else if (ep==2 && bitboardpos==Position::BIN[24]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[24];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==2 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[25]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[31]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[31];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==8 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[23];
                mvSize[cd]++;                        
        }
    }    
*/
}

/*********************************************************************
 * Prüft ob eine Figur an der übergebenen Position am Schachbrett    *
 * vom Gegner bedroht wird. Primär verwendet um zu überprüfen ob der *
 * König der aktuellen Farbe im Schach steht.                        *
 *********************************************************************/
inline bool Position::opponent_target_bitboard2(const unsigned long &bitboard) {
    
    unsigned long bb=bitboard;
    
    int x=bitScanForward(bb);
    
    bool opponent = !onMove;
    
    // pawns
    if (PAWN_ATTACKS[onMove][x] & pawns[opponent])
        return true;
    
    // knights (direkt)
    for (unsigned int i=KN_TARGET3x_FROM[x];i<KN_TARGET3x_TO[x];i++) {
        if (KN_TARGET3x[i] & knights[opponent])
            return true;
    }
    


    // north (linie)
    for (unsigned int i=TARGET_Nx_FROM[x];i<TARGET_Nx_TO[x];i++) {
        if (TARGET_Nx[i] & all) {
            if (TARGET_Nx[i] & rooks[opponent])
                return true;
            if (TARGET_Nx[i] & queens[opponent])
                return true;
            break;
        }
    }
    
    // east
    for (unsigned int i=TARGET_Ex_FROM[x];i<TARGET_Ex_TO[x];i++) {
        if (TARGET_Ex[i] & all) {
            if (TARGET_Ex[i] & rooks[opponent])
                return true;
            if (TARGET_Ex[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // south
    for (unsigned int i=TARGET_Sx_FROM[x];i<TARGET_Sx_TO[x];i++) {
        if (TARGET_Sx[i] & all) {
            if (TARGET_Sx[i] & rooks[opponent])
                return true;
            if (TARGET_Sx[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // west
    for (unsigned int i=TARGET_Wx_FROM[x];i<TARGET_Wx_TO[x];i++) {
        if (TARGET_Wx[i] & all) {
            if (TARGET_Wx[i] & rooks[opponent])
                return true;
            if (TARGET_Wx[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // north-east
    for (unsigned int i=TARGET_NEx_FROM[x];i<TARGET_NEx_TO[x];i++) {
        if (TARGET_NEx[i] & all) {
            if (TARGET_NEx[i] & bishops[opponent])
                return true;
            if (TARGET_NEx[i] & queens[opponent])
                return true;
            break;
        }
    }    

    // south-east
    for (unsigned int i=TARGET_SEx_FROM[x];i<TARGET_SEx_TO[x];i++) {
        if (TARGET_SEx[i] & all) {
            if (TARGET_SEx[i] & bishops[opponent])
                return true;
            if (TARGET_SEx[i] & queens[opponent])
                return true;
            break;
        }
    }        
    
    // south-west
    for (unsigned int i=TARGET_SWx_FROM[x];i<TARGET_SWx_TO[x];i++) {
        if (TARGET_SWx[i] & all) {
            if (TARGET_SWx[i] & bishops[opponent])
                return true;
            if (TARGET_SWx[i] & queens[opponent])
                return true;
            break;
        }
    }        
    
    // north-west
    for (unsigned int i=TARGET_NWx_FROM[x];i<TARGET_NWx_TO[x];i++) {
        if (TARGET_NWx[i] & all) {
            if (TARGET_NWx[i] & bishops[opponent])
                return true;
            if (TARGET_NWx[i] & queens[opponent])
                return true;
            break;
        }
    }        

    // king (direkt)
    for (unsigned int i=K_TARGET2x_FROM[x];i<K_TARGET2x_TO[x];i++) {
        if (K_TARGET2x[i] & kings[opponent])
            return true;
    }        
    
    return false; 
}

string Position::getFenShort() {
    string fen;
    
    // 1. group: figures
    int spaces=0;
    for (int i=1;i<=64;i++) {
        if ((Position::BIN[i-1]&all)!=0L) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            if ((kings[0]&Position::BIN[i-1])!=0L) fen+="K";
            else if ((queens[0]&Position::BIN[i-1])!=0L) fen+="Q";
            else if ((bishops[0]&Position::BIN[i-1])!=0L) fen+="B";
            else if ((knights[0]&Position::BIN[i-1])!=0L) fen+="N";
            else if ((rooks[0]&Position::BIN[i-1])!=0L) fen+="R";
            else if ((pawns[0]&Position::BIN[i-1])!=0L) fen+="P";
            else if ((kings[1]&Position::BIN[i-1])!=0L) fen+="k";
            else if ((queens[1]&Position::BIN[i-1])!=0L) fen+="q";
            else if ((bishops[1]&Position::BIN[i-1])!=0L) fen+="b";
            else if ((knights[1]&Position::BIN[i-1])!=0L) fen+="n";
            else if ((rooks[1]&Position::BIN[i-1])!=0L) fen+="r";
            else if ((pawns[1]&Position::BIN[i-1])!=0L) fen+="p";
        } else
            spaces++;

        if (i%8==0 && i<64) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            fen+="/";
        }
        if (i==64 && spaces>0) {
                fen+=int2String(spaces);
        }
            
    }
    
    return fen;
}

/************************************************************
 * Konvertiert die Daten der aktuell gespeicherten Position *
 * in einen FEN-String.                                     *
 ************************************************************/
string Position::getFen() {
    
    string fen;
    
    // 1. group: figures
    int spaces=0;
    for (int i=1;i<=64;i++) {
        if ((Position::BIN[i-1]&all)!=0L) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            if ((kings[0]&Position::BIN[i-1])!=0L) fen+="K";
            else if ((queens[0]&Position::BIN[i-1])!=0L) fen+="Q";
            else if ((bishops[0]&Position::BIN[i-1])!=0L) fen+="B";
            else if ((knights[0]&Position::BIN[i-1])!=0L) fen+="N";
            else if ((rooks[0]&Position::BIN[i-1])!=0L) fen+="R";
            else if ((pawns[0]&Position::BIN[i-1])!=0L) fen+="P";
            else if ((kings[1]&Position::BIN[i-1])!=0L) fen+="k";
            else if ((queens[1]&Position::BIN[i-1])!=0L) fen+="q";
            else if ((bishops[1]&Position::BIN[i-1])!=0L) fen+="b";
            else if ((knights[1]&Position::BIN[i-1])!=0L) fen+="n";
            else if ((rooks[1]&Position::BIN[i-1])!=0L) fen+="r";
            else if ((pawns[1]&Position::BIN[i-1])!=0L) fen+="p";
        } else
            spaces++;

        if (i%8==0 && i<64) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            fen+="/";
        }
        if (i==64 && spaces>0) {
                fen+=int2String(spaces);
        }
            
    }

    // 2. group: color on move
    if (onMove)
        fen+=" b";
    else
        fen+=" w";

    // 3. group: castles
    if (castlingRules==0)
        fen+=" -";
    else {
        fen+=" ";
        if (castlingRules&1)   fen+="K";
        if (castlingRules&2)    fen+="Q";
        if (castlingRules&4)   fen+="k";
        if (castlingRules&8)    fen+="q";        
    }

    // 4. group: en passant
    if (ep==0)  fen+=" -";
    else if(ep==1)  fen+=" a6";
    else if (ep==2) fen+=" b6";
    else if (ep==3) fen+=" c6";
    else if (ep==4) fen+=" d6";
    else if (ep==5) fen+=" e6";
    else if (ep==6) fen+=" f6";
    else if (ep==7) fen+=" g6";
    else if (ep==8) fen+=" h6";
    else if (ep==9)     fen+=" a3";
    else if (ep==10)    fen+=" b3";
    else if (ep==11)    fen+=" c3";
    else if (ep==12)    fen+=" d3";
    else if (ep==13)    fen+=" e3";
    else if (ep==14)    fen+=" f3";
    else if (ep==15)    fen+=" g3";
    else if (ep==16)    fen+=" h3";

    return fen;
}

/*********************************************************
 * konvertiert den Zug mit dem übergebenen Index in eine *
 * lesbare Form, zb: a3a4, g7g8q, o.ä.                   *
 *********************************************************/
string Position::getMoveString(int i) {
    string val;

    bool isPromotion=false;
    
    val+=get_hum(bitScanForward(mvD[cd][i][0]));
    val+=get_hum(bitScanForward(mvD[cd][i][1]));
    
    if ((mvD[cd][i][0]&pawns[onMove])!=0L && (mvD[cd][i][1]&promotionRow[onMove])!=0L)
        isPromotion=true;
    
    if (!isPromotion)
        return val;
    
    if ((i%4)==0)    
        val+="q";
    else if ((i%4)==1)    
        val+="r";
    else if ((i%4)==2)    
        val+="b";
    else if ((i%4)==3)
        val+="n";    
    
    return val;
}

/********************************************************
 * konvertiert den Zug mit dem übergebenen Index in die *
 * algebraische Kurznotation. Noch unvollständig.       *
 ********************************************************/
string Position::getMoveStringAlgShort(int i) {
    string val;
    val+=get_hum(bitScanForward(mvD[cd][i][1]));
    return val;
}

/******************************************
 * Konvertiert einen Index [0-63] in eine *
 * Schachbrett-Koordinate [a-h][1-8]      *
 ******************************************/
string Position::get_hum(int k) {
    string str;

    if (k%8==0) str+="a";
    if (k%8==1) str+="b";
    if (k%8==2) str+="c";
    if (k%8==3) str+="d";
    if (k%8==4) str+="e";
    if (k%8==5) str+="f";
    if (k%8==6) str+="g";
    if (k%8==7) str+="h";
    str+=int2String(8-k/8);
    return str;
}