/* 
 * File:   Position.h
 * Author: peter
 *
 * Created on 31. Jänner 2012, 18:23
 */

#ifndef POSITION_H
#define	POSITION_H

#include <string>
#include <list>
#include <iostream>
#include <stdio.h>
#include <cstdlib>
#include "typedefs.h"

using namespace std;

class Position {
public:
    Position();
    Position(const string &fen);
    string getFen();
    string getFenShort();
    int getMMChilds(string moveArray[], int moveIndex[]);
    void getPerftInfo(int curr_depth, int depth, Perftres *perft);
    void getDivide(int color, int curr_depth, const int &max_depth, const bool &modus, long &sum, long &total);
    int negaMax(int alpha, int beta, int curr_depth, int depth, bool isPv, bool canNull);
    void printBoard();
    string getBoardString();
    void calc_moves();
    void calcCaptures();
    void doBackup();
    void undoMove(const unsigned long &from, const unsigned long &to, const int &promotion, const unsigned short int &capture);
    unsigned short int move_figure (const unsigned long &from, const unsigned long &to, const int &promotion);
    bool isEndgame();
    void getInfo();    
    string getMoveString(int i);
    string getMoveStringAlgShort(int i);
    
    unsigned long aw;
    unsigned long ab;
    unsigned long all;
    unsigned long kings[2];
    unsigned long pawns[2];
    unsigned long bishops[2];
    unsigned long knights[2];
    unsigned long queens[2];
    unsigned long rooks[2];    
    short int ep;
    bool onMove;
    char castlingRules;
    unsigned long hashKey;

    // Variablen für Suchbaum
    int cd;
    backup myBackup[128];
    unsigned long mvD[128][256][2]; // 256*2*8=4096 bytes = 4 kilobytes => bei maxdepth=64: 256 kilobytes = 0,25 megabyte
    unsigned short int mvSize[128];
    
    // Index vom besten Zug (negaMax), sofern zuvor berechnet
    int bestMoveIndex;
    
    
    unsigned int sum;
    
    
    // statische bitboards
    static const int index64[64];
    static const unsigned short int castlingR1[];
    static const unsigned long promotionRow[];
    static const unsigned long TARGET_Nx[];
    static const unsigned int TARGET_Nx_FROM[];
    static const unsigned int TARGET_Nx_TO[];
    static const unsigned long TARGET_Ex[];
    static const unsigned int TARGET_Ex_FROM[];
    static const unsigned int TARGET_Ex_TO[];    
    static const unsigned long TARGET_Sx[];
    static const unsigned int TARGET_Sx_FROM[];
    static const unsigned int TARGET_Sx_TO[];        
    static const unsigned long TARGET_Wx[];
    static const unsigned int TARGET_Wx_FROM[];
    static const unsigned int TARGET_Wx_TO[];        
    static const unsigned long TARGET_NEx[];
    static const unsigned int TARGET_NEx_FROM[];
    static const unsigned int TARGET_NEx_TO[];       
    static const unsigned long TARGET_SEx[];
    static const unsigned int TARGET_SEx_FROM[];
    static const unsigned int TARGET_SEx_TO[];       
    static const unsigned long TARGET_SWx[];
    static const unsigned int TARGET_SWx_FROM[];
    static const unsigned int TARGET_SWx_TO[];       
    static const unsigned long TARGET_NWx[];
    static const unsigned int TARGET_NWx_FROM[];
    static const unsigned int TARGET_NWx_TO[];           
    static const unsigned long BIN[];
    static const unsigned long K_TARGET2x[];
    static const unsigned int K_TARGET2x_FROM[];
    static const unsigned int K_TARGET2x_TO[];
    static const unsigned long KN_TARGET3x[];
    static const unsigned int KN_TARGET3x_FROM[];
    static const unsigned int KN_TARGET3x_TO[];
    static const unsigned int KN_RUN_FROM[];
    static const unsigned int KN_RUN_TO[];
    static const unsigned long TARGET_BLACK_V1[];
    static const unsigned long TARGET_BLACK_V2[];
    static const unsigned long TARGET_BLACK_V3[];
    static const unsigned long TARGET_BLACK_V4[];
    static const unsigned long TARGET_WHITE_V1[];
    static const unsigned long TARGET_WHITE_V2[];
    static const unsigned long TARGET_WHITE_V3[];
    static const unsigned long TARGET_WHITE_V4[];
    static const unsigned long PAWN_ATTACKS[2][64];
    
    bool inCheck() {
        return opponent_target_bitboard2(kings[onMove]);
    }
    
    bool isCheckMate();
    
    list<string> getMovePositions();
    
private:
    int quiesce(int alpha, int beta, int curr_depth);
    int evaluate(bool side2move);
    
    void sortMoves(unsigned short int sortedIndizes[], int cd, int bestmoveHash);
    unsigned short int sortMovesQuiesce(unsigned short int sortedIndizes[], int cd);
    void getMoveScores(short int moveScores[], int cd);
    
    
    inline void rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos);
    inline void updateBitboardLongs();    
    inline void split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures);
    inline void split_moves_direct_captures(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void split_moves_captures__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void addPawnSingles_black(const int &z);
    inline void addPawnSingles_white(const int &z);
    inline void addPawnSingles_black_captures(const int &z);
    inline void addPawnSingles_white_captures(const int &z);
    string get_hum(int k);
    inline int bitScanForward(const unsigned long &bb) {
        const unsigned long debruijn64 = 0x07EDD5E59A4E28C2;
        return Position::index64[((bb & -bb) * debruijn64) >> 58];
        
//        const int index64[64] = {
//            0, 47,  1, 56, 48, 27,  2, 60,
//           57, 49, 41, 37, 28, 16,  3, 61,
//           54, 58, 35, 52, 50, 42, 21, 44,
//           38, 32, 29, 23, 17, 11,  4, 62,
//           46, 55, 26, 59, 40, 36, 15, 53,
//           34, 51, 20, 43, 31, 22, 10, 45,
//           25, 39, 14, 33, 19, 30,  9, 24,
//           13, 18,  8, 12,  7,  6,  5, 63
//        };
//
//        
//        const unsigned long debruijn64 = 0x03f79d71b4cb0a89;
//           return index64[((bb ^ (bb-1)) * debruijn64) >> 58];        

    };
    inline int popCount(unsigned long x) { // too slow
//        int count = 0;
//        for (int i = 0; i < 64; i++, x >>= 1)
//                count += (int)x & 1;
//        return count;
        
      int c = 0;

      while (x) {
        c++;
        x &= x - 1;
      }
      return (c);        
        
    }
    inline bool opponent_target_bitboard2(const unsigned long &bitboard);
};

#endif	/* POSITION_H */