/* 
 * File:   Position.cpp
 * Author: peter
 * 
 * Created on 31. Jänner 2012, 18:23
 */

#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "Position.h"
#include "HashT.h"
#include "MoveList.h"
#include "Timer.h"
#include "globals.h"

using namespace std;

extern HashT hashTable;
extern MoveList moveList;
extern Timer timer;

extern unsigned short int distances[64][64];

/************************************
 * Initialisiert eine Startposition *
 ************************************/
Position::Position() {
    ep=0;
    kings[1]=Position::BIN[4];
    queens[1]=Position::BIN[3];
    bishops[1]=Position::BIN[2] | Position::BIN[5];
    knights[1]=Position::BIN[1] | Position::BIN[6];
    rooks[1]=Position::BIN[0] | Position::BIN[7];
    pawns[1]=Position::BIN[8] | Position::BIN[9] | Position::BIN[10] | Position::BIN[11] | Position::BIN[12] | Position::BIN[13] | Position::BIN[14] | Position::BIN[15];
    kings[0]=Position::BIN[60];
    queens[0]=Position::BIN[59];
    bishops[0]=Position::BIN[58] | Position::BIN[61];
    knights[0]=Position::BIN[57] | Position::BIN[62];
    rooks[0]=Position::BIN[56] | Position::BIN[63];
    pawns[0]=Position::BIN[48] | Position::BIN[49] | Position::BIN[50] | Position::BIN[51] | Position::BIN[52] | Position::BIN[53] | Position::BIN[54] | Position::BIN[55];
    updateBitboardLongs();
    castlingRules=15;
    onMove=false;
    
    hashKey = hashTable.getHashKey(this);
}

/*****************************************************************
 * Initialisiert eine Position durch den übergebenen FEN-String. *
 *****************************************************************/
Position::Position(const string &fen) {
    
    castlingRules=0;
    
    ep=0;
//    targetmap=0L;

    aw=0L;
    ab=0L;
    all=0L;
    
    pawns[0]=0L;
    pawns[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    bishops[0]=0L;
    bishops[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    queens[0]=0L;
    queens[1]=0L;
    knights[0]=0L;
    knights[1]=0L;
    kings[0]=0L;
    kings[1]=0L;
    
    size_t found;
    found=fen.find(" ");
    string fengrp0 = fen.substr(0,int(found));
    string rest = fen.substr(int(found)+1,fen.size()-int(found));
    string fengrp1 = rest.substr(0,1);
    rest = rest.substr(2,rest.size()-2);
    found=rest.find(" ");
    string fengrp2 = rest.substr(0,int(found));
    rest = rest.substr(int(found)+1,rest.size()-int(found));
    string fengrp3 = rest;

    // 1. group: figures
    int pos=0;
    for (unsigned int i=0;i<fengrp0.length();i++) {
        if (fengrp0.at(i)=='/') continue;
        else if (fengrp0.at(i)=='1') pos=pos+1;
        else if (fengrp0.at(i)=='2') pos=pos+2;
        else if (fengrp0.at(i)=='3') pos=pos+3;
        else if (fengrp0.at(i)=='4') pos=pos+4;
        else if (fengrp0.at(i)=='5') pos=pos+5;
        else if (fengrp0.at(i)=='6') pos=pos+6;
        else if (fengrp0.at(i)=='7') pos=pos+7;
        else if (fengrp0.at(i)=='8') pos=pos+8;
        else if (fengrp0.at(i)=='k') { kings[1]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='q') { queens[1]=queens[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='n') { knights[1]=knights[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='b') { bishops[1]=bishops[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='r') { rooks[1]=rooks[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='p') { pawns[1]=pawns[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='K') { kings[0]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='Q') { queens[0]=queens[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='N') { knights[0]=knights[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='B') { bishops[0]=bishops[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='R') { rooks[0]=rooks[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='P') { pawns[0]=pawns[0]|Position::BIN[pos]; pos++;}
    }
    
    // 2. group: color on move, fengrp1
    if (fengrp1=="b")
        onMove=true;
    else
        onMove=false;

    // 3. group: castles
    if (fengrp2.at(0)!='-') {
        for (unsigned int i=0;i<fengrp2.size();i++) {
            if (fengrp2.at(i)=='K')   castlingRules^=1; // white_castling_short=true;
            if (fengrp2.at(i)=='Q')   castlingRules^=2; // white_castling_long=true;
            if (fengrp2.at(i)=='k')   castlingRules^=4; // black_castling_short=true;
            if (fengrp2.at(i)=='q')   castlingRules^=8; // black_castling_long=true;            
        }
    }
    
    // 4. group: en passant
    int tmp_ep=0;
    if (fengrp3.length()==2) {
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='6') tmp_ep=1;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='6') tmp_ep=2;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='6') tmp_ep=3;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='6') tmp_ep=4;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='6') tmp_ep=5;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='6') tmp_ep=6;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='6') tmp_ep=7;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='6') tmp_ep=8;
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='3') tmp_ep=9;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='3') tmp_ep=10;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='3') tmp_ep=11;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='3') tmp_ep=12;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='3') tmp_ep=13;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='3') tmp_ep=14;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='3') tmp_ep=15;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='3') tmp_ep=16;
    }
    ep=tmp_ep;
    
    updateBitboardLongs();
    
    hashKey = hashTable.getHashKey(this);
}

/***************************
 * Rekursions-Funktion für *
 * pseudolegal Moves       *
 ***************************/
void Position::getChildsPseudo(int depth, long &sum) {
    cd=depth;
    calc_moves();
    if (depth>0) {
        doBackup();
        for (int i=0;i<mvSize[depth];i++) {
//            unsigned short int ret=pos.move_figure(pos.mvD[depth][i][0],pos.mvD[depth][i][1],pos.mvP[depth*256+i]);
            unsigned short int ret=move_figure(mvD[depth][i][0],mvD[depth][i][1],i%4);
            getChildsPseudo(depth-1, sum);
            cd=depth;
            // pos.undoMove(depth);
//            pos.undoMove(pos.mvD[depth][i][0],pos.mvD[depth][i][1],pos.mvP[depth*256+i],ret);
            undoMove(mvD[depth][i][0],mvD[depth][i][1],i%4,ret);
        }
    } else
        sum+=mvSize[depth]; // TODO ist das zählen hier wirklich richtig?
}

/**************************************************************
 * Wird ein UCI-Command oder ein Cmd des Benutzers            *
 * nicht verstanden, so wird diese Funktion aufgerufen,       *
 * um zu prüfen, ob es sich bei dem Cmd um einen zur Position *
 * gültigen Zug handelt (siehe main.cpp)                      *
 **************************************************************/
int Position::getMMChilds(string moveArray[], int moveIndex[]) {
    cd=0;
    doBackup();
    calc_moves();
    int cntMoves=0;
    for (int i=0;i<mvSize[cd];i++) {
        unsigned short int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);        
        if (!catchKing) {
            moveArray[cntMoves]=getMoveString(i);
            moveIndex[cntMoves]=i;
            cntMoves++;
        }
    }
    return cntMoves;
}

/**********************************************************************
 * Gibt zur gegebenen Position und Tiefe die Anzahl der Nodes         *
 * für jeden möglichen Zug an.                                        *
 * divide [n] ... Anzahl der Blätter/Nodes in Tiefe n                 *
 * divide2 [n] ... Summe aller Nodes bis inkl. Tiefe n                *
 * Diese Funktionen sind speziel zum Debugging des Zug-Generators     *
 * wertvoll. An dieser Stelle einen großen Dank an                    *
 * Roman Hartmann (http://www.rocechess.ch/roce.html),dessen Programm *
 * ich selbst verwendet hatte um meine Engine zu debuggen             *
 **********************************************************************/
void Position::getDivide(int color, int curr_depth, const int &depth, const bool &modus, long &sum, long &total) {
    cd = curr_depth-1;
    calc_moves();
    doBackup();
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            unsigned short int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);            
            if (curr_depth==1)
                sum=0;
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            if (!catchKing) {
                getDivide(abs(color-1), (curr_depth+1),depth, modus, sum, total);
                cd--;
                if (curr_depth==1)
                    cout << getMoveString(i) << " " << sum << "     " << getFen() << endl;                                
            } else {
                if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
                    sum--;
                    total--;
                }                
            }
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);            
        }
    }
    if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
        sum+=mvSize[cd];
        total+=mvSize[cd];
    }
}

/*************************************************************************
 * Führt einen Performance-Test zur aktuellen Position aus und           *
 * liefert ausführliche Daten zu Captures, En-Passants, Promotions, etc. *
 *************************************************************************/
void Position::getPerftInfo(int curr_depth, int depth, Perftres *perft) {
    
    cd = curr_depth-1;
    doBackup();
    calc_moves();
    
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            unsigned short int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
            
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            
            if (!catchKing) {
                getPerftInfo((curr_depth+1),depth, perft);
                cd--;

                // count captures
//                if (ret)
//                    perft->perftres[1+(curr_depth-1)*7]++;                    

                // count ep-moves
                // if (mvD[cd][i][2]!=0L) { // <-- funktioniert nicht mehr, kill / [2] nicht mehr vorhanden
                if (mvD[cd][i][1]!=0L) {                
                    perft->perftres[4+(curr_depth-1)*7]++;
                  //  perft->perftres[1+(curr_depth-1)*7]++;
                } 

                // count promotions
//                if ([cd*256+i]>0)
//                    perft->perftres[5+(curr_depth-1)*7]++;                                
                
                // count castlings
                if ((mvD[cd][i][1]&kings[!onMove])!=0L && (mvD[cd][i][0]<<2==mvD[cd][i][1] || mvD[cd][i][0]>>2==mvD[cd][i][1]))
                    perft->perftres[6+(curr_depth-1)*7]++;                     
            
            } else
                perft->perftres[(curr_depth-1)*7]--;                
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
        }
    }    

    perft->perftres[(curr_depth-1)*7]+=mvSize[cd];
    
    if (perft!=NULL) {
        /*        


        // count checks
        Position tmp_pos3(*tmp_pos);
        if (opponent_target_bitboard(tmp_pos3,tmp_pos->kings[abs(color-1)],abs(color-1)))
            perft->perftres[2+(curr_depth-1)*7]++;                

        */
    }    
}

void Position::testHash(int curr_depth, int depth, long &sum) {
    
    sum++;
    cd = curr_depth;    
    calc_moves();
    doBackup();  
    
    for (int i=mvSize[cd]-1;i>=0;i--) {
        unsigned short int ret = move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);        
        
        if (hashKey!=hashTable.getHashKey(this))
            cout << "hash-key failed" << endl;
        
        if (curr_depth<depth)
            testHash(cd+1,depth,sum);
        cd=curr_depth;        
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
        
    }
}

/*****************************************************************************
 * Sortiert das Array mit den möglichen Zügen einer Position.                *
 * Züge mit Captures werden bevorzugt.                                       *
 * Wert eines Zuges ist abhängig welcher Figuren-Typ welchen gegnerischen    *
 * Figuren-Typ schlägt, dh. Bauer schlägt Dame ist ein wertvollerer Zug als  *
 * Dame schlägt Bauer, usw.                                                  *
 * Die Sortierung dient dazu, beim NegaMax früher bzw. bessere Cutoffs       *
 * zu bekommen, somit sollte sich die Zeit der Zugsuche deutlich reduzieren. *
 *****************************************************************************/
void Position::sortMoves(unsigned short int sortedIndizes[], int cd, int bestmoveHash) {
    unsigned short int cntMoves=mvSize[cd];
    short int moveValues[cntMoves];

    /*************************************************************
     * Jedem Zug einen Score zuweisen. Züge ohne Captures        *
     * werden mit dem Wert -2000 bewertet. Züge mit Captures mit *
     * einem Wert abhängig davon, welcher Figurentyp welchen     *
     * gegnerischen Figurentyp schlägt.                          *
     *************************************************************/
    for (int i=0;i<cntMoves;i++) {
        
        if ((mvD[cd][i][1]&all)==0L) { // all kann ersetzt werden
            moveValues[i]=-2000;
            continue;
        }
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else                                            fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else                                               toValue=1;        

        moveValues[i]=toValue-fromValue;
    }
    
    /**************************************************
     * Index des besten Zuges aus dem Hash markieren. *
     **************************************************/
    if (bestmoveHash>=0 && bestmoveHash<cntMoves)
        moveValues[bestmoveHash]=10000;
    
    /********************
     * Werte sortieren. *
     ********************/
    for (int i=0;i<cntMoves;i++)
        sortedIndizes[i]=i;
    
//    /* Bubblesort */
//    for (int n=cntMoves;n>1;n--) {
//        for (int k=0;k<n-1;k++) {
//            if (moveValues[sortedIndizes[k]]>moveValues[sortedIndizes[k+1]]) {
//                int tmp=sortedIndizes[k];
//                sortedIndizes[k]=sortedIndizes[k+1];
//                sortedIndizes[k+1]=tmp;
//            }
//        }
//    }
    
     Heap::buildHeap(sortedIndizes,moveValues,cntMoves);    
}

unsigned short int Position::sortMovesQuiesce(unsigned short int sortedIndizes[], int cd) {
    unsigned short int cntMoves=mvSize[cd];
    short int moveValues[cntMoves];

    // calculate a value for every move
    for (int i=0;i<cntMoves;i++) {
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else                                            fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else                                               toValue=1;        

        moveValues[i]=toValue-fromValue;
    }
    
    for (int i=0;i<cntMoves;i++)
        sortedIndizes[i]=i;
    
    Heap::buildHeap(sortedIndizes,moveValues,cntMoves);
    
    return 0;
}


void Position::getMoveScores(short int moveScores[], int cd) {
    for (int i=0;i<mvSize[cd];i++) {
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else                                            fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else                                               toValue=1;        

        moveScores[i]=toValue-fromValue;
    }    
}




bool Position::isCheckMate() {
       
    calc_moves();
    doBackup();
    int i;
    cd = 0;
    for (i=mvSize[cd]-1;i>=0;i--) {
        unsigned short int ret = move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
        
        /***********************************************************************************
         * Prüft ob der Zug legal ist. Das ist dann der Fall, wenn der gegnerische Spieler *
         * anschließend nicht den König schlagen kann. Da diese Überprüfung sehr aufwendig *
         * ist, wird sie nur in der untersten Ebene durchgeführt, damit die Engine nur     *
         * einen legalen Zug als besten Zug wählen kann.                                   *
         ***********************************************************************************/
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        if (catchKing) {
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
            continue;
        }
        
        
        return false; // Zug gültig, dh. Spielfarbe kann mit diesem Zug dem Schach entweichen. Daher kann es nicht Schachmatt sein.
        
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
    }
    return true; // Alle prinzipiell möglichen Züge sind ungültig, dh. Spielfarbe kann nicht aus dem Schach fliehen. Daher Schachmatt.
}

list<string> Position::getMovePositions() {
    list<string> positions;    

    cd = 0;
    calc_moves();
    doBackup();
    int i;
    for (i=mvSize[cd]-1;i>=0;i--) {


        unsigned short int ret = move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
        
        /***********************************************************************************
         * Prüft ob der Zug legal ist. Das ist dann der Fall, wenn der gegnerische Spieler *
         * anschließend nicht den König schlagen kann. Da diese Überprüfung sehr aufwendig *
         * ist, wird sie nur in der untersten Ebene durchgeführt, damit die Engine nur     *
         * einen legalen Zug als besten Zug wählen kann.                                   *
         ***********************************************************************************/
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        if (catchKing) {
            undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
            continue;
        }

//        printBoard();
//        cout << bitScanForward(mvD[cd][i][0]) << "," << bitScanForward(mvD[cd][i][1]) << endl;
//        cout << getFenShort() << endl;
//        cout << pawns[!onMove] << endl;
//        cout << kings[!onMove] << endl;
//        cout << pawns[onMove] << endl;
//        cout << kings[onMove] << endl;
//        cout << endl;        

        
        
        positions.push_back(getFenShort());
        
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);
    }    
    
//    abort();
    return positions;
}


/******************************************************
 * Gibt zurück, ob es sich bei der aktuellen Position *
 * um eine Position im Endspiel handelt oder nicht.   *
 * Ein Endspiel ist gegeben, wenn:                    *
 * - keine der Damen mehr im Spiel sind,              *
 * - einer der Spieler weniger als 8 Figuren hat      *
 ******************************************************/
bool Position::isEndgame() {
    if (queens[0]==0UL && queens[1]==0UL)
        return true;
    if (popCount(aw)<8 || popCount(ab)<8)
        return true;
    return false;
}




/************************************
 * Speichert die aktuelle Position. *
 ************************************/
void Position::doBackup() {
    myBackup[cd].ep=ep;
    myBackup[cd].castlingRules=castlingRules;
}

/************************************
 * Lädt eine gespeicherte Position. *
 ************************************/
void Position::undoMove(const unsigned long &from, const unsigned long &to, const int &promotion, const unsigned short int &capture) {

    // richtigen aktuellen/alten hashkey setzen
//    hashKey = hashTable.getHashKey(this);    

    // castling-rules aus hashKey entfernen
    short int cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];                    
    
    bool notOnMove=onMove;
    onMove=!onMove;
    
    hashKey^=hashTable.side2move;
    
    // backmove figures
    if ((pawns[onMove]&to)!=0L) {
        pawns[onMove]^=(from|to);
        
        hashKey^=hashTable.phash[onMove][bitScanForward(from)];
        hashKey^=hashTable.phash[onMove][bitScanForward(to)];
        
        // ep-move
        if (capture==CAPTURE_NONE) {
            if (to==from>>9) { // kann eigentlich zu if-elses gemacht werden
                pawns[notOnMove]^=(from>>1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
            }
            if (to==from>>7) {
                pawns[notOnMove]^=(from<<1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
            }
            if (to==from<<9) {
                pawns[notOnMove]^=(from<<1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
            }
            if (to==from<<7) {
                pawns[notOnMove]^=(from>>1);
                hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
            }
        }
    } else if ((knights[onMove]&to)!=0L) {
        knights[onMove]^=(from|to);
        hashKey^=hashTable.nhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.nhash[onMove][bitScanForward(to)];        
    } else if ((bishops[onMove]&to)!=0L) {
        bishops[onMove]^=(from|to);
        hashKey^=hashTable.bhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.bhash[onMove][bitScanForward(to)];        
    } else if ((rooks[onMove]&to)!=0L) {
        rooks[onMove]^=(from|to);
        hashKey^=hashTable.rhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.rhash[onMove][bitScanForward(to)];        
    } else if ((queens[onMove]&to)!=0L) {
        queens[onMove]^=(from|to);
        hashKey^=hashTable.qhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.qhash[onMove][bitScanForward(to)];        
    } else if ((kings[onMove]&to)!=0L) {
        if ((from&1152921504606846992UL)!=0L && (to&4899916394579099716UL)!=0L) {
            if (to==BIN[2]) {
                rochade(from,to,BIN[0],BIN[3]);
                hashKey^=hashTable.rhash[onMove][0];
                hashKey^=hashTable.rhash[onMove][3];            
            } else if (to==BIN[6]) {
                rochade(from,to,BIN[7],BIN[5]);
                hashKey^=hashTable.rhash[onMove][7];
                hashKey^=hashTable.rhash[onMove][5];                            
            } else if (to==BIN[58]) {
                rochade(from,to,BIN[56],BIN[59]);
                hashKey^=hashTable.rhash[onMove][56];
                hashKey^=hashTable.rhash[onMove][59];                            
            } else if (to==BIN[62]) {
                rochade(from,to,BIN[63],BIN[61]);
                hashKey^=hashTable.rhash[onMove][63];
                hashKey^=hashTable.rhash[onMove][61];
            }
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];            
        } else {
            kings[onMove]^=(from|to);            
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];            
        }
    }
    
    // captures
    if ((capture&CAPTURE_MASK)==CAPTURE_PAWN) {
        pawns[notOnMove]|=to;
        hashKey^=hashTable.phash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_KNIGHT) {
        knights[notOnMove]|=to;
        hashKey^=hashTable.nhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_BISHOP) {
        bishops[notOnMove]|=to;
        hashKey^=hashTable.bhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_ROOK) {
        rooks[notOnMove]|=to;
        hashKey^=hashTable.rhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_QUEEN) {
        queens[notOnMove]|=to;
        hashKey^=hashTable.qhash[notOnMove][bitScanForward(to)];
    } else if ((capture&CAPTURE_MASK)==CAPTURE_KING) {
        kings[notOnMove]|=to;
        hashKey^=hashTable.khash[notOnMove][bitScanForward(to)];
    }
    
    // recover promotion
    if ((capture&32768)!=0) { // isPromotionMask gesetzt?
        pawns[onMove]^=from;
        hashKey^=hashTable.phash[onMove][bitScanForward(from)];
        if (promotion==0) {
            queens[onMove]^=from;
            hashKey^=hashTable.qhash[onMove][bitScanForward(from)];
        } else if (promotion==1) {
            rooks[onMove]^=from;
            hashKey^=hashTable.rhash[onMove][bitScanForward(from)];
        } else if (promotion==2) {
            bishops[onMove]^=from;
            hashKey^=hashTable.bhash[onMove][bitScanForward(from)];
        } else if (promotion==3) {
            knights[onMove]^=from;
            hashKey^=hashTable.nhash[onMove][bitScanForward(from)];
        }
    }

    
    if (ep>0)
        hashKey^=hashTable.files[ep%8];    
    
    // recover castling-rules & ep-flag
    ep=myBackup[cd].ep;
    castlingRules=myBackup[cd].castlingRules;        
    updateBitboardLongs();
    
    if (ep>0)
        hashKey^=hashTable.files[ep%8];    
    

    
    cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];            
    
    
    
//    if (hashKey!=hashTable.getHashKey(this)) {
////        cout << oldboard << endl;
//        printBoard();
//        getInfo();
//        cout << "from: " << bitScanForward(from) << endl;
//        cout << "to: " << bitScanForward(to) << endl;
//        cout << "key1: " << hashKey << endl;
//        cout << "key2: " << hashTable.getHashKey(this) << endl;
//        cout << "hashkey error" << endl << endl;
//        abort();        
//    }
    
    
    // den sicher richtigen hashKey setzen
//    hashKey = hashTable.getHashKey(this);
    
}

/************************************************************
 * Führt einen Zug aus und updated entsprechend den Angaben *
 * des Zuges die notwendigen Bitboards.                     *
 ************************************************************/
unsigned short int Position::move_figure (const unsigned long &from, const unsigned long &to, const int &promotion) {

    // aufräumen
    unsigned short int capture=CAPTURE_NONE;
    bool notOnMove=!onMove;
    if (ep>0) {
        hashKey^=hashTable.files[ep%8];
        ep=0;
    }

    // richtigen aktuellen/alten hashkey setzen
//    hashKey = hashTable.getHashKey(this);
    
    // schachbrett-bild speichern
//    string oldboard = getBoardString();
    
    // castling-rules aus hashKey entfernen
    short int cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];                
    
    // kill target figure (if there's one)
    if ((to & all)!=0L) { // all kann ersetzt werden
        if ((pawns[notOnMove]&to)!=0L) {
            pawns[notOnMove]^=to;
            hashKey^=hashTable.phash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_PAWN;
        } else if ((knights[notOnMove]&to)!=0L) {
            knights[notOnMove]^=to;
            hashKey^=hashTable.nhash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_KNIGHT;
        } else if ((bishops[notOnMove]&to)!=0L) {
            bishops[notOnMove]^=to;
            hashKey^=hashTable.bhash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_BISHOP;
        } else if ((queens[notOnMove]&to)!=0L) {
            queens[notOnMove]^=to;
            hashKey^=hashTable.qhash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_QUEEN;
        } else if ((rooks[notOnMove]&to)!=0L) {
            rooks[notOnMove]^=to;
            hashKey^=hashTable.rhash[notOnMove][bitScanForward(to)];
            // Castling Möglichkeit entfernen
            if (to==72057594037927936L) castlingRules&=13; // white_castling_long=false;
            else if (to==9223372036854775808UL) castlingRules&=14; // white_castling_short=false;
            else if (to==1L) castlingRules&=7; // black_castling_long=false;
            else if (to==128L) castlingRules&=11; // black_castling_short=false;                
            capture=CAPTURE_ROOK;
        }
        else if ((kings[notOnMove]&to)!=0L) {
            kings[notOnMove]^=to;
            hashKey^=hashTable.khash[notOnMove][bitScanForward(to)];
            capture=CAPTURE_KING;
        }
    }

    // move figure
    if ((pawns[onMove]&from)!=0L) {
                
        // promote figure
        if ((to&promotionRow[onMove])!=0L) {            
            pawns[onMove] ^= from; // remove pawn
            hashKey^=hashTable.phash[onMove][bitScanForward(from)];
            if (promotion==0) { // add promoted figure
                queens[onMove] |= to;
                hashKey^=hashTable.qhash[onMove][bitScanForward(to)];
            } else if (promotion==1) {
                rooks[onMove] |= to;
                hashKey^=hashTable.rhash[onMove][bitScanForward(to)];
            } else if (promotion==2) {
                bishops[onMove] |= to;
                hashKey^=hashTable.bhash[onMove][bitScanForward(to)];
            } else if (promotion==3) {
                knights[onMove] |= to;
                hashKey^=hashTable.nhash[onMove][bitScanForward(to)];
            }
            capture|=32768; // set isPromotionMask
        } else {
            // ep-move
            if ((all&to)==0L) { // all kann ersetzt werden
                if (to==from>>9) { // hier könnten eigentlich if-elses gemacht werden
                    pawns[notOnMove]^=(from>>1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
                }
                if (to==from>>7) {
                    pawns[notOnMove]^=(from<<1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
                }
                if (to==from<<9) {
                    pawns[notOnMove]^=(from<<1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from<<1)];
                }
                if (to==from<<7) {
                    pawns[notOnMove]^=(from>>1);
                    hashKey^=hashTable.phash[notOnMove][bitScanForward(from>>1)];
                }
            }            
            pawns[onMove]^=(from|to);

            hashKey^=hashTable.phash[onMove][bitScanForward(from)];
            hashKey^=hashTable.phash[onMove][bitScanForward(to)];
            
            if (from==Position::BIN[48] && to==Position::BIN[32])   ep=9;
            else if (from==Position::BIN[49] && to==Position::BIN[33])   ep=10;
            else if (from==Position::BIN[50] && to==Position::BIN[34])   ep=11;
            else if (from==Position::BIN[51] && to==Position::BIN[35])   ep=12;
            else if (from==Position::BIN[52] && to==Position::BIN[36])   ep=13;
            else if (from==Position::BIN[53] && to==Position::BIN[37])   ep=14;
            else if (from==Position::BIN[54] && to==Position::BIN[38])   ep=15;
            else if (from==Position::BIN[55] && to==Position::BIN[39])   ep=16;
            else if (from==Position::BIN[8] && to==Position::BIN[24])    ep=1;
            else if (from==Position::BIN[9] && to==Position::BIN[25])    ep=2;
            else if (from==Position::BIN[10] && to==Position::BIN[26])   ep=3;
            else if (from==Position::BIN[11] && to==Position::BIN[27])   ep=4;
            else if (from==Position::BIN[12] && to==Position::BIN[28])   ep=5;
            else if (from==Position::BIN[13] && to==Position::BIN[29])   ep=6;
            else if (from==Position::BIN[14] && to==Position::BIN[30])   ep=7;
            else if (from==Position::BIN[15] && to==Position::BIN[31])   ep=8;
        }
    } else if ((queens[onMove]&from)!=0L) {
        queens[onMove]^=(from|to);
        hashKey^=hashTable.qhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.qhash[onMove][bitScanForward(to)];
    } else if ((bishops[onMove]&from)!=0L) {
        bishops[onMove]^=(from|to);
        hashKey^=hashTable.bhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.bhash[onMove][bitScanForward(to)];
    } else if ((knights[onMove]&from)!=0L) {
        knights[onMove]^=(from|to);
        hashKey^=hashTable.nhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.nhash[onMove][bitScanForward(to)];
    } else if ((kings[onMove]&from)!=0L) {
        
        // do castling
        // update king- and rook-position
        if ((from&1152921504606846992UL)!=0L && (to&4899916394579099716UL)!=0L) {
        
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];
            if (to==BIN[2]) {
                rochade(from,to,BIN[0],BIN[3]);
                hashKey^=hashTable.rhash[onMove][0];
                hashKey^=hashTable.rhash[onMove][3];
            } else if (to==BIN[6]) {
                rochade(from,to,BIN[7],BIN[5]);
                hashKey^=hashTable.rhash[onMove][7];
                hashKey^=hashTable.rhash[onMove][5];
            } else if (to==BIN[58]) {
                rochade(from,to,BIN[56],BIN[59]);
                hashKey^=hashTable.rhash[onMove][56];
                hashKey^=hashTable.rhash[onMove][59];
            } else if (to==BIN[62]) {
                rochade(from,to,BIN[63],BIN[61]);
                hashKey^=hashTable.rhash[onMove][63];
                hashKey^=hashTable.rhash[onMove][61];
            }

            // remove rochade possibilities
            castlingRules&=castlingR1[onMove];

//            updateBitboardLongs();
//            onMove=!onMove;
//            return capture;
        } else { // no castling        
            kings[onMove]^=(from|to);
            hashKey^=hashTable.khash[onMove][bitScanForward(from)];
            hashKey^=hashTable.khash[onMove][bitScanForward(to)];
            if (from==1152921504606846976L) castlingRules&=12; // castling white entfernen
            else if (from==16L) castlingRules&=3; // castling black entfernen
        }
    } else if ((rooks[onMove]&from)!=0L) {
        rooks[onMove]^=(from|to);
        hashKey^=hashTable.rhash[onMove][bitScanForward(from)];
        hashKey^=hashTable.rhash[onMove][bitScanForward(to)];
        if (from==72057594037927936L) castlingRules&=13; // white_castling_long
        else if (from==9223372036854775808UL) castlingRules&=14; // white_castling_short
        else if (from==1L) castlingRules&=7; // black_castling_long
        else if (from==128L) castlingRules&=11; // black_castling_short
    }
    
    // bitboars updaten und spieler am zug wechseln
    updateBitboardLongs();
    onMove=!onMove;
    
    // castling-rules zu hashkey hinzufügen
    cast=0;
    cast^=castlingRules;
    hashKey^=hashTable.castling[cast];            
    
    // spieler am zug wechseln
    hashKey^=hashTable.side2move;

    // ggf. en-passant hash setzen
    if (ep>0)
        hashKey^=hashTable.files[ep%8];    

    // überprüfung ob hashkey korrekt berechnet wurde
//    if (hashKey!=hashTable.getHashKey(this)) {
////        cout << oldboard << endl;
//        printBoard();
//        getInfo();
//        cout << "from: " << bitScanForward(from) << endl;
//        cout << "to: " << bitScanForward(to) << endl;
//        cout << "key1: " << hashKey << endl;
//        cout << "key2: " << hashTable.getHashKey(this) << endl;
//        cout << "hashkey error" << endl << endl;
//        abort();
//    }
    
    // den sicher richtigen hashkey setzen
//    hashKey=hashTable.getHashKey(this);
    
    return capture;
}

/**********************************************************
 * Ausgabe-Funktion für das "print"-Command. Zeichnet den *
 * aktuellen Zustand des Schachbretts.                    *
 **********************************************************/
void Position::printBoard() {
    cout << getBoardString();
}

string Position::getBoardString() {
    string ret="";
    ret.append("  a b c d e f g h\n");
    for (int k=0;k<8;k++) {
        ret.append(int2String(8-k));
        for (int i=0;i<8;i++) {
            int position=k*8+i;
            if (figOnPos(kings[0],position))            ret.append(" K");
            else if (figOnPos(kings[1],position))       ret.append(" k");
            else if (figOnPos(queens[0],position))      ret.append(" Q");
            else if (figOnPos(queens[1],position))      ret.append(" q");
            else if (figOnPos(knights[0],position))     ret.append(" N");
            else if (figOnPos(knights[1],position))     ret.append(" n");
            else if (figOnPos(bishops[0],position))     ret.append(" B");
            else if (figOnPos(bishops[1],position))     ret.append(" b");        
            else if (figOnPos(pawns[0],position))       ret.append(" P");
            else if (figOnPos(pawns[1],position))       ret.append(" p");        
            else if (figOnPos(rooks[0],position))       ret.append(" R");        
            else if (figOnPos(rooks[1],position))       ret.append(" r");
            else                                        ret.append(" .");
        }
        ret.append("\n");
    }    
    return ret;
}

/*******************************************************
 * Ausgabe-Funktion für das "info"-Command. Diverse    *
 * Angaben zur aktuellen Position: CastlingRules, etc. *
 *******************************************************/
void Position::getInfo() {
    
    cout << getFen() << endl << endl;
    cout << "en passant: " << ep << endl << endl;
    cout << "white castling short: " << getBoolString(castlingRules&1) << endl;
    cout << "white castling long:  " << getBoolString(castlingRules&2) << endl;
    cout << "black castling short: " << getBoolString(castlingRules&4) << endl;
    cout << "black castling long:  " << getBoolString(castlingRules&8) << endl;    
    cout << endl;
    
    cout << "on Move: ";
    if (onMove)
        cout << "black";
    else
        cout << "white";
    
    cout << endl;
}

/***********************************************
 * Eine Rochade durchführen, dh. Bitboards von *
 * König und Turm entsprechend updaten.        *
 ***********************************************/
inline void Position::rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos) {
    kings[onMove]^=(old_king_pos|new_king_pos);
    rooks[onMove]^=(old_rook_pos|new_rook_pos);
}

/****************************************************
 * Updated die 3 Bitboards mit allen Figuren (all), *
 * weißen Figuren (aw) und schwarzen Figuren (ab).  *
 ****************************************************/
inline void Position::updateBitboardLongs() {
    aw=kings[0] | queens[0] | rooks[0] | bishops[0] | knights[0] | pawns[0];
    ab=kings[1] | queens[1] | rooks[1] | bishops[1] | knights[1] | pawns[1];
    all=ab|aw;    // nicht notwendig falls all nicht mehr notwendig ;-)
}

/*******************************************************************
 * Hauptfunktion zur Generierung aller möglichen Züge.             *
 * Dabei wird nicht überprüft, welche Züge auch tatsächlich gültig *
 * sind, da diese Berechnung viel Zeit in Anspruch nimmt und diese *
 * Beachtung bei der Zugsuche nicht immer notwendig ist.           *
 *******************************************************************/
void Position::calc_moves() {
    
    mvSize[cd]=0;
    
    unsigned long bitboard;    
    unsigned long myFigures;
    unsigned long opFigures;
    
    if (onMove) {
        // all figures bitboard from moving / opponent color
        myFigures=ab;
        opFigures=aw;
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_black(x);
            bitboard-=Position::BIN[x];
        }        
    } else {
        // all figures bitboard from moving / opponent color
        myFigures=aw;
        opFigures=ab;        
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_white(x);
            bitboard-=Position::BIN[x];
        }
    }
    
    // rooks
    bitboard = rooks[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // bishops
    bitboard = bishops[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // queens
    bitboard = queens[onMove];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);                        
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }         
    
    // knights
    bitboard = knights[onMove];
    while(bitboard != 0L) {
        int x=bitScanForward(bitboard);
        split_moves_direct(x, Position::KN_TARGET3x, Position::KN_TARGET3x_FROM[x], Position::KN_TARGET3x_TO[x], myFigures);
        bitboard-=Position::BIN[x];
    }       
    
    /*********
     * König *
     *********/
    int x = bitScanForward(kings[onMove]);
    split_moves_direct(x, Position::K_TARGET2x, Position::K_TARGET2x_FROM[x], Position::K_TARGET2x_TO[x], myFigures);

    /***********************
     * König/Turm Rochaden *
     ***********************/
    // white short, K
    if (!onMove
            && (castlingRules&1) // castling allowed
            && (all & 6917529027641081856L) == 0L // fields are free between king and rook; all kann ersetzt werden
            && opponent_target_bitboard2(1152921504606846976)==false
            && opponent_target_bitboard2(2305843009213693952)==false
            && opponent_target_bitboard2(4611686018427387904)==false
            && (pawns[1] & 69805794224242688L) == 0L) { // check if there's a pawn promotion move targeting an involved field

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=4611686018427387904L;
            mvSize[cd]++;
    }

    // white long, Q
    if (!onMove
            && (castlingRules&2)
            && (all & 1008806316530991104L) == 0L // all kann ersetzt werden
            && opponent_target_bitboard2(288230376151711744)==false 
            && opponent_target_bitboard2(576460752303423488)==false 
            && opponent_target_bitboard2(1152921504606846976)==false 
            && (pawns[1] & 17451448556060672L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=288230376151711744L;
            mvSize[cd]++;
    }

    // black short, k
    if (onMove
            && (castlingRules&4)
            && (all & 96L) == 0L // all kann ersetzt werden
            && opponent_target_bitboard2(16)==false 
            && opponent_target_bitboard2(32)==false 
            && opponent_target_bitboard2(64)==false 
            && (pawns[0] & 63488L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=64L;
            mvSize[cd]++;
    }

    // black long, q
    if (onMove
            && (castlingRules&8)
            && (all & 14L) == 0L // all kann ersetzt werden
            && opponent_target_bitboard2(4)==false 
            && opponent_target_bitboard2(8)==false 
            && opponent_target_bitboard2(16)==false 
            && (pawns[0] & 15872L) == 0L) {

            mvD[cd][mvSize[cd]][0]=kings[onMove];
            mvD[cd][mvSize[cd]][1]=4L;
            mvSize[cd]++;            
    }        
        
}

/************************************************
 * Funktion zur Generierung aller Capture-Züge. *
 * Wird in der Quiescence-Suche verwendet.      *
 ************************************************/
void Position::calcCaptures() {
    
    mvSize[cd]=0;
    
    unsigned long bitboard;    
    unsigned long opFigures;
    
    if (onMove) {
        // all figures bitboard from moving / opponent color
        opFigures=aw;
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_black_captures(x);
            bitboard-=Position::BIN[x];
        }        
    } else {
        // all figures bitboard from moving / opponent color
        opFigures=ab;        
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_white_captures(x);
            bitboard-=Position::BIN[x];
        }
    }
    
    /********
     * Turm *
     ********/
    bitboard = rooks[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves_captures__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    /**********
     * Läufer *
     **********/
    bitboard = bishops[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves_captures__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    /********
     * Dame *
     ********/
    bitboard = queens[onMove];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves_captures__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);                        
        split_moves_captures__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves_captures__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }

    /************
     * Springer *
     ************/
    bitboard = knights[onMove];
    while(bitboard != 0L) {
        int x=bitScanForward(bitboard);
        split_moves_direct_captures(x, Position::KN_TARGET3x, Position::KN_TARGET3x_FROM[x], Position::KN_TARGET3x_TO[x], opFigures);
        bitboard-=Position::BIN[x];
    }

    /*********
     * König *
     *********/
    int x = bitScanForward(kings[onMove]);
    split_moves_direct_captures(x, Position::K_TARGET2x, Position::K_TARGET2x_FROM[x], Position::K_TARGET2x_TO[x], opFigures);    
}

/****************************************************************
 * Generierte Capture-Züge für König und Springer aus aktueller *
 * Position und vorgefertigter Bitboards.                       *
 ****************************************************************/
inline void Position::split_moves_direct_captures(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & op) != 0L) { // only capture opponent color figures
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;
        }
    }
}

/****************************************************************
 * Generiert Züge für König und Springer aus aktueller Position *
 * und vorgefertigter Bitboards.                                *
 ****************************************************************/
inline void Position::split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & myFigures) == 0L) { // don't fight my color figures
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;
        }
    }
}

/*******************************************************************
 * Generiert Züge für Läufer, Turm und Dame aus aktueller Position *
 * und vorgefertigter Bitboards.                                   *
 *******************************************************************/
inline void Position::split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & all) == 0L) { // all kann ersetzt werden
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;                
        } else {
            if ((target[i] & op) != 0L) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[z];
                mvD[cd][mvSize[cd]][1]=target[i];
                mvSize[cd]++;                    
            }
            break;
        }
    }
}

/******************************************************************
 * Generiert Capture-Züge für Läufer, Turm und Dame aus aktueller *
 * Position und vorgefertigter Bitboards.                         *
 ******************************************************************/
inline void Position::split_moves_captures__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & all) == 0L) // all kann ersetzt werden
            continue;
        else {
            if ((target[i] & op) != 0L) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[z];
                mvD[cd][mvSize[cd]][1]=target[i];
                mvSize[cd]++;                    
            }
            break;
        }
    }
}

/**********************************************************
 * Die Züge der schwarzen Bauern zur Zugliste hinzufügen. *
 **********************************************************/
inline void Position::addPawnSingles_black(const int &z) {
    unsigned long startpos_black=65280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (Position::TARGET_BLACK_V1[z] & all) == 0UL) {// one step forward: just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                mvSize[cd]++;
                if ((Position::TARGET_BLACK_V1[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (Position::TARGET_BLACK_V3[z] & aw) != 0UL ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                mvSize[cd]++;        
                if ((Position::TARGET_BLACK_V3[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (Position::TARGET_BLACK_V4[z] & aw) != 0L ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_BLACK_V4[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;                    
                }
    }

    if (    (bitboardpos & startpos_black) != 0L && (Position::TARGET_BLACK_V2[z] & all) == 0L && (Position::TARGET_BLACK_V1[z] & all) == 0L) { // two steps forward: both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V2[z];
                mvSize[cd]++;        
    }    
    
    // en passant moves
    if (ep>8 && ep<=16) { // possible ep for black
        if (ep==9 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[40];
                mvSize[cd]++;
        }
        else if (ep==10 && bitboardpos==Position::BIN[32]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[32];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==10 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[39]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[39];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;
        }
        else if (ep==16 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[47];
                mvSize[cd]++;
        }
    }    
    
}

/*******************************************************
 * Die Züge der weißen Bauern zur Zugliste hinzufügen. *
 *******************************************************/
inline void Position::addPawnSingles_white(const int &z) {
    unsigned long startpos_white=71776119061217280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (TARGET_WHITE_V1[z] & all) == 0L) { // just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V1[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V1[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V1[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V1[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (TARGET_WHITE_V3[z] & ab) != 0L ) { // just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V3[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V3[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;                    
                }                
    }
    if (    (TARGET_WHITE_V4[z] & ab) != 0L ) {// just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V4[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;                    
                }                
    }
    
    if (    (bitboardpos & startpos_white) != 0L && (TARGET_WHITE_V2[z] & all) == 0L && (TARGET_WHITE_V1[z] & all) == 0L) { // both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V2[z];
                mvSize[cd]++;
    }   
    
    // en passant moves
    if (ep>0 && ep<=8) { // possible ep for white
        if (ep==1 && bitboardpos==Position::BIN[25]  ) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[16];
                mvSize[cd]++;            
        }
        else if (ep==2 && bitboardpos==Position::BIN[24]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[24];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==2 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[25]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[31]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[31];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==8 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[23];
                mvSize[cd]++;                        
        }
    }    
}

/******************************************************************
 * Die Capture-Züge der schwarzen Bauern zur Zugliste hinzufügen. *
 ******************************************************************/
inline void Position::addPawnSingles_black_captures(const int &z) {
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (Position::TARGET_BLACK_V3[z] & aw) != 0UL ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                mvSize[cd]++;        
                if ((Position::TARGET_BLACK_V3[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                    mvSize[cd]++;                    
                }
    }
    if (    (Position::TARGET_BLACK_V4[z] & aw) != 0L ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_BLACK_V4[z] & 18374686479671623680UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                    mvSize[cd]++;                    
                }
    }    
}

/***************************************************************
 * Die Capture-Züge der weißen Bauern zur Zugliste hinzufügen. *
 ***************************************************************/
inline void Position::addPawnSingles_white_captures(const int &z) {
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (TARGET_WHITE_V3[z] & ab) != 0L ) { // just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V3[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V3[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V3[z];
                    mvSize[cd]++;                    
                }                
    }
    if (    (TARGET_WHITE_V4[z] & ab) != 0L ) {// just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V4[z];
                mvSize[cd]++;
                if ((Position::TARGET_WHITE_V4[z] & 255UL) != 0L) {
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;
                    mvD[cd][mvSize[cd]][0]=bitboardpos;
                    mvD[cd][mvSize[cd]][1]=Position::TARGET_WHITE_V4[z];
                    mvSize[cd]++;                    
                }                
    }
    
}

/*********************************************************************
 * Prüft ob eine Figur an der übergebenen Position am Schachbrett    *
 * vom Gegner bedroht wird. Primär verwendet um zu überprüfen ob der *
 * König der aktuellen Farbe im Schach steht.                        *
 *********************************************************************/
bool Position::opponent_target_bitboard2(const unsigned long &bitboard) {
    
    unsigned long bb=bitboard;
    
    int x=bitScanForward(bb);
    
    bool opponent = !onMove;
    
    // pawns
    if (PAWN_ATTACKS[onMove][x] & pawns[opponent])
        return true;
    
    // knights (direkt)
    for (unsigned int i=KN_TARGET3x_FROM[x];i<KN_TARGET3x_TO[x];i++) {
        if (KN_TARGET3x[i] & knights[opponent])
            return true;
    }
    


    // north (linie)
    for (unsigned int i=TARGET_Nx_FROM[x];i<TARGET_Nx_TO[x];i++) {
        if (TARGET_Nx[i] & all) {
            if (TARGET_Nx[i] & rooks[opponent])
                return true;
            if (TARGET_Nx[i] & queens[opponent])
                return true;
            break;
        }
    }
    
    // east
    for (unsigned int i=TARGET_Ex_FROM[x];i<TARGET_Ex_TO[x];i++) {
        if (TARGET_Ex[i] & all) {
            if (TARGET_Ex[i] & rooks[opponent])
                return true;
            if (TARGET_Ex[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // south
    for (unsigned int i=TARGET_Sx_FROM[x];i<TARGET_Sx_TO[x];i++) {
        if (TARGET_Sx[i] & all) {
            if (TARGET_Sx[i] & rooks[opponent])
                return true;
            if (TARGET_Sx[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // west
    for (unsigned int i=TARGET_Wx_FROM[x];i<TARGET_Wx_TO[x];i++) {
        if (TARGET_Wx[i] & all) {
            if (TARGET_Wx[i] & rooks[opponent])
                return true;
            if (TARGET_Wx[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // north-east
    for (unsigned int i=TARGET_NEx_FROM[x];i<TARGET_NEx_TO[x];i++) {
        if (TARGET_NEx[i] & all) {
            if (TARGET_NEx[i] & bishops[opponent])
                return true;
            if (TARGET_NEx[i] & queens[opponent])
                return true;
            break;
        }
    }    

    // south-east
    for (unsigned int i=TARGET_SEx_FROM[x];i<TARGET_SEx_TO[x];i++) {
        if (TARGET_SEx[i] & all) {
            if (TARGET_SEx[i] & bishops[opponent])
                return true;
            if (TARGET_SEx[i] & queens[opponent])
                return true;
            break;
        }
    }        
    
    // south-west
    for (unsigned int i=TARGET_SWx_FROM[x];i<TARGET_SWx_TO[x];i++) {
        if (TARGET_SWx[i] & all) {
            if (TARGET_SWx[i] & bishops[opponent])
                return true;
            if (TARGET_SWx[i] & queens[opponent])
                return true;
            break;
        }
    }        
    
    // north-west
    for (unsigned int i=TARGET_NWx_FROM[x];i<TARGET_NWx_TO[x];i++) {
        if (TARGET_NWx[i] & all) {
            if (TARGET_NWx[i] & bishops[opponent])
                return true;
            if (TARGET_NWx[i] & queens[opponent])
                return true;
            break;
        }
    }        

    // king (direkt)
    for (unsigned int i=K_TARGET2x_FROM[x];i<K_TARGET2x_TO[x];i++) {
        if (K_TARGET2x[i] & kings[opponent])
            return true;
    }        
    
    return false; 
}

string Position::getFenShort() {
    string fen;
    
    // 1. group: figures
    int spaces=0;
    for (int i=1;i<=64;i++) {
        if ((Position::BIN[i-1]&all)!=0L) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            if ((kings[0]&Position::BIN[i-1])!=0L) fen+="K";
            else if ((queens[0]&Position::BIN[i-1])!=0L) fen+="Q";
            else if ((bishops[0]&Position::BIN[i-1])!=0L) fen+="B";
            else if ((knights[0]&Position::BIN[i-1])!=0L) fen+="N";
            else if ((rooks[0]&Position::BIN[i-1])!=0L) fen+="R";
            else if ((pawns[0]&Position::BIN[i-1])!=0L) fen+="P";
            else if ((kings[1]&Position::BIN[i-1])!=0L) fen+="k";
            else if ((queens[1]&Position::BIN[i-1])!=0L) fen+="q";
            else if ((bishops[1]&Position::BIN[i-1])!=0L) fen+="b";
            else if ((knights[1]&Position::BIN[i-1])!=0L) fen+="n";
            else if ((rooks[1]&Position::BIN[i-1])!=0L) fen+="r";
            else if ((pawns[1]&Position::BIN[i-1])!=0L) fen+="p";
        } else
            spaces++;

        if (i%8==0 && i<64) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            fen+="/";
        }
        if (i==64 && spaces>0) {
                fen+=int2String(spaces);
        }
            
    }
    
    return fen;
}

/************************************************************
 * Konvertiert die Daten der aktuell gespeicherten Position *
 * in einen FEN-String.                                     *
 ************************************************************/
string Position::getFen() {
    
    string fen;
    
    // 1. group: figures
    int spaces=0;
    for (int i=1;i<=64;i++) {
        if ((Position::BIN[i-1]&all)!=0L) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            if ((kings[0]&Position::BIN[i-1])!=0L) fen+="K";
            else if ((queens[0]&Position::BIN[i-1])!=0L) fen+="Q";
            else if ((bishops[0]&Position::BIN[i-1])!=0L) fen+="B";
            else if ((knights[0]&Position::BIN[i-1])!=0L) fen+="N";
            else if ((rooks[0]&Position::BIN[i-1])!=0L) fen+="R";
            else if ((pawns[0]&Position::BIN[i-1])!=0L) fen+="P";
            else if ((kings[1]&Position::BIN[i-1])!=0L) fen+="k";
            else if ((queens[1]&Position::BIN[i-1])!=0L) fen+="q";
            else if ((bishops[1]&Position::BIN[i-1])!=0L) fen+="b";
            else if ((knights[1]&Position::BIN[i-1])!=0L) fen+="n";
            else if ((rooks[1]&Position::BIN[i-1])!=0L) fen+="r";
            else if ((pawns[1]&Position::BIN[i-1])!=0L) fen+="p";
        } else
            spaces++;

        if (i%8==0 && i<64) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            fen+="/";
        }
        if (i==64 && spaces>0) {
                fen+=int2String(spaces);
        }
            
    }

    // 2. group: color on move
    if (onMove)
        fen+=" b";
    else
        fen+=" w";

    // 3. group: castles
    if (castlingRules==0)
        fen+=" -";
    else {
        fen+=" ";
        if (castlingRules&1)   fen+="K";
        if (castlingRules&2)    fen+="Q";
        if (castlingRules&4)   fen+="k";
        if (castlingRules&8)    fen+="q";        
    }

    // 4. group: en passant
    if (ep==0)  fen+=" -";
    else if(ep==1)  fen+=" a6";
    else if (ep==2) fen+=" b6";
    else if (ep==3) fen+=" c6";
    else if (ep==4) fen+=" d6";
    else if (ep==5) fen+=" e6";
    else if (ep==6) fen+=" f6";
    else if (ep==7) fen+=" g6";
    else if (ep==8) fen+=" h6";
    else if (ep==9)     fen+=" a3";
    else if (ep==10)    fen+=" b3";
    else if (ep==11)    fen+=" c3";
    else if (ep==12)    fen+=" d3";
    else if (ep==13)    fen+=" e3";
    else if (ep==14)    fen+=" f3";
    else if (ep==15)    fen+=" g3";
    else if (ep==16)    fen+=" h3";

    return fen;
}

/*********************************************************
 * konvertiert den Zug mit dem übergebenen Index in eine *
 * lesbare Form, zb: a3a4, g7g8q, o.ä.                   *
 *********************************************************/
string Position::getMoveString(int i) {
    string val;

    bool isPromotion=false;
    
    val+=get_hum(bitScanForward(mvD[cd][i][0]));
    val+=get_hum(bitScanForward(mvD[cd][i][1]));
    
    if ((mvD[cd][i][0]&pawns[onMove])!=0L && (mvD[cd][i][1]&promotionRow[onMove])!=0L)
        isPromotion=true;
    
    if (!isPromotion)
        return val;
    
    if ((i%4)==0)    
        val+="q";
    else if ((i%4)==1)    
        val+="r";
    else if ((i%4)==2)    
        val+="b";
    else if ((i%4)==3)
        val+="n";    
    
    return val;
}

/********************************************************
 * konvertiert den Zug mit dem übergebenen Index in die *
 * algebraische Kurznotation. Noch unvollständig.       *
 ********************************************************/
string Position::getMoveStringAlgShort(int i) {
    string val;
    val+=get_hum(bitScanForward(mvD[cd][i][1]));
    return val;
}

/******************************************
 * Konvertiert einen Index [0-63] in eine *
 * Schachbrett-Koordinate [a-h][1-8]      *
 ******************************************/
string Position::get_hum(int k) {
    string str;

    if (k%8==0) str+="a";
    if (k%8==1) str+="b";
    if (k%8==2) str+="c";
    if (k%8==3) str+="d";
    if (k%8==4) str+="e";
    if (k%8==5) str+="f";
    if (k%8==6) str+="g";
    if (k%8==7) str+="h";
    str+=int2String(8-k/8);
    return str;
}