/* 
 * File:   main.cpp
 * Author: peter
 *
 * Created on 31. Jänner 2012, 18:18
 */

#include <cstdlib>
#include <iostream>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <iomanip>
#include <math.h>
#include "Position.h"
#include "main.h"
#include "Perft.h"
#include "Egdb.h"
#include "globals.h"
#include "HashT.h"
#include "EpdSuite.h"

using namespace std;

FILE * pFile;

void log(string txt);

HashT hashTable;
Perft perft;

int main(int argc, char** argv) {

    cout << "argc=" << argc << endl;
    for (int i=0; i < argc; i++)
        cout << "argv[" << i << "]=" << argv[i] << endl;
    cout << endl;

    /*
    cout << "--------------------------------------------" << endl;
    cout << "--                                        --" << endl;
    cout << "--   Welcome to Magdalena                 --" << endl;
    cout << "--   Chess-Engine                         --" << endl;
    cout << "--                                        --" << endl;
    cout << "--   programmed by Peter Garscha          --" << endl;
    cout << "--   peter [at] rlsux [dot] com           --" << endl;
    cout << "--                                        --" << endl;
    cout << "--------------------------------------------" << endl << endl;
       */
    
    Position *myPosition = new Position();
    hashTable.init(pow(2,25));
    
    pFile = fopen ("/home/peter/magdalena-log.txt","a");
    if (pFile==NULL)
        cout << "can't write to log file" << endl;
    
    char input[256];
    string check;
    
    bool loop=true;
    while (loop) {
        check.clear();
        cin.getline(input,256);
        check.append(input);
        string logStr=check;
        log(logStr);
        
        if (check=="uci") {
            cout << "id name Magdalena\n";
            cout << "id author Peter Garscha\n";
            cout << "uciok\n";
        }
        
        else if (check.size()==0) {}
        
        else if (check=="isready")
            cout << "readyok\n";
        
        
        else if (check=="stop") {}
        
        else if (check=="debug on")
            cout << "debugging on now" << endl;
        
        else if (check=="debug off")
            cout << "debugging off now" << endl;
        
        else if (check=="ucinewgame")
            cout << "ucinewgame ok" << endl;
        
        else if (check=="perft")
            perft.doPerft((*myPosition).getFen(),5);

        else if (check=="perft all") {
            perft.doPerft("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1",4);
            perft.doPerft("r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq -",4);
            perft.doPerft("8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - -",4);
            perft.doPerft("r3k2r/Pppp1ppp/1b3nbN/nP6/BBP1P3/q4N2/Pp1P2PP/R2Q1RK1 w kq - 0 1",4);
            perft.doPerft("r2q1rk1/pP1p2pp/Q4n2/bbp1p3/Np6/1B3NBn/pPPP1PPP/R3K2R b KQ - 0 1",4);
        }
        
        else if (check=="perft list")
            perft.doPerftList();

        else if (check.substr(0,5)=="perft" && check.size()>5) {
            int depth=string2int(check.substr(6,check.size()));
            if (depth<8 && depth>0)
                perft.doPerft((*myPosition).getFen(),depth);
            else
                cout << "given depth is invalid for perft-test" << endl;
        }
        
        else if (check.substr(0,6)=="pseudo" && check.size()>6) {
            int depth=string2int(check.substr(7,check.size()));
            if (depth<8 && depth>0)
                perft.doPseudoPerft(*myPosition,depth);
            else
                cout << "given depth is invalid for pseudo-test" << endl;
        }
        
        else if (check=="eval" || check.substr(0,2)=="go") {
            // hashTable->updateAges();
            hashTable.clearHash();
            int sum=0;
            
            bool side2move=(*myPosition).onMove;
            
            timeval start,end;
            gettimeofday(&start,0);             
            int score = (*myPosition).negaMax(side2move,-100000, 100000, 1, 7, sum);
            gettimeofday(&end,0);
            double seconds = (end.tv_sec-start.tv_sec) + (end.tv_usec-start.tv_usec)*0.000001;            
            if (check=="eval") {
                // cout << "Move: " << move << endl;
                cout << "Move: " << (*myPosition).getMoveString((*myPosition).bestMoveIndex) << endl;
                cout << "Score: " << score << endl;
                cout << "Positions: " << sum << endl;
                cout.setf(ios::fixed, ios::floatfield);
                cout << "Time: " << setprecision(3) << seconds << " seconds" << endl;
            } else {
                // cout << "bestmove " << move << endl;
                cout << "bestmove " << (*myPosition).getMoveString((*myPosition).bestMoveIndex) << endl;
            }
        }
        
        else if (check=="egdb") {
            Egdb egdb;
            egdb.build();
        }
        
        else if (check=="print")
            (*myPosition).printBoard();
        
        else if (check=="info")
            (*myPosition).getInfo();
               
        else if (check=="epdsuite") {
            EpdSuite::testCompleteSuite();
            
        }

        else if (check.substr(0,6)=="divide") {
            bool modus=false;
            if (check.substr(0,7)=="divide2")
                modus=true;
            
            size_t found;
            found=check.find(" ");
            
            string strDepth = check.substr(int(found)+1,check.size()-int(found));
            int depth = atoi(strDepth.c_str());
            
            if (depth>0) {
                long sum=0;
                long total=0;
                int color;
                if ((*myPosition).onMove==true)
                    color=1;
                else color=0;
         
                (*myPosition).getDivide(color,1,depth,modus,sum,total);
                cout << "Nodes: " << total << endl;
            } else {
                cout << "no valid depth given. must be a valid number greater than 0" << endl;
            }
        }
        
        else if (check=="new") {
            delete myPosition;
            myPosition=new Position();
        }
        
        else if (check=="fen")
            cout << endl << "fen: " << (*myPosition).getFen() << endl << endl;
        
        else if (check.substr(0,13)=="position fen ") {
            string newFen = check.substr(13,check.size());
            delete myPosition;
            myPosition=new Position(newFen);
        }
        
        else if (check.substr(0,8)=="setboard") {
            if (check.size()>9) {
                string newFen = check.substr(9,check.size()-9);
                cout << newFen << endl;
                delete myPosition;
                myPosition=new Position(newFen);
            }
        }
        
        else if (check=="quit" || check=="exit") {
            cout << endl << "Goodbye!" << endl;
            loop=false;
        }
        
        else if (check=="help" || check=="?") {
            cout << "divide [n] ........... divide position with given depth" << endl;
            cout << "divide2 [n] .......... divide position with given depth" << endl;
            cout << "egdb ................. build end game database" << endl;
            cout << "eval ................. evaluate position" << endl;
            cout << "exit ................. exit engine" << endl;
            cout << "fen .................. print fen of current position" << endl;
            cout << "help ................. show this menu" << endl;
            cout << "info ................. print additional info about current position" << endl;
            cout << "new .................. reset position" << endl;
            cout << "perft [n] ............ standard performance test" << endl;
            cout << "                       (with current/given position)" << endl;
            cout << "perft all ............ performance test with 5 positions" << endl;
            cout << "perft list ........... test 126 positions to depth 6 (takes a while)" << endl;
            cout << "print ................ print board of current position" << endl;
            cout << "setboard [fen] ....... set current position to given fen" << endl;
            cout << "quit ................. exit engine" << endl;
            cout << "? .................... show this menu" << endl;
        }
        
        else {
            string moveArray[256];
            int cntMoves = myPosition->getMMChilds(moveArray);
            if (cntMoves==0) cout << "no valid move available or checkmate" << endl;
            bool found=false;
            for (int i=0;i<cntMoves;i++) {
                if (moveArray[i]==check) {
                    found=!found;
                    break;
                }
            }
            if (found) cout << "valid move" << endl;
            else cout << "invalid move / unknown command: " << check << endl;
        }
    }
    
    delete myPosition;
    
    if (pFile!=NULL)
        fclose(pFile);
   
    return 0;
}

void log(string txt) {
    if (txt.size()>0 && pFile!=NULL) {
        txt.append("\n");
        fputs (txt.data(),pFile);
    }
}