/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionEvalFastQuiesce.cpp
 * Author: peter
 * 
 * Created on 20. Februar 2014, 18:32
 */
#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "main.h"
#include "Timer.h"
#include "PositionEvalFastQuiesce.h"

using namespace std;

extern Timer timer;

int PositionEvalFastQuiesce::negaMax(int alpha, int beta, int curr_depth, int depth) { //, bool canNull) {
    sum++;   
    cd = curr_depth;
    if (!timer.isTimeout() && (sum&1023)==0)
        timer.checkTimeout();    
    if (timer.isTimeout())
        return 0;
    if (curr_depth==depth)
//        return evaluate(onMove);
        return quiesce(alpha,beta,curr_depth);
    
//    moveList.setMoveAt(curr_depth,hashKey,"",0);
    
    /*********************************************************************************
     * Überprüfe ob die aktuelle Position bereits im Spiel- bzw. Suchverlauf vorkam. *
     * Ggf. wird ein Score für ein Draw bzw. ein entsprechender Contempt-Faktor      *
     * zurück gegeben.                                                               *
     *********************************************************************************/

//    if (cd>1 && moveList.isRepetition(cd)) {
//        int contemptOpening=-50;
//        int contemptMiddlegame=-25;
//        int contempt=0;
//        if (moveList.getNumHalfMoves()<20)
//            contempt=contemptOpening;
//        else if (!isEndgame())
//            contempt=contemptMiddlegame;
//        if (cd%2==1)
//            contempt=-contempt;
//        return contempt;
//    }
    
    /****************************************************************************
     * Suche in der HashTable nach der aktuellen Position und prüfe ggf. ob der *
     * gespeicherte Score verwendet werden kann.                                *
     ****************************************************************************/
    int bestmoveHash=-1;
    

    int score=0;  
 
    calc_moves();
    unsigned short int sortedIndizes[256];
    sortMoves(sortedIndizes,cd,bestmoveHash);
    
    doBackup();
    int maxWert=alpha;    
    for (int i=mvSize[cd]-1;i>=0;i--) {
        unsigned short int ret = move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4);
        if (curr_depth==0) {
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]); // TODO wollen wir das in jeder tiefe?
            onMove=!onMove;
            if (catchKing) {
                undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);
                continue;
            }
        }

        score = -negaMax(-beta,-maxWert,(curr_depth+1),depth);
        cd=curr_depth;
        undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);        
       
        if (score > maxWert) {
            maxWert = score;
            if (curr_depth==0)
                bestMoveIndex=sortedIndizes[i];
            if (maxWert>=beta) {
                maxWert=beta;
                break;
            }
        }
    }
    return maxWert;
}

int PositionEvalFastQuiesce::quiesce(int alpha, int beta, int curr_depth) {
       
    cd = curr_depth;

    sum++;

    if (!timer.isTimeout() && (sum&1023)==0)
        timer.checkTimeout();    
    if (timer.isTimeout())
        return 0;    
    
    
    int standPat = evaluate(onMove);
    if (standPat>=beta)
        return beta;
    if(standPat>alpha)
        alpha=standPat;

    calcCaptures();
    if (!mvSize[cd])
        return alpha;

    int bestmoveHash=-1;
    unsigned short int sortedIndizes[256];
    sortMoves(sortedIndizes,cd,bestmoveHash);

//    unsigned short int moveScores[256];
//    getMoveScores(moveScores,cd);
    
    doBackup();
    for (int i=mvSize[cd]-1;i>=0;i--) {    
        
//        if ((mvD[cd][sortedIndizes[i]][1]&kings[!onMove])!=0)
//            return 20000;
        
        unsigned short int ret=move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4);
        int score = -quiesce(-beta,-alpha,(curr_depth+1));
        cd--;
        undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);
        if (score > alpha) {
            if (score>=beta) 
                return beta;
            alpha=score;
        }
    }
    return alpha;
}