/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionMonteCarloTreeSearch.h
 * Author: peter
 *
 * Created on 20. Februar 2014, 18:32
 */

#ifndef POSITIONMONTECARLOTREESEARCH_H
#define	POSITIONMONTECARLOTREESEARCH_H

#include <stdlib.h>
#include <time.h>
#include "Position.h"

//! A Monte-Carlo-Tree-Search Game-Mode
/*!
 * This class doesn't have an evaluation-function, but uses
 * a monte-carlo-tree-search-method to calculate the best move
 * from a given position. This class uses the MCTSNode-Class.
*/
class PositionMonteCarloTreeSearch : public Position {
public:
    string getName() {return "MonteCarloTreeSearch"; };
    void go(unsigned int depth);
    int getMCTSChilds(string moveArray[], int moveIndex[], unsigned long moveFrom[], unsigned long moveTo[], int movePromotion[]);
    int doRandomMove();
    int doRandomMoveSimilarity();
    int doRandomMoveSimilarityA();
    int doRandomMoveSimilarityB();
    void evaluateSkips(int mode);
    void sortChilds();
//    void quicksort(int left, int right);
//    int quicksortSplit(int left, int right);
//    void swapMoves(int i, int k);
    double skipSortValue[256];
    unsigned short int sortedIndizes[256];
};

#endif	/* POSITIONMONTECARLOTREESEARCH_H */