/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionPureMonteCarloCatch.cpp
 * Author: peter
 * 
 * Created on 20. Februar 2014, 18:32
 */
#include <iostream>
#include <iomanip> 
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "main.h"
#include "Timer.h"
#include "PositionPureMonteCarloCatch.h"

using namespace std;

extern Timer timer;

void PositionPureMonteCarloCatch::go(unsigned int depth) {
    (void)depth; // depth wird hier nicht benötigt
    srand(time(NULL));
    string moveArray[256];
    int moveIndex[256];
    int cntMoves = getMMChildsCatch(moveArray,moveIndex);
    if (cntMoves==0) {
        cntMoves = getMMChilds(moveArray,moveIndex);
    }
    
    int cd=0;
    doBackup();
    double resValues[cntMoves];
    while(1) {
        if (timer.isTimeout() || timer.checkTimeout())
            break;
        for (int i=0;i<cntMoves;i++) {
            int ret=move_figure(mvD[cd][moveIndex[i]][0],mvD[cd][moveIndex[i]][1],moveIndex[i]%4);
            playoutPosition(i,10);
            undoMove(mvD[cd][moveIndex[i]][0],mvD[cd][moveIndex[i]][1],moveIndex[i]%4,ret);
        }
    }

    for (int i=0;i<cntMoves;i++) {
        resValues[i]=(double)result[i][0]/result[i][1];
    }
    double minValue=resValues[0];
    double maxValue=resValues[0];
    int minValueIndex=0;
    int maxValueIndex=0;
    for (int i=1;i<cntMoves;i++) {
        if (resValues[i]>maxValue) {
            maxValue=resValues[i];
            maxValueIndex=i;
        } else if (resValues[i]<minValue) {
            minValue=resValues[i];
            minValueIndex=i;
        }
    }
    
    /************************************
     * Das Ergebnis der Suche ausgeben. *
     ************************************/
    timer.stopTimer();
    double seconds = timer.getStartEndDiffSeconds();
    cout << "Time: " << seconds << " seconds" << endl;
    onMove==false?cout << "bestmove " << moveArray[maxValueIndex] << endl:cout << "bestmove " << moveArray[minValueIndex] << endl;
}

//int PositionPureMonteCarloCatch::doRandomMove() {
//    cd=0;
//    calc_moves();
//    int cntMoves=mvSize[cd];
//    if (cntMoves==0)
//        return onMove==0?1:2; // nomove for white=1, nomove for black=2;
//    int randomMove=rand()%cntMoves;
//    move_figure(mvD[0][randomMove][0],mvD[0][randomMove][1],randomMove%4);
//    return 0;
//}
//
//bool PositionPureMonteCarloCatch::isKingWhiteLost() {
//    return (kings[0]==0);
//}
//
//bool PositionPureMonteCarloCatch::isKingBlackLost() {
//    return (kings[1]==0);
//}
//
//void PositionPureMonteCarloCatch::pmcTest(const string &currentFen) {
//    cout << "Pure-MonteCarlo-Test" << endl;
//    srand(time(NULL));
//    cout << "FEN: " << currentFen << endl;
//    int gamesToPlay=10000;
//    int cntWhiteWins=0;
//    int cntBlackWins=0;
//    int cntNoMovesWhite=0;
//    int cntNoMovesBlack=0;
//    timer.resetTimer(0,0);
//    for (int i=0;i<gamesToPlay;i++) {
//        PositionPureMonteCarloCatch *pmc;
//        pmc = new PositionPureMonteCarloCatch();
//        pmc->setFen(currentFen);
//        int cntMoves=0;
//        while(1) {
//            int noMove=pmc->doRandomMove();
//            cntMoves++;            
//            if (noMove==1) {
//                cntNoMovesWhite++;
//                break;
//            } else if (noMove==2) {
//                cntNoMovesBlack++;
//                break;
//            } else if  (cntMoves>80)
//                break;
//            else if (pmc->isKingWhiteLost()) {
//                cntBlackWins++;
//                break;
//            } else if (pmc->isKingBlackLost()) {
//                cntWhiteWins++;
//                break;
//            }
//        }
//        delete pmc;
//    }    
//    timer.stopTimer();
//    double seconds = timer.getStartEndDiffSeconds();
//    cout << "Games: " << gamesToPlay << endl;
//    cout << "White wins: " << cntWhiteWins << endl;
//    cout << "Black wins: " << cntBlackWins << endl;
//    cout << "No Moves White: " << cntNoMovesWhite << endl;
//    cout << "No Moves Black: " << cntNoMovesBlack << endl;
//    cout << "Time: " << seconds << " seconds" << endl;    
//    cout << "=============================";
//    cout << "Legal Moves:" << endl;
//    PositionPureMonteCarloCatch *pmc = new PositionPureMonteCarloCatch();
//    pmc->setFen(currentFen);
//    string moveArray[256];
//    int moveIndex[256];
//    int cntMoves = pmc->getMMChilds(moveArray,moveIndex);
//    int cd=0;
//    pmc->doBackup();
//    double resValues[cntMoves];
//    for (int i=0;i<cntMoves;i++) {
//        int ret=pmc->move_figure(pmc->mvD[cd][moveIndex[i]][0],pmc->mvD[cd][moveIndex[i]][1],moveIndex[i]%4);
//        pmc->playoutPosition(i,1000);
//        resValues[i]=(double)pmc->result[i][0]/pmc->result[i][1];
//        cout << moveArray[i] << "\t" << std::setprecision(5) << resValues[i] << "\t\t" << pmc->getFen() << endl;
//        pmc->undoMove(pmc->mvD[cd][moveIndex[i]][0],pmc->mvD[cd][moveIndex[i]][1],moveIndex[i]%4,ret);
//    }
//    double minValue=resValues[0];
//    double maxValue=resValues[0];
//    int minValueIndex=0;
//    int maxValueIndex=0;
//    for (int i=1;i<cntMoves;i++) {
//        if (resValues[i]>maxValue) {
//            maxValue=resValues[i];
//            maxValueIndex=i;
//        } else if (resValues[i]<minValue) {
//            minValue=resValues[i];
//            minValueIndex=i;
//        }
//    }
//    pmc->onMove==false?cout << "bestmove " << moveArray[maxValueIndex] << endl:cout << "bestmove " << moveArray[minValueIndex] << endl;
//}
//
//void PositionPureMonteCarloCatch::playoutPosition(int moveIndex, int games) {
//    string currentFen = getFen();
//    srand(time(NULL));
//    int gamesToPlay=games;
//    int cntWhiteWins=0;
//    int cntBlackWins=0;
//    for (int i=0;i<gamesToPlay;i++) {
//        PositionPureMonteCarloCatch *pmc;
//        pmc = new PositionPureMonteCarloCatch();
//        pmc->setFen(currentFen);
//        int cntMoves=0;
//        while(1) {
//            pmc->doRandomMove();
//            cntMoves++;
//            if  (cntMoves>80)
//                break;
//            else if (pmc->isKingWhiteLost()) {
//                cntBlackWins++;
//                break;
//            } else if (pmc->isKingBlackLost()) {
//                cntWhiteWins++;
//                break;
//            }
//        }
//        delete pmc;
//    }
//    result[moveIndex][0]=cntWhiteWins;
//    result[moveIndex][1]=cntBlackWins;
//}
