/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionUCTSearchSimilarityB.cpp
 * Author: peter
 * 
 * Created on 20. Februar 2014, 18:32
 */
#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "main.h"
#include "Timer.h"
#include "PositionUCTSearchSimilarityB.h"
#include "UCTNodeSimilarityB.h"

using namespace std;

extern Timer timer;

void PositionUCTSearchSimilarityB::go(unsigned int depth) {
    (void)depth; // depth wird hier nicht benötigt
    srand(time(NULL));
    UCTNodeSimilarityB *rootNode = new UCTNodeSimilarityB();
    int iterations=0;
    while(1) {
        PositionUCTSearchSimilarity *rootPos = new PositionUCTSearchSimilarityB(*this);
        UCTNodeSimilarityB *selected = rootNode->select(rootPos);
        UCTNodeSimilarityB *expanded = selected->expand(rootPos);
        double result = expanded->simulate(rootPos);
        expanded->update(result);
        iterations++;
        if (iterations%100==0) {
            if (timer.isTimeout() || timer.checkTimeout())
                break;
        }
        delete rootPos;
    }
    string foundMove=rootNode->getMove();
    delete rootNode;
    
    /************************************
     * Das Ergebnis der Suche ausgeben. *
     ************************************/
    timer.stopTimer();
    double seconds = timer.getStartEndDiffSeconds();
    cout << "Move: " << foundMove << endl;
    cout << "Time: " << seconds << " seconds" << endl;
    cout << "bestmove " << foundMove << endl;
}