/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   UCTNodeSimilarityA.h
 * Author: peter
 *
 * Created on 10. April 2014, 07:45
 */

#ifndef UCTNODESIMILARITYA_H
#define	UCTNODESIMILARITYA_H

#include <string>
#include <iostream>
#include <stdio.h>
#include <cstdlib>
#include <cmath>
#include "PositionMonteCarloTreeSearch.h"

using namespace std;

//! Node Class for UCT-like Monte Carlo Tree Search with Similarity
/*!
 * Node Class for UCT-like Monte Carlo Tree Search. This class also handles
 * the basic MCTS-functions like select, expand, simulate and update. This class
 * is similar to MCTS-Node, but has modified select-function.
 * Additionally to UCT, it uses a hash-position-similarity function
 * to prune the game-tree.
*/
class UCTNodeSimilarityA {
public:
    UCTNodeSimilarityA();
    UCTNodeSimilarityA(string lastMoveString, unsigned long lastMoveFrom, unsigned long lastMoveTo, int lastMovePromotion, UCTNodeSimilarityA *parent);
    string getMove(); // most visited child
    virtual ~UCTNodeSimilarityA();
    UCTNodeSimilarityA* select(PositionMonteCarloTreeSearch *rootPos);
    UCTNodeSimilarityA* expand(PositionMonteCarloTreeSearch *rootPos);
    double simulate(PositionMonteCarloTreeSearch *rootPos);
    void update(double value);
    bool hasChildren();
    bool isNotFullyExpanded();
    UCTNodeSimilarityA* getChild(int i);
    UCTNodeSimilarityA* createAndAddChild(string lastMoveString, unsigned long lastMoveFrom, unsigned long lastMoveTo, int lastMovePromotion, PositionMonteCarloTreeSearch *rootPos);
//    static const unsigned int movestats[256][256];
//    static const unsigned int movestatsGood[256][256];
//    static const double movestatsGoodDouble[256][256];
    static const int similarMoves[256][8];
private:
    void evaluateSkips(PositionMonteCarloTreeSearch *currentPos);
    void sortChilds();
    void quicksort(int left, int right);
    int quicksortSplit(int left, int right);
    // move data
    unsigned long lastMoveFrom;
    unsigned long lastMoveTo;
    int lastMovePromotion;
    string lastMoveString;
    // node data
    int cntMoves;
    int numChilds;
    double totalValue;
    double visits;
    UCTNodeSimilarityA *parent;
    UCTNodeSimilarityA *childs[256];
    double skipSortValue;
};

#endif	/* UCTNODESIMILARITYA_H */