/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   main.cpp
 * Author: peter
 *
 * Created on 31. Jänner 2012, 18:18
 */

#include <cstdlib>
#include <iostream>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <iomanip>
#include <math.h>
#include <vector>
#include <ctime>
#include "Position.h"
#include "Timer.h"
#include "string"
#include "main.h"
#include <sstream>
#include "PositionEval.h"
#include "PositionEvalFast.h"
#include "PositionEvalQuiesce.h"
#include "PositionEvalFastQuiesce.h"
#include "PositionKingCheck.h"
#include "PositionAspiration.h"
#include "PositionRandom.h"
#include "PositionRandomCatch.h"
#include "PositionPureMonteCarlo.h"
#include "PositionPureMonteCarloCatch.h"
#include "PositionMonteCarloTreeSearch.h"
#include "PositionUCTSearch.h"
#include "PositionUCTSearchSimilarity.h"
#include "PositionUCTSearchSimilarityA.h"
#include "PositionUCTSearchSimilarityB.h"
#include "PositionUCTPlayoutSimilarity.h"
#include "PositionUCTPlayoutSimilarityA.h"
#include "PositionUCTPlayoutSimilarityB.h"

using namespace std;

void doPseudoPerft(Position &pos, int max_depth);
void doPureMonteCarloTest();
void getHash(string check);

Timer timer;
Position *myPosition;
unsigned short int distances[64][64];
int engineMode=6; // aspiration-mode is default game play mode, seems to be the strongest
string uctfile="";

int main(int argc, char** argv) {
    
    bool beSilent=false;
    
    for (int i=0; i < argc; i++) {
        string arg="";
        arg.append(argv[i]);
        if (arg.substr(0,6)=="silent")
            beSilent=true;
    }        
    
    if (!beSilent) {    
        cout << "***************************************************" << endl;
        cout << "* Madlenka Copyright (C) 2012-2014 Peter Garscha  *" << endl;
        cout << "* This program comes with ABSOLUTELY NO WARRANTY. *" << endl;
        cout << "* This is free software, and you are welcome to   *" << endl;
        cout << "* redistribute it under certain conditions.       *" << endl;
        cout << "* See the files COPYING and README coming with    *" << endl;
        cout << "* the source code of this program.                *" << endl;
        cout << "***************************************************" << endl << endl;
        cout << "███╗   ███╗ █████╗ ██████╗ ██╗     ███████╗███╗   ██╗██╗  ██╗ █████╗ " << endl;
        cout << "████╗ ████║██╔══██╗██╔══██╗██║     ██╔════╝████╗  ██║██║ ██╔╝██╔══██╗" << endl;
        cout << "██╔████╔██║███████║██║  ██║██║     █████╗  ██╔██╗ ██║█████╔╝ ███████║" << endl;
        cout << "██║╚██╔╝██║██╔══██║██║  ██║██║     ██╔══╝  ██║╚██╗██║██╔═██╗ ██╔══██║" << endl;
        cout << "██║ ╚═╝ ██║██║  ██║██████╔╝███████╗███████╗██║ ╚████║██║  ██╗██║  ██║" << endl;
        cout << "╚═╝     ╚═╝╚═╝  ╚═╝╚═════╝ ╚══════╝╚══════╝╚═╝  ╚═══╝╚═╝  ╚═╝╚═╝  ╚═╝" << endl << endl;
        cout << "Madlenka is a Chess-Engine for Linux." << endl;
        cout << "type 'help' for a list of available commands" << endl;
        cout << "or 'uci' to enter the uci-mode." << endl << endl;
    }
    
    for (int i=0; i < argc; i++) {
        string arg="";
        arg.append(argv[i]);
        if (arg.substr(0,4)=="mode" && i<argc-1)
            setMode(argv[i+1]);
        if (arg.substr(0,7)=="uctfile" && i<argc-1)
            uctfile=argv[i+1];
    }    
    
    initEngine();
    setPosition("position startpos");
    
    if (!beSilent)
        cout << "engine mode: " << myPosition->getName() << endl << endl;
    
    char input[4096];
    string check;
    bool ucimode=false;
    bool loop=true;
    while (loop) {
        check.clear();
        if (!ucimode && !beSilent)
            cout << "madlenka> ";
        cin.getline(input,4096);
        check.append(input);
        check = nicerCmd(check);
        
        if (check=="uci") {
            cout << "\n";
            cout << "id name Madlenka - " << myPosition->getName() << "\n";
            cout << "id author Peter Garscha\n";
            cout << "uciok\n";
            ucimode=true;
        }
        
        else if (check.size()==0) {}
        
        else if (check=="isready")
            cout << "readyok\n";
        
        else if (check=="stop") {}
        
        else if (check=="debug on")
            cout << "debugging on now" << endl;
        
        else if (check=="debug off")
            cout << "debugging off now" << endl;
        
        else if (check.substr(0,6)=="divide")
            divide(check);
        
        else if (check.substr(0,7)=="gethash")
            getHash(check);
        
        else if (check=="ucinewgame")
            cout << "ucinewgame ok" << endl;

        else if (check.substr(0,4)=="eval" || check.substr(0,2)=="go")
            evalgo(check);
        
        else if (check=="print")
            (*myPosition).printBoard();
        
        else if (check=="info")
            (*myPosition).getInfo();
        
        else if (check=="fen")
            cout << "fen: " << (*myPosition).getFen() << endl;
        
        else if (check=="perftlist")
            doPerftList();        
        
        else if (check=="pmc")
            doPureMonteCarloTest();
        
        else if (check.substr(0,8)=="position")
            setPosition(check);
        
        else if (check.substr(0,6)=="pseudo" && check.size()>6) {
            int depth=string2int(check.substr(7,check.size()));
            if (depth<10 && depth>0)
                doPseudoPerft(*myPosition,depth);
            else
                cout << "given depth is invalid for pseudo-test" << endl;
        }        
        
        else if (check=="quit" || check=="exit") {
            cout << endl << "Goodbye!" << endl;
            loop=false;
        }
        
        else if (check=="help" || check=="?") {
            cout << "divide [depth] ....... nodes to depth" << endl;
            cout << "divide2 [depth] ...... accumulated nodes to depth" << endl;
            cout << "eval ................. evaluate position" << endl;
            cout << "exit ................. exit engine" << endl;
            cout << "fen .................. print fen of current position" << endl;
            cout << "help ................. show this menu" << endl;
            cout << "info ................. print additional info about current position" << endl;
            cout << "perftlist ............ check 126 perft positions (move-generator correctness)" << endl;
            cout << "print ................ print board of current position" << endl;
            cout << "pseudo [depth] ....... speed test (pseudo legal)" << endl;
            cout << "pmc .................. pure monte carlo test" << endl;
            cout << "quit ................. exit engine" << endl;
            cout << "? .................... show this menu" << endl;            
        }
        
        else {
            string moveArray[256];
            int moveIndex[256];
            int cntMoves = myPosition->getMMChilds(moveArray,moveIndex);
            if (cntMoves==0)
                cout << "no valid move available or checkmate" << endl;
            bool found=false;
            for (int i=0;i<cntMoves;i++) {
                if (moveArray[i]==check) {
                    found=!found;
                    break;
                }
            }
            if (found)
                cout << "valid move" << endl;
        }
    }

    delete myPosition;
    return 0;
}

int string2int(string str) {
    char const *c = str.c_str();
    int k = atoi(c);
    return k;
}

long string2long(string str) {
    char const *c=str.c_str();
    long k=atol(c);    
    return k;    
}

string int2String(int value) {
    string s;
    stringstream out;
    out << value;
    s = out.str();
    return s;
}

bool figOnPos(const unsigned long &figure, const int &position) {
    return ((figure & Position::BIN[position]) != 0L);
}

string nicerCmd(string command) {

    while (true) {
        int pos = command.find("\n");
        if (pos>=0) command=command.substr(0,pos)+" "+command.substr(pos+1);
        else break;
    }    
    
    while (true) {
        int pos = command.find("\\n");
        if (pos>=0) command=command.substr(0,pos)+command.substr(pos+2);
        else break;
    }

    while (true) {
        int pos = command.find("\t");
        if (pos>=0) command=command.substr(0,pos)+" "+command.substr(pos+1);
        else break;
    }    
    
    while (true) {
        int pos = command.find("\\t");
        if (pos>=0) command=command.substr(0,pos)+command.substr(pos+2);
        else break;
    }    
    
    while (true) {
        int pos = command.find("  ");
        if (pos>=0) command=command.substr(0,pos)+command.substr(pos+1);
        else break;
    }            
    
    return command;
}

void initEngine() {
    for (int i=0;i<64;i++) {
        for (int j=0;j<64;j++) {
            int row_i=i/8;
            int row_j=j/8;
            int col_i=i%8;
            int col_j=j%8;
            distances[i][j]=14-(abs(row_i-row_j)+abs(col_i-col_j))/2;
        }
    }  
}

void setPosition(string check) {
    vector <string> fields;
    while(true) {
        int pos=check.find_first_of(" ");
        if (pos>=0) {
            fields.push_back(check.substr(0,pos));
            check=check.substr(pos+1);
        } else {
            fields.push_back(check);
            break;
        }
    }

    if (fields.size()>2 && fields[1]=="fen") {
        string newFen=fields[2];
        for (unsigned int i=3;i<fields.size();i++) {
            if (fields[i]!="moves") {
                newFen.append(" ");
                newFen.append(fields[i]);
            } else
                break;
        }
        cout << newFen << endl;
        newPosition();
        myPosition->setFen(newFen);
    }

    if (fields.size()>1 && fields[1]=="startpos") {
        newPosition();
    }

    bool doMoves=false;
    for (unsigned int i=0;i<fields.size();i++) {
        if (!doMoves && fields[i]=="moves") {
            doMoves=true;
            continue;
        }
        if (doMoves) {
            string moveArray[256];
            int moveIndex[256];
            int cntMoves = myPosition->getMMChilds(moveArray, moveIndex);
            bool found=false;
            int realMoveIndex;
            for (int k=0;k<cntMoves;k++) {
                if (moveArray[k]==fields[i]) {
                    found=!found;
                    realMoveIndex=moveIndex[k]; // notwendig, weil nicht immer alle Züge gültig sind
                    break;
                }
            }
            if (found)
                myPosition->move_figure(myPosition->mvD[0][realMoveIndex][0],myPosition->mvD[0][realMoveIndex][1],realMoveIndex%4);
        }               
    }
}

void setMode(string mode) {
    int modeint=string2int(mode);
    if (modeint>=1 && modeint<=18)
        engineMode=modeint;
//    if (mode=="1")
//        engineMode=string2int(mode);
//    else if (mode=="2")
//        engineMode=string2int(mode);
//    else if (mode=="3")
//        engineMode=string2int(mode);
//    else if (mode=="4")
//        engineMode=string2int(mode);
//    else if (mode=="5")
//        engineMode=string2int(mode);
//    else if (mode=="6")
//        engineMode=string2int(mode);
//    else if (mode=="7")
//        engineMode=string2int(mode);
//    else if (mode=="8")
//        engineMode=string2int(mode);
//    else if (mode=="9")
//        engineMode=string2int(mode);
//    else if (mode=="10")
//        engineMode=string2int(mode);
//    else if (mode=="11")
//        engineMode=string2int(mode);
//    else if (mode=="12")
//        engineMode=string2int(mode);
//    else if (mode=="13")
//        engineMode=string2int(mode);
//    else if (mode=="14")
//        engineMode=string2int(mode);
//    else if (mode=="15")
//        engineMode=string2int(mode);
//    else if (mode=="16")
//        engineMode=string2int(mode);
//    else if (mode=="17")
//        engineMode=string2int(mode);
//    else if (mode=="18")
//        engineMode=string2int(mode);
}

void newPosition() {
    delete myPosition;
    switch(engineMode) {
        case 1:
            myPosition=new PositionEvalFast();
            break;
        case 2:
            myPosition=new PositionEval();
            break;
        case 3:
            myPosition=new PositionEvalFastQuiesce();
            break;
        case 4:
            myPosition=new PositionEvalQuiesce();
            break;
        case 5:
            myPosition=new PositionKingCheck();
            break;
        case 6:
            myPosition=new PositionAspiration();
            break;
        case 7:
            myPosition=new PositionRandom();
            break;
        case 8:
            myPosition=new PositionRandomCatch();
            break;
        case 9:
            myPosition=new PositionPureMonteCarlo();
            break;
        case 10:
            myPosition=new PositionPureMonteCarloCatch();
            break;
        case 11:
            myPosition=new PositionMonteCarloTreeSearch();
            break;
        case 12:
            myPosition=new PositionUCTSearch();
            break;
        case 13:
            myPosition=new PositionUCTSearchSimilarity();
            break;
        case 14:
            myPosition=new PositionUCTSearchSimilarityA();
            break;
        case 15:
            myPosition=new PositionUCTSearchSimilarityB();
            break;
        case 16:
            myPosition=new PositionUCTPlayoutSimilarity();
            break;
        case 17:
            myPosition=new PositionUCTPlayoutSimilarityA();
            break;
        case 18:
            myPosition=new PositionUCTPlayoutSimilarityB();
            break;
    }
}

void divide(string check) {
    bool modus=false;
    if (check.substr(0,7)=="divide2")
        modus=true;

    size_t found;
    found=check.find(" ");

    string strDepth = check.substr(int(found)+1,check.size()-int(found));
    int depth = atoi(strDepth.c_str());

    if (depth>0) {
        long sum=0;
        long total=0;
        int color;
        if ((*myPosition).onMove==true)
            color=1;
        else color=0;

        (*myPosition).getDivide(color,1,depth,modus,sum,total);
        cout << "Nodes: " << total << endl;
    } else {
        cout << "no valid depth given. must be a valid number greater than 0" << endl;
    }    
}

void doPseudoPerft(Position &pos, int max_depth) {
    timeval start,end;
    gettimeofday(&start,0);    
    
    long sum=0;
    // pos.getChildsPseudo(max_depth-1,sum);
    pos.getChildsPseudo(max_depth-1,sum);
    
    gettimeofday(&end,0);
    double seconds = (end.tv_sec-start.tv_sec) + (end.tv_usec-start.tv_usec)*0.000001;
    
    cout.setf(ios::fixed, ios::floatfield);
    cout << endl << "perft result (" << max_depth << "):" << endl;
    cout << "nodes: " << sum << endl;
    cout << "time: " << setprecision(3) << seconds << " seconds, nps: " << setprecision(0) << double(sum/seconds) << endl << endl;    
}

void doPerftList() {
    string fenlist[][7]={   
        {"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1","20","400","8902","197281","4865609","119060324"},
        {"r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq - 0 1","48","2039","97862","4085603","193690690","8031647685"},
        {"4k3/8/8/8/8/8/8/4K2R w K - 0 1","15","66","1197","7059","133987","764643"},
        {"4k3/8/8/8/8/8/8/R3K3 w Q - 0 1","16","71","1287","7626","145232","846648"},
        {"4k2r/8/8/8/8/8/8/4K3 w k - 0 1","5","75","459","8290","47635","899442"},
        {"r3k3/8/8/8/8/8/8/4K3 w q - 0 1","5","80","493","8897","52710","1001523"},
        {"4k3/8/8/8/8/8/8/R3K2R w KQ - 0 1","26","112","3189","17945","532933","2788982"},
        {"r3k2r/8/8/8/8/8/8/4K3 w kq - 0 1","5","130","782","22180","118882","3517770"},
        {"8/8/8/8/8/8/6k1/4K2R w K - 0 1","12","38","564","2219","37735","185867"},
        {"8/8/8/8/8/8/1k6/R3K3 w Q - 0 1","15","65","1018","4573","80619","413018"},
        {"4k2r/6K1/8/8/8/8/8/8 w k - 0 1","3","32","134","2073","10485","179869"},
        {"r3k3/1K6/8/8/8/8/8/8 w q - 0 1","4","49","243","3991","20780","367724"},
        {"r3k2r/8/8/8/8/8/8/R3K2R w KQkq - 0 1","26","568","13744","314346","7594526","179862938"},
        {"r3k2r/8/8/8/8/8/8/1R2K2R w Kkq - 0 1","25","567","14095","328965","8153719","195629489"},
        {"r3k2r/8/8/8/8/8/8/2R1K2R w Kkq - 0 1","25","548","13502","312835","7736373","184411439"},
        {"r3k2r/8/8/8/8/8/8/R3K1R1 w Qkq - 0 1","25","547","13579","316214","7878456","189224276"},
        {"1r2k2r/8/8/8/8/8/8/R3K2R w KQk - 0 1","26","583","14252","334705","8198901","198328929"},
        {"2r1k2r/8/8/8/8/8/8/R3K2R w KQk - 0 1","25","560","13592","317324","7710115","185959088"},
        {"r3k1r1/8/8/8/8/8/8/R3K2R w KQq - 0 1","25","560","13607","320792","7848606","190755813"},
        {"4k3/8/8/8/8/8/8/4K2R b K - 0 1","5","75","459","8290","47635","899442"},
        {"4k3/8/8/8/8/8/8/R3K3 b Q - 0 1","5","80","493","8897","52710","1001523"},
        {"4k2r/8/8/8/8/8/8/4K3 b k - 0 1","15","66","1197","7059","133987","764643"},
        {"r3k3/8/8/8/8/8/8/4K3 b q - 0 1","16","71","1287","7626","145232","846648"},
        {"4k3/8/8/8/8/8/8/R3K2R b KQ - 0 1","5","130","782","22180","118882","3517770"},
        {"r3k2r/8/8/8/8/8/8/4K3 b kq - 0 1","26","112","3189","17945","532933","2788982"},
        {"8/8/8/8/8/8/6k1/4K2R b K - 0 1","3","32","134","2073","10485","179869"},
        {"8/8/8/8/8/8/1k6/R3K3 b Q - 0 1","4","49","243","3991","20780","367724"},
        {"4k2r/6K1/8/8/8/8/8/8 b k - 0 1","12","38","564","2219","37735","185867"},
        {"r3k3/1K6/8/8/8/8/8/8 b q - 0 1","15","65","1018","4573","80619","413018"},
        {"r3k2r/8/8/8/8/8/8/R3K2R b KQkq - 0 1","26","568","13744","314346","7594526","179862938"},
        {"r3k2r/8/8/8/8/8/8/1R2K2R b Kkq - 0 1","26","583","14252","334705","8198901","198328929"},
        {"r3k2r/8/8/8/8/8/8/2R1K2R b Kkq - 0 1","25","560","13592","317324","7710115","185959088"},
        {"r3k2r/8/8/8/8/8/8/R3K1R1 b Qkq - 0 1","25","560","13607","320792","7848606","190755813"},
        {"1r2k2r/8/8/8/8/8/8/R3K2R b KQk - 0 1","25","567","14095","328965","8153719","195629489"},
        {"2r1k2r/8/8/8/8/8/8/R3K2R b KQk - 0 1","25","548","13502","312835","7736373","184411439"},
        {"r3k1r1/8/8/8/8/8/8/R3K2R b KQq - 0 1","25","547","13579","316214","7878456","189224276"},
        {"8/1n4N1/2k5/8/8/5K2/1N4n1/8 w - - 0 1","14","195","2760","38675","570726","8107539"},
        {"8/1k6/8/5N2/8/4n3/8/2K5 w - - 0 1","11","156","1636","20534","223507","2594412"},
        {"8/8/4k3/3Nn3/3nN3/4K3/8/8 w - - 0 1","19","289","4442","73584","1198299","19870403"},
        {"K7/8/2n5/1n6/8/8/8/k6N w - - 0 1","3","51","345","5301","38348","588695"},
        {"k7/8/2N5/1N6/8/8/8/K6n w - - 0 1","17","54","835","5910","92250","688780"},
        {"8/1n4N1/2k5/8/8/5K2/1N4n1/8 b - - 0 1","15","193","2816","40039","582642","8503277"},
        {"8/1k6/8/5N2/8/4n3/8/2K5 b - - 0 1","16","180","2290","24640","288141","3147566"},
        {"8/8/3K4/3Nn3/3nN3/4k3/8/8 b - - 0 1","4","68","1118","16199","281190","4405103"},
        {"K7/8/2n5/1n6/8/8/8/k6N b - - 0 1","17","54","835","5910","92250","688780"},
        {"k7/8/2N5/1N6/8/8/8/K6n b - - 0 1","3","51","345","5301","38348","588695"},
        {"B6b/8/8/8/2K5/4k3/8/b6B w - - 0 1","17","278","4607","76778","1320507","22823890"},
        {"8/8/1B6/7b/7k/8/2B1b3/7K w - - 0 1","21","316","5744","93338","1713368","28861171"},
        {"k7/B7/1B6/1B6/8/8/8/K6b w - - 0 1","21","144","3242","32955","787524","7881673"},
        {"K7/b7/1b6/1b6/8/8/8/k6B w - - 0 1","7","143","1416","31787","310862","7382896"},
        {"B6b/8/8/8/2K5/5k2/8/b6B b - - 0 1","6","106","1829","31151","530585","9250746"},
        {"8/8/1B6/7b/7k/8/2B1b3/7K b - - 0 1","17","309","5133","93603","1591064","29027891"},
        {"k7/B7/1B6/1B6/8/8/8/K6b b - - 0 1","7","143","1416","31787","310862","7382896"},
        {"K7/b7/1b6/1b6/8/8/8/k6B b - - 0 1","21","144","3242","32955","787524","7881673"},
        {"7k/RR6/8/8/8/8/rr6/7K w - - 0 1","19","275","5300","104342","2161211","44956585"},
        {"R6r/8/8/2K5/5k2/8/8/r6R w - - 0 1","36","1027","29215","771461","20506480","525169084"},
        {"7k/RR6/8/8/8/8/rr6/7K b - - 0 1","19","275","5300","104342","2161211","44956585"},
        {"R6r/8/8/2K5/5k2/8/8/r6R b - - 0 1","36","1027","29227","771368","20521342","524966748"},
        {"6kq/8/8/8/8/8/8/7K w - - 0 1","2","36","143","3637","14893","391507"},
        {"6KQ/8/8/8/8/8/8/7k b - - 0 1","2","36","143","3637","14893","391507"},
        {"K7/8/8/3Q4/4q3/8/8/7k w - - 0 1","6","35","495","8349","166741","3370175"},
        {"6qk/8/8/8/8/8/8/7K b - - 0 1","22","43","1015","4167","105749","419369"},
        {"6KQ/8/8/8/8/8/8/7k b - - 0 1","2","36","143","3637","14893","391507"},
        {"K7/8/8/3Q4/4q3/8/8/7k b - - 0 1","6","35","495","8349","166741","3370175"},
        {"8/8/8/8/8/K7/P7/k7 w - - 0 1","3","7","43","199","1347","6249"},
        {"8/8/8/8/8/7K/7P/7k w - - 0 1","3","7","43","199","1347","6249"},
        {"K7/p7/k7/8/8/8/8/8 w - - 0 1","1","3","12","80","342","2343"},
        {"7K/7p/7k/8/8/8/8/8 w - - 0 1","1","3","12","80","342","2343"},
        {"8/2k1p3/3pP3/3P2K1/8/8/8/8 w - - 0 1","7","35","210","1091","7028","34834"},
        {"8/8/8/8/8/K7/P7/k7 b - - 0 1","1","3","12","80","342","2343"},
        {"8/8/8/8/8/7K/7P/7k b - - 0 1","1","3","12","80","342","2343"},
        {"K7/p7/k7/8/8/8/8/8 b - - 0 1","3","7","43","199","1347","6249"},
        {"7K/7p/7k/8/8/8/8/8 b - - 0 1","3","7","43","199","1347","6249"},
        {"8/2k1p3/3pP3/3P2K1/8/8/8/8 b - - 0 1","5","35","182","1091","5408","34822"},
        {"8/8/8/8/8/4k3/4P3/4K3 w - - 0 1","2","8","44","282","1814","11848"},
        {"4k3/4p3/4K3/8/8/8/8/8 b - - 0 1","2","8","44","282","1814","11848"},
        {"8/8/7k/7p/7P/7K/8/8 w - - 0 1","3","9","57","360","1969","10724"},
        {"8/8/k7/p7/P7/K7/8/8 w - - 0 1","3","9","57","360","1969","10724"},
        {"8/8/3k4/3p4/3P4/3K4/8/8 w - - 0 1","5","25","180","1294","8296","53138"},
        {"8/3k4/3p4/8/3P4/3K4/8/8 w - - 0 1","8","61","483","3213","23599","157093"},
        {"8/8/3k4/3p4/8/3P4/3K4/8 w - - 0 1","8","61","411","3213","21637","158065"},
        {"k7/8/3p4/8/3P4/8/8/7K w - - 0 1","4","15","90","534","3450","20960"},
        {"8/8/7k/7p/7P/7K/8/8 b - - 0 1","3","9","57","360","1969","10724"},
        {"8/8/k7/p7/P7/K7/8/8 b - - 0 1","3","9","57","360","1969","10724"},
        {"8/8/3k4/3p4/3P4/3K4/8/8 b - - 0 1","5","25","180","1294","8296","53138"},
        {"8/3k4/3p4/8/3P4/3K4/8/8 b - - 0 1","8","61","411","3213","21637","158065"},
        {"8/8/3k4/3p4/8/3P4/3K4/8 b - - 0 1","8","61","483","3213","23599","157093"},
        {"k7/8/3p4/8/3P4/8/8/7K b - - 0 1","4","15","89","537","3309","21104"},
        {"7k/3p4/8/8/3P4/8/8/K7 w - - 0 1","4","19","117","720","4661","32191"},
        {"7k/8/8/3p4/8/8/3P4/K7 w - - 0 1","5","19","116","716","4786","30980"},
        {"k7/8/8/7p/6P1/8/8/K7 w - - 0 1","5","22","139","877","6112","41874"},
        {"k7/8/7p/8/8/6P1/8/K7 w - - 0 1","4","16","101","637","4354","29679"},
        {"k7/8/8/6p1/7P/8/8/K7 w - - 0 1","5","22","139","877","6112","41874"},
        {"k7/8/6p1/8/8/7P/8/K7 w - - 0 1","4","16","101","637","4354","29679"},
        {"k7/8/8/3p4/4p3/8/8/7K w - - 0 1","3","15","84","573","3013","22886"},
        {"k7/8/3p4/8/8/4P3/8/7K w - - 0 1","4","16","101","637","4271","28662"},
        {"7k/3p4/8/8/3P4/8/8/K7 b - - 0 1","5","19","117","720","5014","32167"},
        {"7k/8/8/3p4/8/8/3P4/K7 b - - 0 1","4","19","117","712","4658","30749"},
        {"k7/8/8/7p/6P1/8/8/K7 b - - 0 1","5","22","139","877","6112","41874"},
        {"k7/8/7p/8/8/6P1/8/K7 b - - 0 1","4","16","101","637","4354","29679"},
        {"k7/8/8/6p1/7P/8/8/K7 b - - 0 1","5","22","139","877","6112","41874"},
        {"k7/8/6p1/8/8/7P/8/K7 b - - 0 1","4","16","101","637","4354","29679"},
        {"k7/8/8/3p4/4p3/8/8/7K b - - 0 1","5","15","102","569","4337","22579"},
        {"k7/8/3p4/8/8/4P3/8/7K b - - 0 1","4","16","101","637","4271","28662"},
        {"7k/8/8/p7/1P6/8/8/7K w - - 0 1","5","22","139","877","6112","41874"},
        {"7k/8/p7/8/8/1P6/8/7K w - - 0 1","4","16","101","637","4354","29679"},
        {"7k/8/8/1p6/P7/8/8/7K w - - 0 1","5","22","139","877","6112","41874"},
        {"7k/8/1p6/8/8/P7/8/7K w - - 0 1","4","16","101","637","4354","29679"},
        {"k7/7p/8/8/8/8/6P1/K7 w - - 0 1","5","25","161","1035","7574","55338"},
        {"k7/6p1/8/8/8/8/7P/K7 w - - 0 1","5","25","161","1035","7574","55338"},
        {"3k4/3pp3/8/8/8/8/3PP3/3K4 w - - 0 1","7","49","378","2902","24122","199002"},
        {"7k/8/8/p7/1P6/8/8/7K b - - 0 1","5","22","139","877","6112","41874"},
        {"7k/8/p7/8/8/1P6/8/7K b - - 0 1","4","16","101","637","4354","29679"},
        {"7k/8/8/1p6/P7/8/8/7K b - - 0 1","5","22","139","877","6112","41874"},
        {"7k/8/1p6/8/8/P7/8/7K b - - 0 1","4","16","101","637","4354","29679"},
        {"k7/7p/8/8/8/8/6P1/K7 b - - 0 1","5","25","161","1035","7574","55338"},
        {"k7/6p1/8/8/8/8/7P/K7 b - - 0 1","5","25","161","1035","7574","55338"},
        {"3k4/3pp3/8/8/8/8/3PP3/3K4 b - - 0 1","7","49","378","2902","24122","199002"},
        {"8/Pk6/8/8/8/8/6Kp/8 w - - 0 1","11","97","887","8048","90606","1030499"},
        {"n1n5/1Pk5/8/8/8/8/5Kp1/5N1N w - - 0 1","24","421","7421","124608","2193768","37665329"},
        {"8/PPPk4/8/8/8/8/4Kppp/8 w - - 0 1","18","270","4699","79355","1533145","28859283"},
        {"n1n5/PPPk4/8/8/8/8/4Kppp/5N1N w - - 0 1","24","496","9483","182838","3605103","71179139"},
        {"8/Pk6/8/8/8/8/6Kp/8 b - - 0 1","11","97","887","8048","90606","1030499"},
        {"n1n5/1Pk5/8/8/8/8/5Kp1/5N1N b - - 0 1","24","421","7421","124608","2193768","37665329"},
        {"8/PPPk4/8/8/8/8/4Kppp/8 b - - 0 1","18","270","4699","79355","1533145","28859283"},
        {"n1n5/PPPk4/8/8/8/8/4Kppp/5N1N b - - 0 1","24","496","9483","182838","3605103","71179139"}
    };
    
    int maxDepth=6;
    for (int i=0;i<126;i++) {
        long results[6]={string2long(fenlist[i][1]),string2long(fenlist[i][2]),string2long(fenlist[i][3]),string2long(fenlist[i][4]),string2long(fenlist[i][5]),string2long(fenlist[i][6])};   
        string fen = fenlist[i][0];        
        newPosition();
        myPosition->setFen(fen);
        Perftres *result = new Perftres(maxDepth);
        myPosition->getPerftInfo(0,maxDepth,result);
        cout << fen << "          ";
        for (int i=0;i<maxDepth;i++) {
            if (result->perftres[7*i]==results[i]) cout << "OK  ";
            else {
                cout << "ERR(" << result->perftres[7*i] << "," << results[i] << ")  ";
            }
        }
        cout << endl;
        delete result;
    }
}

void evalgo(string check) {
    /***********************************************
     * Das übergebene Command verarbeiten und die  *
     * einzelnen Teile im vector fields speichern. *
     ***********************************************/
    vector <string> fields;

    while(true) {
        int pos=check.find_first_of(" ");
        if (pos>=0) {
            fields.push_back(check.substr(0,pos));
            check=check.substr(pos+1);
        } else {
            fields.push_back(check);
            break;
        }
    }            

    unsigned int depth=32;
    unsigned int btime=0;
    unsigned int wtime=0;
    unsigned int movestogo=0;

    for (unsigned int i=0;i<fields.size();i++) {
        if (fields[i]=="depth" && (i+1)<fields.size()) {
            unsigned int newDepth=string2int(fields[i+1]);
            if (newDepth>0)
                depth=newDepth;
            continue;
        } else if (fields[i]=="wtime" && (i+1)<fields.size()) {
            wtime=string2int(fields[i+1]);
            continue;
        } else if (fields[i]=="btime" && (i+1)<fields.size()) {
            btime=string2int(fields[i+1]);
            continue;
        } else if (fields[i]=="movestogo" && (i+1)<fields.size()) {
            movestogo=string2int(fields[i+1]);
            continue;
        }                
    }

    cout << "search params: wtime=" << wtime << ", btime=" << btime << ", movestogo=" << movestogo << ", depth=" << depth << endl;

    /***************************************************
     * Suche vorbereiten, dh. TranspositionAge erhöhen *
     * TimeManagement festlegen, usw.                  *
     ***************************************************/
    bool side2move=(*myPosition).onMove;

    if (movestogo==0)   movestogo=1;

    unsigned long timeForMove;
    if (side2move && btime)             timeForMove=btime;
    else if (!side2move && wtime)       timeForMove=wtime;
    else                                timeForMove=9999999;

    timer.resetTimer(timeForMove,movestogo);

    myPosition->go(depth);    
}

void doPureMonteCarloTest() {
    string currentFen=myPosition->getFen();
    PositionPureMonteCarlo::pmcTest(currentFen);
}

void getHash(string check) {
    
    // get fen- and move-part, set position to fen
    string fen=check.substr(8);
    myPosition->setFen(fen);
    fen=myPosition->getFen();
    string move=check.substr(check.find(",")+1);
    
    // evaluate position
    bool promotionPossible=myPosition->isPromotionPossible();
    bool rochadePossible=myPosition->isRochadePossible();
    bool capturePossible=myPosition->isCapturePossible();
    bool isEndgame=myPosition->isEndgame();
    bool hasQueen=myPosition->hasQueen();
    bool hasBothRooks=myPosition->hasBothRooks();
    bool hasBothBishops=myPosition->hasBothBishops();
    bool hasBothKnights=myPosition->hasBothKnights();
    
    // endcode booleans in an 8-bit-char
    int positionHash=0;
    positionHash^=promotionPossible;
    positionHash<<=1;
    positionHash^=rochadePossible;
    positionHash<<=1;
    positionHash^=capturePossible;
    positionHash<<=1;
    positionHash^=isEndgame;
    positionHash<<=1;
    positionHash^=hasQueen;
    positionHash<<=1;
    positionHash^=hasBothRooks;
    positionHash<<=1;
    positionHash^=hasBothBishops;
    positionHash<<=1;
    positionHash^=hasBothKnights;
    
    // evaluate move
    
    string moveArray[256];
    int moveIndex[256];
    int cntMoves=myPosition->getMMChilds(moveArray,moveIndex);
    unsigned long moveFrom=0, moveTo=0;
    for (int i=0;i<cntMoves;i++) {
        if (moveArray[i]==move) {
            moveFrom=myPosition->mvD[0][moveIndex[i]][0];
            moveTo=myPosition->mvD[0][moveIndex[i]][1];
            break;
        }
    }
        
    bool isKingMove=myPosition->getKingUnsignedLong(myPosition->onMove)&moveFrom;
    bool isQueenMove=myPosition->getQueenUnsignedLong(myPosition->onMove)&moveFrom;
    bool isPawnMove=myPosition->getPawnsUnsignedLong(myPosition->onMove)&moveFrom;
    bool isBishopMove=myPosition->getBishopsUnsignedLong(myPosition->onMove)&moveFrom;
    bool isRookMove=myPosition->getRooksUnsignedLong(myPosition->onMove)&moveFrom;
    bool isKnightMove=myPosition->getKnightsUnsignedLong(myPosition->onMove)&moveFrom;
    bool isMoveCenter=moveTo&66229406269440UL;
    bool isMoveCapture=myPosition->getAllUnsignedLong()&moveTo; // en-passant wird hier nicht betrachtet
//    bool isRochade=false;
//    bool isPromotion=false;
//    bool isMoveForward=false;
//    bool isMoveBackward=false;

    // endcode booleans in an 8-bit-char
    int moveHash=0;
    moveHash^=isKingMove;
    moveHash<<=1;
    moveHash^=isQueenMove;
    moveHash<<=1;
    moveHash^=isPawnMove;
    moveHash<<=1;
    moveHash^=isBishopMove;
    moveHash<<=1;
    moveHash^=isRookMove;
    moveHash<<=1;
    moveHash^=isKnightMove;
    moveHash<<=1;
    moveHash^=isMoveCenter;
    moveHash<<=1;
    moveHash^=isMoveCapture;
    
    // return results
    cout<<"("<<promotionPossible<<","<<rochadePossible<<","<<capturePossible<<","<<isEndgame<<","<<
    hasQueen<<","<<hasBothRooks<<","<<hasBothBishops<<","<<hasBothKnights<<")|";
    
    cout << positionHash << "|" << move << "|" << "(" << moveFrom << "," << moveTo << ")|" << moveHash << endl;
}

/*! \mainpage Overview
 *
 * \section about About Madlenka
 *
 * Madlenka is an UCI-OpenSource-Chess-Engine developed for Linux.\n\n
 * The SourceCode of Madlenka can be found on github.com:\n
 * https://github.com/petergar/madlenka\n\n
 * If you have any questions or comments to my engine,
 * don't hesitate to contact me!
 * \n\n
 * e-mail: peter [at] rlsux [dot] com
 *
 */